
// Simple utils library
//    2005 gama

#include <string.h>
#include <ctype.h>
#include <stdlib.h>
#include <pspiofilemgr.h>

#include "utils.h"

int is_ext(const char *s, const char *ext)
{
	const char *sext = strrchr(s, '.');
    return sext && stricmp(ext, sext) == 0;
}

int is_zip(const char *s)
{
	return is_ext(s, ".zip") || is_ext(s, ".rar") || is_ext(s, ".rsn");
}

int is_playlist(const char *s)
{
    return is_ext(s, ".m3u");
}

// TODO - fix warning "pointer target type"
char *get_filename( const char *path )
{
	char *fname = strrchr( path, '/' );
	if( fname == NULL ) return path;
    return strrchr( path, '/' ) + 1;
}

char *get_prefix_extension( const char *filename, int to_upper )
{
    char ext[256];
    char *tmp = get_basename(filename);

	// Remove any remaining sufixes
	if( tmp && strchr(tmp, '.') ) {
		char tmp2[256];
		strncpy(tmp2, tmp, 256);
		free(tmp);

		tmp = get_basename(tmp2);
	}

    if( tmp ) {
		int len = strlen(tmp), i;
        for(i = 0; i < len; i++) {
			ext[i] = to_upper ? toupper( tmp[i] ) : tmp[i];
		}
		ext[i] = '\0';

		return strdup(ext);
    }

	return strdup("<null>");
}

char *get_extension( const char *filename, int to_upper )
{
	const unsigned int ext_len = 10;
	char ext[ext_len];

	const char* p = strrchr( filename, '.' );
	if ( p && strlen( p ) <= ext_len ) {
		int i;
  		for ( i = 0; i < strlen(p); i++ )
   	        ext[i] = to_upper ? toupper( p[i+1] ) : p [i+1];
	}
	else return strdup("<null>");

	return strdup( ext );
}

int is_extension( const char *filename, const char *ext )
{
	char *sext = get_extension(filename, 1);
	int len1 = strlen(sext), len2 = strlen(sext);
	int result = ((len1 == len2) && (strcmp(ext, sext) == 0)) ? 1 : 0;
	free( sext );
	return result;
}

int is_prefix_extension( const char *filename, const char *ext )
{
	char *prefix = get_prefix_extension(filename, 1);
	int len1 = strlen(prefix), len2 = strlen(ext);
	int result = ((len1 == len2) && strcmp(ext, prefix) == 0) ? 1 : 0;
	free(prefix);
	return result;
}

char *get_basename( const char *filename )
{
    char base[256];
    char *tmp = strrchr(filename, '.');
    if( tmp ) {
        strncpy(base, filename, strlen(filename) - strlen(tmp));
		base[strlen(filename)-strlen(tmp)] = '\0';

		char *tmp2 = strrchr(base, '/');
        if( tmp2 ) strcpy(base, tmp2+1);
		return strdup(base);
    }

	// Then remove path
	tmp = strrchr(filename, '/');
    if( tmp ) {
		strcpy(base, tmp+1);
		return strdup(base);
	}

	// then just return the same file name
	return strdup(filename);
}

char *get_basedir( const char *filename )
{
	char base[256];
    char *tmp = strrchr(filename, '/');
    if( tmp ) {
        strncpy(base, filename, strlen(filename) - strlen(tmp));
		base[strlen(filename)-strlen(tmp)] = '\0';
		return strdup(base);
    }

	return strdup("<null>");
}

int strcount(const char *str, char c)
{
	int count = 0, i;
	int len = strlen(str);
	for(i = 0; i < len; i++) if( str[i] == c ) count++;

	return count;
}

#ifdef PSP
static char *i2a(unsigned i, char *a, unsigned r)
{
	if (i/r > 0) a = i2a(i/r,a,r);
	*a = "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ"[i%r];
	return a+1;
}

char *itoa(int i, char *a, int r)
{
	if ((r < 2) || (r > 36)) r = 10;
	if (i<0) {
		*a = '-';
		*i2a(-(unsigned)i,a+1,r) = 0;
	} else *i2a(i,a,r) = 0;
	return a;
}
#endif

int _fgets(SceUID fd, char *line, int max_len)
{
	int i = 0, rc = 0;
	char c;
	while( (rc = sceIoRead(fd, &c, 1)) > 0 ) {
		if( c == '\n' ) break;
		line[i] = c;
		// check max length
		if( ++i >= max_len ) break;
	}

	line[i] = '\0';
	return rc;
}

int utf8_strlen(const char *str)
{
	int len = 0;
	unsigned char c, *ptr = (unsigned char *)str;
	while( *ptr != '\0' ) {
		c = *ptr & 0xe0;
        if (c < 0x80)
            ptr++;
        else if (c < 0xe0)
            ptr += 2;
        else if (c < 0xf0)
            ptr += 3;

		len ++;
	}

	return len;
}

void utf8_strncpy(char *dst, const char *src, int maxchars)
{
	unsigned char c, *dptr, *sptr;
	int count = 0;

	dptr = (unsigned char *)dst;
	sptr = (unsigned char *)src;

	while( (*sptr != '\0') && (count < maxchars) ) {
		c = *sptr & 0xe0;
        if (c < 0x80) {
			*dptr++ = *sptr++;
		}
        else if (c < 0xe0) {
			*dptr++ = *sptr++;
			*dptr++ = *sptr++;
		}
        else if (c < 0xf0) {
			*dptr++ = *sptr++;
			*dptr++ = *sptr++;
			*dptr++ = *sptr++;
		}

		count ++;
	}

	*dptr = '\0';
}

void utf8_strncat(char *dst, const char *src, int maxchars)
{
	utf8_strncpy(dst + strlen(dst), src, maxchars);
}

int utf8_offset(const char *str, int offset)
{
	int uoffset = 0;
	unsigned char c, *ptr = (unsigned char *)str;
	while( *ptr != '\0' && offset > 0 ) {
		c = *ptr & 0xe0;
        if (c < 0x80) {
            ptr++;
			uoffset++;
		}
        else if (c < 0xe0) {
            ptr += 2;
			uoffset += 2;
		}
        else if (c < 0xf0) {
            ptr += 3;
			uoffset += 3;
		}

		offset--;
	}

	return uoffset;
}
