
/*  Input Plugin for GameMusicGear ~
 *  Copyright (C) 2005 - 2010 gama, rockMax
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/* Default headers */
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <pspkernel.h>
#include "../plugin.h"

#include "faad.h"
#include "mp4ff.h"

#define TAGLIB
#ifdef TAGLIB
#include "taglib/fileref.h"
#include "taglib/tag.h"
#include "taglib/toolkit/tbytevector.h"
#include "taglib/mpeg/mpegfile.h"
#include "taglib/mpeg/id3v2/id3v2tag.h"
#include "taglib/mpeg/id3v2/id3v2frame.h"
#include "taglib/mpeg/id3v2/id3v2header.h"
#include "taglib/mpeg/id3v2/frames/attachedpictureframe.h"
using namespace TagLib;
#endif

// LINKER PATCH
extern "C" {
void *__dso_handle = NULL;
}

#ifdef PSP
PSP_MODULE_INFO("IN_AAC", 0, 1, 1);
PSP_NO_CREATE_MAIN_THREAD();
PSP_HEAP_SIZE_KB( 4 * 1024 );

#ifdef __cplusplus
extern "C" {
#endif
int module_start(SceSize argc, void* argp)
{
 	return 0;
}

int module_stop(int args, void *argp)
{
	return 0;
}
#ifdef __cplusplus
}
#endif
#endif

/* Prototypes */
void init(void);
void reset(void);
int  load( char *filename );
int  load_rw( char *filename, char *addr, int size );
long  seek( long time );
void cleanup( void );
int  write_buffer( signed short *dest, int len );
void get_song_title( char *filename, char **title, int *length );

void get_packet(void);
void send_packet(void);
/* End Prototypes */

/* Variables */
#define SAFE_FREE(x)		if(x) free(x); x = NULL;
#define MAX_CHANNELS 2
typedef struct {
    long bytes_into_buffer;
    long bytes_consumed;
    long file_offset;
    unsigned char *buffer;
    int at_eof;
    FILE *infile;
} aac_buffer;

static int adts_sample_rates[] =
{96000,88200,64000,48000,44100,32000,24000,22050,16000,12000,11025,8000,7350,0,0,0};

// AAC Stuff
NeAACDecHandle         hDecoder;
NeAACDecFrameInfo      frameInfo;
NeAACDecConfiguration *config;
aac_buffer 			   b;

// MP4 Stuff
mp4ff_t 			  *infile = NULL;
mp4ff_callback_t 	  *mp4cb  = NULL;
mp4AudioSpecificConfig mp4ASC;
long sampleId = 0, numSamples;
int track;

FILE 		  *mp4File = NULL;
unsigned char *buffer = NULL;
unsigned int   buffer_size;

// audio buffer stuff
void *tmpbuf = NULL;
int writeposbytes = 0;
int readposbytes  = 0;
int tagsize = 0;

// mp4 file or aac file identifier
int mp4file = 0;

int playing = 0;
int channels = 2;
int samplerate = 44100;
AFormat aformat = FMT_S16_LE;
int kbps = 128;

char title[128];
char artist[128];
char album[128];
char copyright[128];
char format[128];

// extra info
// CHECKME - this must be static?
char *pic_data = NULL;
int   pic_size;

long length;
int subtunes;
/* End Variables */

/** START of Plugin definitions setup */
InputPlugin itable =
{
	/* Populated by Plugin */
	"Advanced Audio Coding",				/* The description of the input plugin */

	init,							/* Called when the plugin is loaded */
	reset, 							/* Called when need to update config */
	load,							/* Load the file */
	NULL,							/* Load the file from data */
	NULL,							/* Set song subtune */
	NULL,							/* Seek to the specified time */
	NULL,							/* Set the equalizer, most plugins won't be able to do this */
	cleanup,						/* Called when gmgear exit */
	write_buffer, 					/* Fill audio buffer */
	get_song_title,					/* Function to grab the title string */

	get_packet,					/* Called when need main app data packet*/
	send_packet,				/* Send data packet to main app */

	NULL,						/* Filled by Main App */
	NULL,						/* Filled by Main App */
	NULL,						/* Filled by Main App */
	0,								/* Filled by Main App */
	LATIN,							/* Filled by Main App */
	CLIENT_PLAYER					/* Player type */
};
/** END of Plugin definitions setup */

/** Single Plugin Export */
#ifdef PSP
InputPlugin *get_iplugin_info()
#else
InputPlugin *dll_get_iplugin_info()
#endif
{
	return &itable;
}

void init(void)
{
	strcpy(format, "AAC");
	subtunes = 1;

	kbps = 0x6C;
	aformat = FMT_S16_LE;

	hDecoder = NULL;
	memset(&b, 0, sizeof(aac_buffer));

	itable.in_packet->ID = NO_PACKET;
}

void reset(void)
{
	strcpy(title, "");
	strcpy(album, "");
	strcpy(copyright, "");
	strcpy(artist, "");

	length = 0;

	// Cleanup
	if( pic_data ) {
		delete [] pic_data;
		pic_data = NULL;
	}

	if( hDecoder ) {
		NeAACDecClose(hDecoder);
		hDecoder = NULL;
	}

	if( mp4file ) {
        if( infile ) {
			mp4ff_close(infile);
			infile = NULL;
		}

		SAFE_FREE(mp4cb);
		SAFE_FREE(buffer);
        if( mp4File ) {
			fclose(mp4File);
			mp4File = NULL;
		}

		sampleId = 0;
	}
	else {
		if (b.infile)
		{
			fclose(b.infile);
			b.infile = NULL;
		}
		SAFE_FREE(b.buffer);
	}

	tmpbuf = NULL;
	itable.in_packet->ID = NO_PACKET;
}

/* EXTRA FUNCTIONS */

static int fill_b_buffer(aac_buffer *b);
static void advance_buffer(aac_buffer *b, int bytes);

int adts_parse(aac_buffer *b, int *bitrate, float *length)
{
    int frames, frame_length;
    int t_framelength = 0;
    int sample_rate = 44100;
    float frames_per_sec, bytes_per_frame;

    /* Read all frames to ensure correct time and bitrate */
    for (frames = 0; /* */; frames++)
    {
        fill_b_buffer(b);

        if (b->bytes_into_buffer > 7)
        {
            /* check syncword */
            if (!((b->buffer[0] == 0xFF)&&((b->buffer[1] & 0xF6) == 0xF0)))
                break;

            if (frames == 0)
                samplerate = adts_sample_rates[(b->buffer[2]&0x3c)>>2];

            frame_length = ((((unsigned int)b->buffer[3] & 0x3)) << 11)
                | (((unsigned int)b->buffer[4]) << 3) | (b->buffer[5] >> 5);

            t_framelength += frame_length;

            if (frame_length > b->bytes_into_buffer)
                break;

            advance_buffer(b, frame_length);
        } else {
            break;
        }
    }

    frames_per_sec = (float)sample_rate/1024.0f;
    if (frames != 0)
        bytes_per_frame = (float)t_framelength/(float)(frames*1000);
    else
        bytes_per_frame = 0;
    *bitrate = (int)(8.0f * bytes_per_frame * frames_per_sec + 0.5f);
    if (frames_per_sec != 0)
        *length = (float)frames/frames_per_sec;
    else
        *length = 1;

    return 1;
}

uint32_t read_callback(void *user_data, void *buffer, uint32_t length)
{
    return fread(buffer, 1, length, (FILE*)user_data);
}

uint32_t seek_callback(void *user_data, uint64_t position)
{
    return fseek((FILE*)user_data, position, SEEK_SET);
}

// FAAD file buffering routines
static int foo = 1;
static int fill_b_buffer(aac_buffer *b)
{
    int bread;

    if (b->bytes_consumed > 0)
    {
        //printf("Hej 1!\n");
        if (b->bytes_into_buffer)
        {
            if(foo)
            {
            printf("bytes_consumed = %d, bytes_into_buffer = %d\n",
                b->bytes_consumed, b->bytes_into_buffer);
                foo = 0;
            }
            memmove((void*)b->buffer, (void*)(b->buffer + b->bytes_consumed),
                b->bytes_into_buffer);
            //printf("Hej 3!\n");
        }

        //printf("Hej 4!\n");
        if (!b->at_eof)
        {
            bread = fread((void*)(b->buffer + b->bytes_into_buffer), 1,
                b->bytes_consumed, b->infile);

            if (bread != b->bytes_consumed)
                b->at_eof = 1;

            b->bytes_into_buffer += bread;
        }

        b->bytes_consumed = 0;

        if (b->bytes_into_buffer > 3)
        {
            if (memcmp(b->buffer, "TAG", 3) == 0)
                b->bytes_into_buffer = 0;
        }
        if (b->bytes_into_buffer > 11)
        {
            if (memcmp(b->buffer, "LYRICSBEGIN", 11) == 0)
                b->bytes_into_buffer = 0;
        }
        if (b->bytes_into_buffer > 8)
        {
            if (memcmp(b->buffer, "APETAGEX", 8) == 0)
                b->bytes_into_buffer = 0;
        }
    }

    return 1;
}

static void advance_buffer(aac_buffer *b, int bytes)
{
    b->file_offset += bytes;
    b->bytes_consumed = bytes;
    b->bytes_into_buffer -= bytes;
}

int GetAACTrack(mp4ff_t *infile)
{
    /* find AAC track */
    int i, rc;
    int numTracks = mp4ff_total_tracks(infile);

    for (i = 0; i < numTracks; i++)
    {
        unsigned char *buff = NULL;
        unsigned int buff_size = 0;
        mp4AudioSpecificConfig mp4ASC;

        mp4ff_get_decoder_config(infile, i, &buff, &buff_size);

        if (buff)
        {
            rc = NeAACDecAudioSpecificConfig(buff, buff_size, &mp4ASC);
            free(buff);

            if (rc < 0)
                continue;
            return i;
        }
    }

    /* can't decode this */
    return -1;
}

static int fill_buffer_aac(signed short *dest, int len)
{
    if (playing)
    {
        int writtenbytes = 0;
        for (;;)
        {
            int bytesleft = len - writtenbytes;
            int bufbytes  = writeposbytes - readposbytes;
            //printf("len = %04d - written = %04d = %04d, " \
                    "read = %04d - write = %04d = %04d\n",\
                len,writtenbytes,bytesleft,readposbytes,writeposbytes,bufbytes);
            if (bytesleft <= bufbytes)
            {
                memcpy((char*)dest + writtenbytes, (char*)tmpbuf + readposbytes, bytesleft);
                readposbytes += bytesleft;
                writtenbytes = len;
                break;
            }
            else
            {
                if( tmpbuf ) {
					memcpy((char*)dest + writtenbytes, (char*)tmpbuf + readposbytes, bufbytes);
					writtenbytes += bufbytes;
				}

                readposbytes = writeposbytes = 0;

                while (!writeposbytes)
                {
                    tmpbuf = NeAACDecDecode(hDecoder, &frameInfo, b.buffer,
                        b.bytes_into_buffer);
                    if (frameInfo.error > 0)
                    {
                        fprintf(stderr, "faadplugin:fill_buffer - Error: %s\n",
                                NeAACDecGetErrorMessage(frameInfo.error));
                        playing = 0;
						return 0;
                    }

                    writeposbytes = frameInfo.samples << 1;
                    advance_buffer(&b, frameInfo.bytesconsumed);
                    if (b.at_eof) {
						playing = 0;
                        return 0;
					}

                    fill_b_buffer(&b);
                }

                if (!writeposbytes)
                    break;
            }
        }

        return writtenbytes;
    }

    return 0;
}

static int fill_buffer_mp4(signed short *dest, int len)
{
	// Finished decoding?
	if( sampleId >= numSamples ) {
		playing = 0;
		return 0;
	}

    if (playing)
    {
        int writtenbytes = 0;
        for (;;)
        {
            int bytesleft = len - writtenbytes;
            int bufbytes  = writeposbytes - readposbytes;
            //printf("len = %04d - written = %04d = %04d, " \
                    "read = %04d - write = %04d = %04d\n",\
                len,writtenbytes,bytesleft,readposbytes,writeposbytes,bufbytes);
            if (bytesleft <= bufbytes)
            {
                memcpy((char*)dest + writtenbytes, (char*)tmpbuf + readposbytes, bytesleft);
                readposbytes += bytesleft;
                writtenbytes = len;
                break;
            }
            else
            {
                if( tmpbuf ) {
					memcpy((char*)dest + writtenbytes, (char*)tmpbuf + readposbytes, bufbytes);
					writtenbytes += bufbytes;
				}

                readposbytes = writeposbytes = 0;

                while (!writeposbytes)
                {
					 /* get acces unit from MP4 file */
					if (buffer) {
						//free(buffer);
						buffer = NULL;
					}
					buffer_size = 0;

					int dur = mp4ff_get_sample_duration(infile, track, sampleId);
					int rc = mp4ff_read_sample(infile, track, sampleId, &buffer,  &buffer_size);
					if (rc == 0) {
						fprintf(stderr, "Reading from MP4 file failed.\n");
						playing = 0;
						return 0;
					}

					sampleId ++;
					tmpbuf = NeAACDecDecode(hDecoder, &frameInfo, buffer, buffer_size);

                    if (frameInfo.error > 0)
                    {
                        fprintf(stderr, "faadplugin:fill_buffer - Error: %s\n",
                                NeAACDecGetErrorMessage(frameInfo.error));
                        playing = 0;
                        return 0;
                    }

                    writeposbytes = frameInfo.samples<<1;
                }

                if (!writeposbytes)
                    break;
            }
        }
        //samples_done += writtenbytes/(2*plugin.channels);
        return writtenbytes;
    }
    return 0;
}

// LOADING ROUTINES
int load_mp4(const char *filename)
{
	/* initialise the callback structure */
    mp4cb = (mp4ff_callback_t *)malloc(sizeof(mp4ff_callback_t));
	if( !mp4cb ) return -7;

    mp4File = fopen(filename, "rb");
	if( !mp4File ) return -8;
    mp4cb->read = read_callback;
    mp4cb->seek = seek_callback;
    mp4cb->user_data = mp4File;

	hDecoder = NeAACDecOpen();
	if( hDecoder == NULL ) return -9;

    /* Set configuration */
    config = NeAACDecGetCurrentConfiguration(hDecoder);
    config->outputFormat = FAAD_FMT_16BIT;
    config->downMatrix   = 1; // 5.1 to 2;
    //config->dontUpSampleImplicitSBR = 1;
    NeAACDecSetConfiguration(hDecoder, config);

	infile = mp4ff_open_read(mp4cb);
    if (!infile)
    {
        /* unable to open file */
        return -10;
    }

    if ((track = GetAACTrack(infile)) < 0)
    {
        NeAACDecClose(hDecoder); hDecoder = NULL;
        mp4ff_close(infile);     infile = NULL;
        SAFE_FREE(mp4cb);
        fclose(mp4File);		 mp4File = NULL;
        return -11;
    }

	buffer = NULL;
    buffer_size = 0;
    mp4ff_get_decoder_config(infile, track, &buffer, &buffer_size);

	long unsigned int b_size = buffer_size, freq;
	unsigned char chn;
    if(NeAACDecInit2(hDecoder, buffer, b_size,
                    &freq, &chn) < 0)
    {
        /* If some error initializing occured, skip the file */
        NeAACDecClose(hDecoder); hDecoder = NULL;
        mp4ff_close(infile);     infile = NULL;
        SAFE_FREE(mp4cb);
        fclose(mp4File);		 mp4File = NULL;
        return -12;
    }

    if (buffer)
    {
        if (NeAACDecAudioSpecificConfig(buffer, buffer_size, &mp4ASC) >= 0)
        {
            if (mp4ASC.frameLengthFlag == 1)  ;
            if (mp4ASC.sbr_present_flag == 1) ;
        }

        free(buffer); buffer = NULL;
    }

	/* GET FILE INFO */
    {
        char *tag = NULL, *item = NULL;
        int k, j;
        char *ot[6] = { "NULL", "MAIN AAC", "LC AAC", "SSR AAC", "LTP AAC", "HE AAC" };
        long samples = mp4ff_num_samples(infile, track);
        float f = 1024.0f;
        float seconds;
        if (mp4ASC.sbr_present_flag == 1)
        {
            f = f * 2.0f;
        }
        seconds = (float)samples*(float)(f-1.0f)/(float)mp4ASC.samplingFrequency;

		channels   = mp4ASC.channelsConfiguration;
		samplerate = mp4ASC.samplingFrequency;
        length     = seconds * 1000;
		kbps 	   = mp4ff_get_avg_bitrate(infile, track) / 1000;

		mp4ff_meta_get_title(infile, &tag);
		if( tag ) {
			strncpy(title, tag, 128);
			free(tag); tag = NULL;
		}

		mp4ff_meta_get_artist(infile, &tag);
		if( tag ) {
			strncpy(artist, tag, 128);
			free(tag); tag = NULL;
		}

		mp4ff_meta_get_album(infile, &tag);
		if( tag ) {
			strncpy(album, tag, 128);
			free(tag); tag = NULL;
		}

		mp4ff_meta_get_title(infile, &tag);
		if( tag ) {
			strncpy(title, tag, 128);
			free(tag); tag = NULL;
		}

		mp4ff_meta_get_date(infile, &tag);
		if( tag ) {
			strncpy(copyright, tag, 128);
			free(tag); tag = NULL;
		}
    }

	sampleId = 0;
	numSamples = mp4ff_num_samples(infile, track);

	writeposbytes = 0;
    readposbytes  = 0;
	buffer = NULL;

	playing = 1;
	return 1;
}

int load_aac(const char *filename)
{
	// Search ID3 Tag
	MPEG::File *tagfile = NULL;
	tagfile = new MPEG::File(filename, true, MPEG::Properties::Accurate);

	if( tagfile ) {
		if( tagfile->tag() && !tagfile->tag()->isEmpty() )
		{
			TagLib::Tag *tag = tagfile->tag();

			// Get tag size and picture
			if( tagfile->ID3v2Tag() ) {

				// Get the list of frames for a specific frame type
				ID3v2::FrameList l = tagfile->ID3v2Tag()->frameListMap()["APIC"];

				if( !l.isEmpty() ) {
					pspDebugScreenPrintf("[Picture Found]\n");
					ID3v2::AttachedPictureFrame *pic =
					static_cast<ID3v2::AttachedPictureFrame *> (l.front());
					pic_size = pic->picture().size();

					if( pic->picture().size() > 0 ) pic_data = new char [pic->picture().size()];

					if( pic_data != NULL )
						memcpy(pic_data, pic->picture().data(), pic->picture().size());
				}
			}

			if( !tag->title().isEmpty()  ) strncpy(title, tag->title().toCString(), 128);
			if( !tag->artist().isEmpty() ) strncpy(artist, tag->artist().toCString(), 128);
			if( !tag->album().isEmpty()  ) strncpy(album, tag->album().toCString(), 128);
			if( tag->year() > 0 		 ) sprintf(copyright, "%d", tag->year());
		}

		delete tagfile;
	}

	// Open file
	memset(&b, 0, sizeof(aac_buffer));
    b.infile = fopen(filename, "rb");
    if (!b.infile)
    {
        return -1;
    }

    fseek(b.infile, 0, SEEK_END);
    int fileread = ftell(b.infile);
    fseek(b.infile, 0, SEEK_SET);

    if (!(b.buffer = (unsigned char*)malloc(FAAD_MIN_STREAMSIZE*MAX_CHANNELS)))
    {
        return -2;
    }

	// clean buffer
	memset(b.buffer, 0, FAAD_MIN_STREAMSIZE*MAX_CHANNELS);

    int bread = fread(b.buffer, 1, FAAD_MIN_STREAMSIZE*MAX_CHANNELS, b.infile);
    b.bytes_into_buffer = bread;
    b.bytes_consumed = 0;
    b.file_offset = 0;


    if (bread != FAAD_MIN_STREAMSIZE*MAX_CHANNELS)
        b.at_eof = 1;

    tagsize = 0;
    if (!memcmp(b.buffer, "ID3", 3))
    {
        pspDebugScreenPrintf( "foo 4.3.1\n");
        tagsize = (b.buffer[6] << 21) | (b.buffer[7] << 14) |
            (b.buffer[8] <<  7) | (b.buffer[9] <<  0);
        tagsize += 10;
		int tagsizetmp;
		pspDebugScreenPrintf( "foo 4.3.2\n");
        for (tagsizetmp = tagsize; tagsizetmp > 0; )
        {
            int tagsizetmp2 = tagsizetmp <= b.bytes_into_buffer ? tagsizetmp
                : b.bytes_into_buffer;
            tagsizetmp -= tagsizetmp2;
            advance_buffer(&b, tagsizetmp2);
            fill_b_buffer(&b);
        }
    }

    hDecoder = NeAACDecOpen();
	if( hDecoder == NULL ) return -3;

    config = NeAACDecGetCurrentConfiguration(hDecoder);
    //config->defSampleRate = 44100; // For RAW files. Which is.. ?
    config->defObjectType = LC;
    config->outputFormat = FAAD_FMT_16BIT; // double check this
    config->downMatrix = 1; // 5.1 to 2
    config->useOldADTSFormat = 0;
    //config->dontUpSampleImplicitSBR = 1;
    NeAACDecSetConfiguration(hDecoder, config);

    /* get AAC infos for printing */
    int header_type = 0;
    if ((b.buffer[0] == 0xFF) && ((b.buffer[1] & 0xF6) == 0xF0))
    {
        float length_f;
        adts_parse(&b, &kbps, &length_f);
        fseek(b.infile, tagsize, SEEK_SET);
        length = (int)(length_f*1000.0f);

        bread = fread(b.buffer, 1, FAAD_MIN_STREAMSIZE*MAX_CHANNELS, b.infile);
        b.at_eof = bread != FAAD_MIN_STREAMSIZE*MAX_CHANNELS ? 1 : 0;
        b.bytes_into_buffer = bread;
        b.bytes_consumed    = 0;
        b.file_offset       = tagsize;

        header_type = 1;
    } else if (memcmp(b.buffer, "ADIF", 4) == 0) {
        int skip_size = (b.buffer[4] & 0x80) ? 9 : 0;
        kbps = ((unsigned int)(b.buffer[4 + skip_size] & 0x0F)<<19) |
            ((unsigned int)b.buffer[5 + skip_size]<<11) |
            ((unsigned int)b.buffer[6 + skip_size]<<3) |
            ((unsigned int)b.buffer[7 + skip_size] & 0xE0);

        float length_f = (float)fileread;
        if (fileread != 0)
        {
            length_f = length_f*8.f/kbps + 0.5f;
        }

        length = (int)length_f;
        kbps = (int)((float)kbps/1000.0f + 0.5f);

        header_type = 2;
    }

    fill_b_buffer(&b);

	unsigned long int freq;
	unsigned char chn;
    if ((bread = NeAACDecInit(hDecoder, b.buffer,
        b.bytes_into_buffer, &freq, &chn)) < 0)
    {
        fprintf(stderr, "faadplugin:init_file - Error initializing decoder library.\n");
        return -1;
    }

	channels   = chn;
    samplerate = freq;

    advance_buffer(&b, bread);
    fill_b_buffer(&b);

    // REMOVE - samples_done  = 0;
    writeposbytes = 0;
    readposbytes  = 0;

	playing = 1;

	return 1;
}
/* END EXTRA FUNCTIONS */

int load( char *file )
{
	// Identify file
	mp4file = 0;
	unsigned char header[8];
    FILE *in = fopen(file, "rb");
    if (!in) {
        return 0;
    }
    fread(header, 1, 8, in);
    fclose(in);

    if (header[4] == 'f' && header[5] == 't' && header[6] == 'y' && header[7] == 'p')
        mp4file = 1;

	if( mp4file ) {
		strncpy(format, "MP4", 128);
		return load_mp4(file);
	}
	else {
		strncpy(format, "AAC", 128);
		return load_aac(file);
	}

	// Never reached
	return 0;
}

int seek( int time )
{
	if( playing ) {
		// Seek function (time in msec)
	}

    return 0;
}

void cleanup( void )
{
	// Cleanup
	reset();
	__psp_free_heap();
}

int write_buffer( signed short *dest, int len )
{
	if(playing)
    {
		// Audio write func
		if( mp4file ) return fill_buffer_mp4(dest, len);
		else return fill_buffer_aac(dest, len);
    }

    return 0;
}

void get_song_title( char *filename, char **title, int *length )
{
}

void get_packet(void)
{
	switch( itable.in_packet->ID ) {
		case NO_PACKET:
		default:
			break;
	};
}

void send_packet(void)
{
	switch( itable.out_packet->ID ) {
		/* song data */
		case SONG_TITLE:
			itable.out_packet->data = title;
			break;
		case SONG_ARTIST:
			itable.out_packet->data = artist;
			break;
		case SONG_ALBUM_GAME:
			itable.out_packet->data = album;
			break;
		case SONG_COPYRIGHT:
			itable.out_packet->data = copyright;
			break;
		case SONG_FORMAT:
			itable.out_packet->data = format;
			break;
		case SONG_LENGTH:
			itable.out_packet->data = &length;
			break;
		case SONG_SUBTUNES:
			itable.out_packet->data = &subtunes;
			break;

		/* audio info */
		case AUDIO_CHANNELS:
			itable.out_packet->data = &channels;
			break;
		case AUDIO_SAMPLERATE:
			itable.out_packet->data = &samplerate;
			break;
		case AUDIO_FORMAT:
			itable.out_packet->data = &aformat;
			break;
		case AUDIO_KBPS:
			itable.out_packet->data = &kbps;
			break;

		/* extra data */
		case ATTACHED_PICTURE:
			itable.out_packet->size = pic_size;
			itable.out_packet->data = pic_data;
			break;
		case NO_PACKET:
		default:
			itable.out_packet->size = 0;
			itable.out_packet->data = NULL;
			break;
	};
}
