
/*  Input Plugin for GameMusicGear ~
 *  Copyright (C) 2005 - 2010 gama, rockMax
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/* Default headers */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/stat.h>

#include <pspkernel.h>
#include "../plugin.h"

#include "ac3.h"

PSP_MODULE_INFO("IN_AC3", 0, 1, 1);
PSP_NO_CREATE_MAIN_THREAD();
PSP_HEAP_SIZE_KB( 6 * 1024 );

#ifdef __cplusplus
extern "C" {
#endif
int module_start(SceSize argc, void* argp)
{
 	return 0;
}

int module_stop(int args, void *argp)
{
	return 0;
}
#ifdef __cplusplus
}
#endif

/* Prototypes */
void init(void);
void reset(void);
int  load( char *filename );
int  load_rw( char *filename, char *addr, int size );
long  seek( long time );
void cleanup( void );
int  write_buffer( signed short *dest, int len );
void get_song_title( char *filename, char **title, int *length );

void get_packet(void);
void send_packet(void);
/* End Prototypes */

/* Variables */
ac3_frame_t *ac3_frame = NULL;
ac3_config_t ac3_config;

#define CHUNK_SIZE 2048
uint_8 buf[CHUNK_SIZE];
FILE *in_file = NULL;

char tmpbuf[256 * 6 * 2 * 2];
int writeposbytes = 0;
int readposbytes  = 0;
int	fsize         = 0;

int playing = 0;
int channels = 1;
int samplerate = 44100;
AFormat aformat = FMT_S16_LE;
int kbps = 128;

char title[128];
char artist[128];
char album[128];
char copyright[128];
char format[128];

long length;
int subtunes;
/* End Variables */

/** START of Plugin definitions setup */
InputPlugin itable =
{
	/* Populated by Plugin */
	"AC3 Decoder",				/* The description of the input plugin */

	init,							/* Called when the plugin is loaded */
	reset, 							/* Called when need to update config */
	load,							/* Load the file */
	NULL,							/* Load the file from data */
	NULL,							/* Set song subtune */
	NULL,							/* Seek to the specified time */
	NULL,							/* Set the equalizer, most plugins won't be able to do this */
	cleanup,						/* Called when gmgear exit */
	write_buffer, 					/* Fill audio buffer */
	get_song_title,					/* Function to grab the title string */

	get_packet,					/* Called when need main app data packet*/
	send_packet,				/* Send data packet to main app */

	NULL,						/* Filled by Main App */
	NULL,						/* Filled by Main App */
	NULL,						/* Filled by Main App */
	0,								/* Filled by Main App */
	LATIN,							/* Filled by Main App */
	CLIENT_PLAYER					/* Player type */
};
/** END of Plugin definitions setup */

/** Single Plugin Export */
#ifdef PSP
InputPlugin *get_iplugin_info()
#else
InputPlugin *dll_get_iplugin_info()
#endif
{
	return &itable;
}

void fill_buffer_b(uint_8 **start, uint_8 **end)
{
	uint_32 bytes_read;

	*start = buf;

	bytes_read = fread(*start, 1, CHUNK_SIZE, in_file);

	//FIXME hack...
	if(bytes_read < CHUNK_SIZE)
		playing = 0;

	*end = *start + bytes_read;
}

void init(void)
{
	strcpy(format, "AC3");
	subtunes = 1;

	kbps = 0x6C;
	aformat = FMT_S16_LE;

	itable.in_packet->ID = NO_PACKET;
}

void reset(void)
{
	strcpy(title, "");
	strcpy(album, "");
	strcpy(copyright, "");
	strcpy(artist, "");

	length = 0;
	playing = 0;
	fsize = 0;

	// Cleanup
	if( in_file ) {
		fclose(in_file);
		in_file = NULL;
	}

	itable.in_packet->ID = NO_PACKET;
}

int load( char *file )
{
	// Get file size
	struct stat st;
	memset(&st, 0, sizeof(struct stat));
	if( stat(file, &st) == 0 ) fsize = st.st_size;

	in_file = fopen(file, "r");
	if( !in_file ) return -1;

	ac3_config.fill_buffer_callback = fill_buffer_b;
	ac3_config.num_output_ch = 2;
	ac3_config.flags = 0;

	ac3_init(&ac3_config);

	memset(tmpbuf, 0, sizeof(tmpbuf));

	ac3_frame = ac3_decode_frame();
	if( !ac3_frame)	return -1;

	readposbytes = 0;
    writeposbytes = 256 * 6 * 2 * 2;
	memcpy(tmpbuf, ac3_frame->audio_data, 256 * 6 * 2 * 2);

	// Big enough to play all song
	channels   = ac3_config.num_output_ch;
	samplerate = ac3_frame->sampling_rate;
	kbps	   = ac3_frame->bit_rate;
    length     = fsize*8/kbps;
	if( length <= 0 ) length = 1800 * 1000;  // Just big enough for now (30 min)

	playing = 1;
	return 1;
}

long seek( long time )
{
	if( playing ) {
		// Seek function (time in msec)
	}

    return 0;
}

void cleanup( void )
{
	// Cleanup
	reset();
	__psp_free_heap();
}

int write_buffer( signed short *dest, int len )
{
	if (playing)
    {
        int writtenbytes = 0;
        for (;;)
        {
            int bytesleft = len - writtenbytes;
            int bufbytes  = writeposbytes - readposbytes;

            if (bytesleft <= bufbytes)
            {
                memcpy((char*)dest + writtenbytes, (char*)tmpbuf + readposbytes, bytesleft);
                readposbytes += bytesleft;
                writtenbytes = len;
                break;
            }
            else
            {
                memcpy((char*)dest + writtenbytes, (char*)tmpbuf + readposbytes, bufbytes);
                writtenbytes += bufbytes;
                readposbytes = writeposbytes = 0;

                while (!writeposbytes)
                {
                    ac3_frame = ac3_decode_frame();
                    if( !ac3_frame)
                    {
						// End of song?
                        playing = 0;
                        break;
                    }

                    writeposbytes = 256 * 6 * 2 * 2;
					memcpy(tmpbuf, ac3_frame->audio_data, 256 * 6 * 2 * 2);
                }
            }
        }

        return writtenbytes;
    }

    return 0;
}

void get_song_title( char *filename, char **title, int *length )
{
}

void get_packet(void)
{
	switch( itable.in_packet->ID ) {
		case NO_PACKET:
		default:
			break;
	};
}

void send_packet(void)
{
	switch( itable.out_packet->ID ) {
		/* song data */
		case SONG_TITLE:
			itable.out_packet->data = title;
			break;
		case SONG_ARTIST:
			itable.out_packet->data = artist;
			break;
		case SONG_ALBUM_GAME:
			itable.out_packet->data = album;
			break;
		case SONG_COPYRIGHT:
			itable.out_packet->data = copyright;
			break;
		case SONG_FORMAT:
			itable.out_packet->data = format;
			break;
		case SONG_LENGTH:
			itable.out_packet->data = &length;
			break;
		case SONG_SUBTUNES:
			itable.out_packet->data = &subtunes;
			break;

		/* audio info */
		case AUDIO_CHANNELS:
			itable.out_packet->data = &channels;
			break;
		case AUDIO_SAMPLERATE:
			itable.out_packet->data = &samplerate;
			break;
		case AUDIO_FORMAT:
			itable.out_packet->data = &aformat;
			break;
		case AUDIO_KBPS:
			itable.out_packet->data = &kbps;
			break;

		/* extra data */
		case ATTACHED_PICTURE:
			itable.out_packet->size = 0;
			itable.out_packet->data = NULL;
			break;
		case NO_PACKET:
		default:
			itable.out_packet->size = 0;
			itable.out_packet->data = NULL;
			break;
	};
}
