
/*  Input Plugin for GameMusicGear ~
 *  Copyright (C) 2005 - 2010 gama, rockMax
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/* Default headers */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/stat.h>

#include "dumb.h"

#include <pspkernel.h>
#include "../plugin.h"

PSP_MODULE_INFO("IN_DUMB", 0, 1, 1);
PSP_NO_CREATE_MAIN_THREAD();
PSP_HEAP_SIZE_KB( 6 * 1024 );

#ifdef __cplusplus
extern "C" {
#endif
int module_start(SceSize argc, void* argp)
{
 	return 0;
}

int module_stop(int args, void *argp)
{
	return 0;
}
#ifdef __cplusplus
}
#endif

/* Prototypes */
void init(void);
void reset(void);
int  load( char *filename );
int  load_rw( char *filename, char *addr, int size );
long  seek( long time );
void cleanup( void );
int  write_buffer( signed short *dest, int len );
void get_song_title( char *filename, char **title, int *length );

void get_packet(void);
void send_packet(void);
/* End Prototypes */

/* Definitions */

typedef struct
{
	DUH *duh;
	DUH_SIGRENDERER *file;
	short channels, bits_per_sample, eof, going;
	long samples_per_sec;
	unsigned long position, length;
	int seek_to;
	pid_t pid;
	int in_mem;
}
DuhFile;

/* Variables */
static DuhFile *duh_file = NULL;
int quick_load = 0;
int song_length = 200 * 1000;

int playing = 0;
int channels = 2;
int samplerate = 44100;
AFormat aformat = FMT_S16_LE;
int kbps = 128;

char title[128];
char artist[128];
char album[128];
char copyright[128];
char format[128];

long length;
int subtunes;
/* End Variables */

/** START of Plugin definitions setup */
InputPlugin itable =
{
	/* Populated by Plugin */
	"DUMB 0.9.3",				/* The description of the input plugin */

	init,							/* Called when the plugin is loaded */
	reset, 							/* Called when need to update config */
	load,							/* Load the file */
	load_rw,						/* Load the file from data */
	NULL,							/* Set song subtune */
	seek,							/* Seek to the specified time */
	NULL,							/* Set the equalizer, most plugins won't be able to do this */
	cleanup,						/* Called when gmgear exit */
	write_buffer, 					/* Fill audio buffer */
	get_song_title,					/* Function to grab the title string */

	get_packet,					/* Called when need main app data packet*/
	send_packet,				/* Send data packet to main app */

	NULL,						/* Filled by Main App */
	NULL,						/* Filled by Main App */
	NULL,						/* Filled by Main App */
	0,								/* Filled by Main App */
	LATIN,							/* Filled by Main App */
	CLIENT_PLAYER					/* Player type */
};
/** END of Plugin definitions setup */

/** Single Plugin Export */
InputPlugin *get_iplugin_info()
{
	return &itable;
}

static void duh_init(void)
{
	dumb_register_stdfiles();
}

static void install_callbacks(DUH_SIGRENDERER *sr)
{
	DUMB_IT_SIGRENDERER *itsr = duh_get_it_sigrenderer(sr);
	dumb_it_set_loop_callback(itsr, &dumb_it_callback_terminate, NULL);
	dumb_it_set_xm_speed_zero_callback(itsr, &dumb_it_callback_terminate, NULL);
}

void init(void)
{
	strcpy(format, "DUMB");
	subtunes = 1;

#ifdef DEBUG
	pspDebugScreenInit();
#endif

	/* reset variables */
	duh_file = NULL;
	quick_load = 0;
	song_length = 200 * 1000;

	duh_init();

	kbps = 0x6C;
	aformat = FMT_S16_LE;

	itable.in_packet->ID = NO_PACKET;
}

void reset(void)
{
	channels = ReadIntValue("channels", itable.config);
	samplerate = ReadFreqValue("samplerate", itable.config);
	dumb_it_max_to_mix = ReadIntValue("dumb_it_max_to_mix", itable.config);
	dumb_resampling_quality = ReadIntValue("dumb_resampling_quality", itable.config);
	quick_load = ReadBoolValue("quick_load", itable.config);
	if( quick_load ) song_length = ReadIntValue("song_length", itable.config) * 1000;

	strcpy(title, "");
	strcpy(album, "");
	strcpy(copyright, "");
	strcpy(artist, "");

	length = 0;
	playing = 0;

	// Cleanup
	if( duh_file ) {
		duh_end_sigrenderer(duh_file->file);
		unload_duh(duh_file->duh);
		free(duh_file);
		duh_file = NULL;
	}

	itable.in_packet->ID = NO_PACKET;
}

char *getFileData(char *filename, long *size)
{
	SceUID mFile;
    char *data;
	size_t readbytes;

	mFile = sceIoOpen(filename, PSP_O_RDONLY, 0777);
	if( mFile < 0 ) return NULL;
    *size = sceIoLseek(mFile, 0, PSP_SEEK_END);
	sceIoLseek(mFile, 0, PSP_SEEK_SET);
    data = malloc(sizeof(char) * (*size));
	readbytes = sceIoRead(mFile, data, *size);
	*size = readbytes;

	sceIoClose(mFile);

    return(data);
}

/* LOAD HELPER */
char *fix_tag(char *tag)
{
	/* Remove remaining spaces */
	char *tmp, *ptr;
	tmp = ptr = tag;
	while( *ptr++ ) {
		char c = *ptr;
		if( c != ' ' && c != '\0' ) tmp = ptr + 1;
	}

	*tmp = '\0';

	return tag;
}

int reset_dumb_file(DUMBFILE *file, char *addr, int size)
{
	if( file ) {
		dumbfile_close(file);
		if( (file = dumbfile_open_memory(addr, size)) != NULL ) return 1;
	}

	return 0;
}

int _load(char *addr, long size)
{
	// Sanity check
	if( addr == NULL ) return -10;

	duh_file = malloc(sizeof (DuhFile));
	memset(duh_file, 0, sizeof (DuhFile));

	duh_file->in_mem = 1;
	DUMBFILE *file = dumbfile_open_memory(addr, size);
	if( file == NULL) return -1;

	duh_file->duh = read_duh(file);
	if (!duh_file->duh) {
		if( !reset_dumb_file(file, addr, size) ) return 0;
		duh_file->duh = dumb_read_it(file);
		if (!duh_file->duh) {
			if( !reset_dumb_file(file, addr, size) ) return 0;
			duh_file->duh = dumb_read_xm(file);
			if (!duh_file->duh) {
				if( !reset_dumb_file(file, addr, size) ) return 0;
				duh_file->duh = dumb_read_s3m(file);
				if (!duh_file->duh) {
					if( !reset_dumb_file(file, addr, size) ) return 0;
					duh_file->duh = dumb_read_mod(file);
					if (!duh_file->duh) {
						dumbfile_close(file);
						free(duh_file);
						duh_file = NULL;
						return -2;
					}
				}
			}
		}
	}

	if ((duh_file->file = duh_start_sigrenderer(duh_file->duh, 0, channels, 0)))
	{
		install_callbacks(duh_file->file);
		duh_file->channels = channels;
		duh_file->samples_per_sec = samplerate;
		duh_file->bits_per_sample = 16;
		length = duh_file->length = ((long long)duh_get_length(duh_file->duh) * (1000LL)) >> 16;
		duh_file->position = 0;
		duh_file->going = 1;

		char *tag = duh_get_tag(duh_file->duh, "TITLE");
		fix_tag(tag);
		if( tag && tag[0] ) strncpy(title, fix_tag(tag), 128);
		duh_file->seek_to = -3;
	}
	else {
		dumbfile_close(file);
		free(duh_file);
		duh_file = NULL;
		return -4;
	}

	// Close dumbfile
	dumbfile_close(file);
	return 1;
}

int _load_quick(char *addr, long size)
{
	// Sanity check
	if( addr == NULL ) return -10;

	duh_file = malloc(sizeof (DuhFile));
	memset(duh_file, 0, sizeof (DuhFile));

	duh_file->in_mem = 1;
	DUMBFILE *file = dumbfile_open_memory(addr, size);
	if( file == NULL) return -1;

	duh_file->duh = read_duh(file);
	if (!duh_file->duh) {
		if( !reset_dumb_file(file, addr, size) ) return 0;
		duh_file->duh = dumb_read_it_quick(file);
		if (!duh_file->duh) {
			if( !reset_dumb_file(file, addr, size) ) return 0;
			duh_file->duh = dumb_read_xm_quick(file);
			if (!duh_file->duh) {
				if( !reset_dumb_file(file, addr, size) ) return 0;
				duh_file->duh = dumb_read_s3m_quick(file);
				if (!duh_file->duh) {
					if( !reset_dumb_file(file, addr, size) ) return 0;
					duh_file->duh = dumb_read_mod_quick(file);
					if (!duh_file->duh) {
						dumbfile_close(file);
						free(duh_file);
						duh_file = NULL;
						return -2;
					}
				}
			}
		}
	}

	if ((duh_file->file = duh_start_sigrenderer(duh_file->duh, 0, channels, 0)))
	{
		install_callbacks(duh_file->file);
		duh_file->channels = channels;
		duh_file->samples_per_sec = samplerate;
		duh_file->bits_per_sample = 16;
		length = song_length;
		duh_file->position = 0;
		duh_file->going = 1;

		char *tag = duh_get_tag(duh_file->duh, "TITLE");
		if( tag && tag[0] ) strncpy(title, fix_tag(tag), 128);
		duh_file->seek_to = -3;
	}
	else {
		dumbfile_close(file);
		free(duh_file);
		duh_file = NULL;
		return -4;
	}

	// Close dumbfile
	dumbfile_close(file);
	return 1;
}

int load( char *file )
{
	long size;
	char *data = getFileData(file, &size);
	int ret = quick_load ? _load_quick(data, size) : _load(data, size);

	if( ret > 0 ) {
		char *tmp = strrchr(file, '.');
		if( tmp ) {
			char ext[32];
			strncpy(ext, tmp + 1, 32);
			int len = strlen(ext), i;
			for(i = 0; i < len; i++) ext[i] = toupper(ext[i]);
			strncpy(format, ext, 32);
		}

		playing = 1;
	}

	return ret;
}

int load_rw( char *filename, char *addr, int size )
{
	int ret = quick_load ? _load_quick(addr, size) : _load(addr, size);

	if( ret > 0 ) {
		char *tmp = strrchr(filename, '.');
		if( tmp ) {
			char ext[32];
			strncpy(ext, tmp + 1, 32);
			int len = strlen(ext), i;
			for(i = 0; i < len; i++) ext[i] = toupper(i);
			strncpy(format, ext, 32);
		}

		playing = 1;
	}

	return ret;
}

long seek( long time )
{
	if( playing ) {
		long pos = (long)(((double)time / 1000.0) * 65536.0);
		duh_end_sigrenderer(duh_file->file);
		duh_file->file = duh_start_sigrenderer(duh_file->duh, 0, channels, pos);
		install_callbacks(duh_file->file);
		return time;
	}

    return 0;
}

void cleanup( void )
{
	// Cleanup
	reset();
	__psp_free_heap();
}

int write_buffer( signed short *dest, int len )
{
	if (playing)
    {
        int actual_read = duh_render(duh_file->file, 16, 0, 1.0f, 65536.0f / (float)duh_file->samples_per_sec,
				len / ((duh_file->bits_per_sample / 8) * duh_file->channels), dest);
		actual_read *= (duh_file->bits_per_sample / 8) * duh_file->channels;
		return actual_read;
    }

    return 0;
}

void get_song_title( char *filename, char **title, int *length )
{
}

void get_packet(void)
{
	switch( itable.in_packet->ID ) {
		case NO_PACKET:
		default:
			break;
	};
}

void send_packet(void)
{
	switch( itable.out_packet->ID ) {
		/* song data */
		case SONG_TITLE:
			itable.out_packet->data = title;
			break;
		case SONG_ARTIST:
			itable.out_packet->data = artist;
			break;
		case SONG_ALBUM_GAME:
			itable.out_packet->data = album;
			break;
		case SONG_COPYRIGHT:
			itable.out_packet->data = copyright;
			break;
		case SONG_FORMAT:
			itable.out_packet->data = format;
			break;
		case SONG_LENGTH:
			itable.out_packet->data = &length;
			break;
		case SONG_SUBTUNES:
			itable.out_packet->data = &subtunes;
			break;

		/* audio info */
		case AUDIO_CHANNELS:
			itable.out_packet->data = &channels;
			break;
		case AUDIO_SAMPLERATE:
			itable.out_packet->data = &samplerate;
			break;
		case AUDIO_FORMAT:
			itable.out_packet->data = &aformat;
			break;
		case AUDIO_KBPS:
			itable.out_packet->data = &kbps;
			break;

		/* extra data */
		case ATTACHED_PICTURE:
			itable.out_packet->size = 0;
			itable.out_packet->data = NULL;
			break;
		case NO_PACKET:
		default:
			itable.out_packet->size = 0;
			itable.out_packet->data = NULL;
			break;
	};
}
