/*  Input Plugin for GameMusicGear ~
 *  Copyright (C) 2005 - 2010 gama, rockMax
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/* Default headers */

#include <pspkernel.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <malloc.h>

#include <ctype.h>

#include "../plugin.h"

#include "FLAC/stream_decoder.h"
#include "FLAC/metadata.h"
#include "FLAC/callback.h"
#include "FLAC/format.h"

#define TAGLIB
#ifdef TAGLIB
#include "fileref.h"
#include "tag.h"
#include "tbytevector.h"
#include "id3v2tag.h"
#include "id3v2frame.h"
#include "id3v2header.h"
#include "attachedpictureframe.h"
#include "flacfile.h"
using namespace TagLib;
#endif

// LINKER PATCH
extern "C" {
void *__dso_handle = NULL;
}

PSP_MODULE_INFO("IN_FLAC", 0, 1, 1);
PSP_NO_CREATE_MAIN_THREAD();
PSP_HEAP_SIZE_KB( 1024 * 4 );

#ifdef __cplusplus
extern "C" {
#endif
int module_start(SceSize argc, void* argp)
{
 	return 0;
}

int module_stop(int args, void *argp)
{
	return 0;
}
#ifdef __cplusplus
}
#endif

/* Defines */
#define 	BUFFSIZE 		(32768*2*2)
/* End Defines */

/* Prototypes */
void init(void);
void reset(void);
int load( char *filename );
long seek( long time );
void cleanup( void );
int write_buffer( signed short *dest, int len );
void get_song_title( char *filename, char **title, int *length );

void get_packet(void);
void send_packet(void);

char *get_picture( int *size );
/* End Prototypes */

/* Variables */
FLAC__StreamDecoder *mydecoder = NULL;
signed short *tmpbuf = NULL;
int playing = 0;
long samples_done; // Samples done for each channel
unsigned int buffsize;
unsigned int writeposbytes;
unsigned int readposbytes;

/* Required */
int channels = 2;
int samplerate = 44100;
AFormat aformat = FMT_S16_LE;
int kbps = 128;

char title[128];
char artist[128];
char album[128];
char copyright[128];
char format[128];

long length;
int subtunes;

bool string_utf8;

// picture stuff
char *pic_data;
int   pic_size;

/* End Variables */

/** START of Plugin definitions setup */
InputPlugin itable =
{
	/* Populated by Plugin */
	"Free Lossless Audio Codec",	/* The description of the input plugin */

	init,							/* Called when the plugin is loaded */
	reset, 						/* Called when need to update config */
	load,							/* Load the file */
	NULL,							/* Load the file from data */
	NULL,							/* Set song subtune */
	seek,							/* Seek to the specified time */
	NULL,							/* Set the equalizer, most plugins won't be able to do this */
	cleanup,						/* Called when gmgear exit */
	write_buffer, 					/* Fill audio buffer */
	get_song_title,					/* Function to grab the title string */

	get_packet,					/* Called when need main app data packet*/
	send_packet,				/* Send data packet to main app */

	NULL,						/* Filled by Main App */
	NULL,						/* Filled by Main App */
	NULL,						/* Filled by Main App */
	0,								/* Filled by Main App */
	LATIN,							/* Filled by Main App */
	CLIENT_PLAYER					/* Player type */
};
/** END of Plugin definitions setup */

/** Single Plugin Export */
InputPlugin *get_iplugin_info()
{
	return &itable;
}

/* Decoder Functions */
FLAC__StreamDecoderWriteStatus write_callback(
    const FLAC__StreamDecoder *decoder, const FLAC__Frame *frame,
    const FLAC__int32 *const buffer[], void *client_Data)
{
	if( playing ) {
		int samples = frame->header.blocksize * channels;
		if (samples > buffsize)
		{
			//_log("flacplugin:write_callback - Too big block!\n");
			writeposbytes = 0;
			return FLAC__STREAM_DECODER_WRITE_STATUS_ABORT;
		}
		if (channels == 2)
		{
			for(int i = 0, j = 0; j < samples; i++)
			{
				tmpbuf[j++] = buffer[0][i];
				tmpbuf[j++] = buffer[1][i];
        }
    }
    else // Assumes mono
        for (int i = 0; i < samples; i++)
            tmpbuf[i] = buffer[0][i];

		writeposbytes = samples<<1;
	}

    return FLAC__STREAM_DECODER_WRITE_STATUS_CONTINUE;
}

void metadata_callback(const FLAC__StreamDecoder *decoder,
                       const FLAC__StreamMetadata *metadata, void *client_data)
{
	//sceKernelDelayThread( 10000 );
}

void error_callback(const FLAC__StreamDecoder *decoder,
                    FLAC__StreamDecoderErrorStatus status, void *client_data)
{
   printf("flacplugin:error_callback - libFLAC error\n");
   sceKernelExitGame();
}
/* End Decoder Functions */

void init(void)
{
	strcpy(format, "FLAC");

	length = 0;
	subtunes = 1;

	pic_data = NULL;
	pic_size = 0;

	kbps = 0;
	aformat = FMT_S16_LE;

	if( itable.charset == UTF8 ) string_utf8 = true;
	else string_utf8 = false;

	itable.in_packet->ID = NO_PACKET;
}

void reset(void)
{
	strcpy(title, "");
	strcpy(album, "");
	strcpy(copyright, "");
	strcpy(artist, "");

	length = 200000;
	kbps = 128;

	if( pic_data ) {
		delete [] pic_data;
		pic_data = NULL;
		pic_size = 0;
	}

	if (mydecoder) {
        FLAC__stream_decoder_finish(mydecoder);
        FLAC__stream_decoder_delete(mydecoder);
        mydecoder = NULL;
    }

    if (tmpbuf) {
        free(tmpbuf);
        tmpbuf = NULL;
    }

    playing = 0;
	buffsize = 0;
}

void read_tag(char *file)
{
#ifdef TAGLIB
	FLAC::File *tagfile = NULL;
	tagfile = new FLAC::File(file, true, FLAC::Properties::Accurate);

	if( !tagfile ) {
		pspDebugScreenPrintf("Warning: Cannot create Flac tag\n");
		return;
	}

	if( tagfile->tag() && !tagfile->tag()->isEmpty() )
	{
		TagLib::Tag *tag = tagfile->tag();

			// Get picture
			if( tagfile->ID3v2Tag() ) {
				// Get the list of frames for a specific frame type
				ID3v2::FrameList l = tagfile->ID3v2Tag()->frameListMap()["APIC"];

				if( !l.isEmpty() ) {
					pspDebugScreenPrintf("[Picture Found]\n");
					ID3v2::AttachedPictureFrame *pic =
					static_cast<ID3v2::AttachedPictureFrame *> (l.front());
					pic_size = pic->picture().size();

					if( pic->picture().size() > 0 ) pic_data = new char [pic->picture().size()];

					if( pic_data != NULL )
						memcpy(pic_data, pic->picture().data(), pic->picture().size());
				}
			}

		if( !tag->title().isEmpty() ) strncpy(title, tag->title().toCString(), 128);
		if( !tag->album().isEmpty() ) strncpy(album, tag->album().toCString(), 128);
		if( tag->year() > 0 ) snprintf(copyright, 128, "%d", tag->year());
		if( !tag->artist().isEmpty() ) strncpy(artist, tag->artist().toCString(), 128);
	}

	if( tagfile->audioProperties() )
	{
		AudioProperties *properties = tagfile->audioProperties();
		length = properties->length() * 1000;

		channels = properties->channels();
		samplerate = properties->sampleRate();
		kbps = properties->bitrate();
	}

	delete tagfile;
#endif
}

int load( char *file )
{
    mydecoder = FLAC__stream_decoder_new();
	if( !mydecoder ) return 0;

	read_tag(file);

    if( FLAC__stream_decoder_init_file(
        mydecoder,
        file,
        write_callback,
        metadata_callback,
        error_callback,
        NULL) != FLAC__STREAM_DECODER_INIT_STATUS_OK ) return 0;

    //FLAC__stream_decoder_set_metadata_respond_all(mydecoder); // No good
    FLAC__stream_decoder_set_metadata_ignore_all(mydecoder); // Changes nothing?
    FLAC__stream_decoder_process_until_end_of_metadata(mydecoder);

    if ( channels <= 0 ) channels = FLAC__stream_decoder_get_channels(mydecoder);
    if ( channels > 2 ) channels = 2;
    if ( samplerate <= 0 ) samplerate = FLAC__stream_decoder_get_sample_rate(mydecoder);

    if( length <= 0 ) length = (int)(1000.0f*FLAC__stream_decoder_get_total_samples(mydecoder)/samplerate);

    buffsize *= channels;
    if (!buffsize)
        buffsize = BUFFSIZE;
    if(!(tmpbuf = (short int*)malloc(buffsize*2)))
    {
        reset();
        return -1;
    }

    playing = 1;
    writeposbytes = 0;
    readposbytes  = 0;
    samples_done  = 0;

	return 1;
}

long seek( long time )
{
	 if (playing && time)
    {
        long skip_samples = (long)((float)samplerate*time/1000);
        samples_done = skip_samples;

		FLAC__bool ok = FLAC__stream_decoder_seek_absolute(mydecoder, samples_done);
		readposbytes = writeposbytes = 0;

        return (long)(1000.0f*skip_samples/samplerate);
    }

	return 0;
}

void cleanup( void )
{
	reset();
	__psp_free_heap();
}

int write_buffer( signed short *dest, int len )
{
	if(playing)
    {
        int writtenbytes = 0;
        for (;;)
        {
            int bytesleft = len - writtenbytes;
            int bufbytes  = writeposbytes - readposbytes;

            if (bytesleft <= bufbytes)
            {
                memcpy((char*)dest + writtenbytes, (char*)tmpbuf + readposbytes, bytesleft);
                readposbytes += bytesleft;
                writtenbytes = len;
                break;
            }
            else
            {
                memcpy((char*)dest + writtenbytes, (char*)tmpbuf + readposbytes, bufbytes);
                writtenbytes += bufbytes;
                readposbytes = writeposbytes = 0;
                if (!FLAC__stream_decoder_process_single(mydecoder))
                {
                    playing = 0;
                    return 0;
                }
                if (!writeposbytes)
                    break;
            }
        }
        samples_done += writtenbytes/(2*channels);
        return writtenbytes;
    }

    return 0;
}

void get_song_title( char *filename, char **title, int *length )
{
}

void get_packet(void)
{
	switch( itable.in_packet->ID ) {
		case NO_PACKET:
		default:
			break;
	};
}

void send_packet(void)
{
	switch( itable.out_packet->ID ) {
		/* song data */
		case SONG_TITLE:
			itable.out_packet->data = title;
			break;
		case SONG_ARTIST:
			itable.out_packet->data = artist;
			break;
		case SONG_ALBUM_GAME:
			itable.out_packet->data = album;
			break;
		case SONG_COPYRIGHT:
			itable.out_packet->data = copyright;
			break;
		case SONG_FORMAT:
			itable.out_packet->data = format;
			break;
		case SONG_LENGTH:
			itable.out_packet->data = &length;
			break;
		case SONG_SUBTUNES:
			itable.out_packet->data = &subtunes;
			break;

		/* audio info */
		case AUDIO_CHANNELS:
			itable.out_packet->data = &channels;
			break;
		case AUDIO_SAMPLERATE:
			itable.out_packet->data = &samplerate;
			break;
		case AUDIO_FORMAT:
			itable.out_packet->data = &aformat;
			break;
		case AUDIO_KBPS:
			itable.out_packet->data = &kbps;
			break;

		/* extra data */
		case ATTACHED_PICTURE:
			itable.out_packet->data = get_picture( &itable.out_packet->size );
			break;
		case NO_PACKET:
		default:
			itable.out_packet->size = 0;
			itable.out_packet->data = NULL;
			break;
	};
}


char *get_picture( int *size )
{
	*size = pic_size;
	return pic_data;
}


