
/*  Input Plugin for GameMusicGear ~
 *  Copyright (C) 2005 - 2010 gama, rockMax
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/* Default headers */

#include <pspkernel.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <malloc.h>

#include <ctype.h>

#include "../plugin.h"
#include "../SharedLib/libutils/utils.h"
#include "gme/Music_Emu.h"
#include "gme/gme.h"
#include "gme/blargg_source.h"

PSP_MODULE_INFO("IN_GME", 0, 1, 1);
PSP_NO_CREATE_MAIN_THREAD();
PSP_HEAP_SIZE_KB( 8 * 1024 );

#ifdef __cplusplus
extern "C" {
#endif
int module_start(SceSize argc, void* argp)
{
 	return 0;
}

int module_stop(int args, void *argp)
{
	return 0;
}
#ifdef __cplusplus
}
#endif

/* Prototypes */
void init(void);
void reset(void);
int load( char *filename );
int load_rw( char *filename, char *addr, int size );
void set_tune( int tune );
long seek( long time );
void cleanup( void );
int write_buffer( signed short *dest, int len );
void get_song_title( char *filename, char **title, int *length );

void get_packet(void);
void send_packet(void);
/* End Prototypes */

/* Variables */
Music_Emu *emu = NULL;
gme_err_t error;
gme_info_t *t_info = NULL;
int playing;
int inited;
int force_length;

/* Defualt Stereo Effects */
int enable_accuracy = 0;
int equalizer_effect = 1;
float stereo_depth = 0.2f;
float tempo = 1.0f;
float treble = 0.0f;
long bass = 20;

char filename[256];

typedef short sample_t;

/* Required */
int channels = 2;
int samplerate = 44100;
AFormat aformat = FMT_S16_LE;
int kbps = 128;

char title[128];
char artist[128];
char album[128];
char copyright[128];
char format[128];

long length;
int subtunes;

/** START of Plugin definitions setup */
InputPlugin itable =
{
	"Game_Music_Emu 0.5.2 Plugin",	/* The description of the input plugin */

	init,							/* Called when the plugin is loaded */
	reset, 						/* Called when need to update config */
	load,							/* Load the file */
	load_rw,						/* Load the file from data */
	set_tune,						/* Set song subtune */
	seek,							/* Seek to the specified time */
	NULL,							/* Set the equalizer, most plugins won't be able to do this */
	cleanup,						/* Called when gmgear exit */
	write_buffer, 					/* Fill audio buffer */
	get_song_title,					/* Function to grab the title string */

	get_packet,					/* Called when need main app data packet*/
	send_packet,				/* Send data packet to main app */

	NULL,
	NULL,							/* Filled by Main App */
	NULL,							/* Filled by Main App */
	0,								/* Filled by Main App */
	LATIN,							/* Filled by Main App */
	CLIENT_PLAYER					/* Player type */
};
/** END of Plugin definitions setup */

/** Single Plugin Export */
InputPlugin *get_iplugin_info()
{
	return &itable;
}

void init(void)
{
	force_length = 0;
	kbps = 0xFF;

	itable.in_packet->ID = NO_PACKET;
}

void reset(void)
{
	enable_accuracy = ReadBoolValue("enable_accuracy", itable.config);
	equalizer_effect = ReadBoolValue("equalizer_effect", itable.config);
	force_length = ReadIntValue("force_length", itable.config);
	stereo_depth = ReadFloatValue("stereo_depth", itable.config);
	tempo = ReadFloatValue("tempo", itable.config);
	treble = ReadFloatValue("treble", itable.config);
	bass = ReadIntValue("bass", itable.config);

	// TODO - checkme
	// channels = ReadIntValue("channels", itable.config);
	samplerate = ReadFreqValue("sample_rate", itable.config);

	strcpy(title, "");
	strcpy(album, "");
	strcpy(copyright, "");
	strcpy(artist, "");
	strcpy(format, "");

	length = 0;
	subtunes = 1;

	if( emu ) gme_delete( emu );
	playing = 0;
}

int load( char *file )
{
	if( (error = gme_open_file( file, &emu, samplerate )) ) {
        return 0;
    }
    if( emu == NULL ) return -1;

	/* Enable most emulation accuracy */
	gme_enable_accuracy( emu, enable_accuracy );

	/* Add some stereo enhancement */
	gme_set_stereo_depth( emu, stereo_depth );

	/* Adjust equalizer for crisp, bassy sound */
	if( equalizer_effect )
	{
		gme_equalizer_t eq;
		eq.treble = treble;
		eq.bass   = bass;
		gme_set_equalizer( emu, &eq );
	}

    strcpy(filename, get_filename( file ));

    playing = 1;
    set_tune( 0 );
	return 1;
}

int load_rw( char *file, char *addr, int size )
{
	/* Dirty hack to load vgz from memory */
	if( is_ext(file, ".vgz") )
	{
		char tmpmod[256];
		sprintf(tmpmod, "_%s", file);
		FILE *fp = fopen(tmpmod, "wb");
		if( fp == NULL ) return -2;
		fwrite(addr, 1, size, fp);
		fclose(fp);
		int ret = load(tmpmod);
		remove(tmpmod);
		return ret;
	}

	if( (error = gme_open_data( addr, size, &emu, samplerate )) ) {
		return 0;
    }

	if( emu == NULL ) return -1;

	/* Enable most emulation accuracy */
	gme_enable_accuracy( emu, enable_accuracy );

	/* Add some stereo enhancement */
	gme_set_stereo_depth( emu, stereo_depth );

	/* Adjust equalizer for crisp, bassy sound */
	if( equalizer_effect )
	{
		gme_equalizer_t eq;
		eq.treble = treble;
		eq.bass   = bass;
		gme_set_equalizer( emu, &eq );
	}

	strcpy(filename, get_filename( file ));
    playing = 1;
    set_tune( 0 );
    return 1;
}

void set_tune( int tune )
{
	char *tmp = get_extension(filename, true);
	strncpy(format, tmp, 32);
	free(tmp);

	gme_start_track( emu, tune );
    if( !(error = gme_track_info( emu, &t_info, tune )) )
    {
		if( force_length > 0 ) length = force_length * 1000;
        else if (t_info->length > 0)
            length = t_info->length;
        else if (t_info->intro_length >= 0 && t_info->loop_length > 0)
            length = t_info->intro_length + t_info->loop_length * 2;
        else
            length = (2*60 + 30) * 1000; // 2:30

        if ( t_info->game )
        {
			strncpy( album, t_info->game, 32 );
        }

        if ( t_info->song )
        {
            strncpy( title, t_info->song, 32 );
        }
		/* else
        {
			char *base = get_base(filename);
            sprintf( title, "%s (%d)", base, tune+1 );
			free(base);
        }
		*/

        if ( t_info->author )
        {
            strncpy( artist, t_info->author, 32 );
        }

        if ( t_info->copyright )
        {
            strncpy( copyright, t_info->copyright, 32 );
        }

        subtunes = gme_track_count( emu );
		if( subtunes <= 0 ) subtunes = 1;
    }
}

long seek( long time )
{
	if( (error = gme_seek( emu, time )) ) {
		return 0;
    }

	return gme_tell( emu );
}

void cleanup( void )
{
	if( emu ) gme_delete(emu);
	gme_free_info(t_info);

	playing = 0;

	__psp_free_heap();
}

int write_buffer( signed short *dest, int len )
{
	if(playing)
    {
        gme_play(emu, len>>1, dest);
        // TODO - remove this ? ...
		// if( gme_track_ended( emu ) ) return 0;
        return len;
    }

    return 0;
}

void get_song_title( char *filename, char **title, int *length )
{
}

void get_packet(void)
{
	switch( itable.in_packet->ID ) {
		case NO_PACKET:
		default:
			break;
	};
}

void send_packet(void)
{
	switch( itable.out_packet->ID ) {
		/* song data */
		case SONG_TITLE:
			itable.out_packet->data = title;
			break;
		case SONG_ARTIST:
			itable.out_packet->data = artist;
			break;
		case SONG_ALBUM_GAME:
			itable.out_packet->data = album;
			break;
		case SONG_COPYRIGHT:
			itable.out_packet->data = copyright;
			break;
		case SONG_FORMAT:
			itable.out_packet->data = format;
			break;
		case SONG_LENGTH:
			itable.out_packet->data = &length;
			break;
		case SONG_SUBTUNES:
			itable.out_packet->data = &subtunes;
			break;

		/* audio info */
		case AUDIO_CHANNELS:
			itable.out_packet->data = &channels;
			break;
		case AUDIO_SAMPLERATE:
			itable.out_packet->data = &samplerate;
			break;
		case AUDIO_FORMAT:
			itable.out_packet->data = &aformat;
			break;
		case AUDIO_KBPS:
			itable.out_packet->data = &kbps;
			break;

		/* extra data */
		case ATTACHED_PICTURE:
			itable.out_packet->size = 0;
			itable.out_packet->data = NULL;
			break;
		case NO_PACKET:
		default:
			itable.out_packet->data = NULL;
			break;
	};
}
