/*  Input Plugin for GameMusicGear ~
 *  Copyright (C) 2005 - 2010 gama, rockMax
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/* Default headers */
#include <stdlib.h>
#include <string.h>

#include <pspkernel.h>
#include "../plugin.h"

extern "C" {
#include "hvl_replay.h"
}

PSP_MODULE_INFO("IN_HVL", 0, 1, 1);
PSP_NO_CREATE_MAIN_THREAD();
PSP_HEAP_SIZE_KB( 2 * 1024 );

#ifdef __cplusplus
extern "C" {
#endif
int module_start(SceSize argc, void* argp)
{
 	return 0;
}

int module_stop(int args, void *argp)
{
	return 0;
}
#ifdef __cplusplus
}
#endif

/* Prototypes */
void init(void);
void reset(void);
int  load( char *filename );
int  load_rw( char *filename, char *addr, int size );
void set_tune( int tune );
long  seek( long time );
void cleanup( void );
int  write_buffer( signed short *dest, int len );
void get_song_title( char *filename, char **title, int *length );

void get_packet(void);
void send_packet(void);
/* End Prototypes */

/* Variables */
struct hvl_tune *ht = NULL;
char *tmpbuf = NULL;
int writeposbytes = 0;
int readposbytes  = 0;

int stereosep  = 2;
int playing    = 0;
int channels   = 2;
int samplerate = 44100;
int bps		   = 16;
AFormat aformat = FMT_S16_LE;
int kbps = 128;

char title[128];
char artist[128];
char album[128];
char copyright[128];
char format[128];

long length;
int subtunes;
/* End Variables */

/** START of Plugin definitions setup */
InputPlugin itable =
{
	/* Populated by Plugin */
	"Hively Tracker v1.6",				/* The description of the input plugin */

	init,							/* Called when the plugin is loaded */
	reset, 							/* Called when need to update config */
	load,							/* Load the file */
	load_rw,						/* Load the file from data */
	set_tune,						/* Set song subtune */
	seek,							/* Seek to the specified time */
	NULL,							/* Set the equalizer, most plugins won't be able to do this */
	cleanup,						/* Called when gmgear exit */
	write_buffer, 					/* Fill audio buffer */
	get_song_title,					/* Function to grab the title string */

	get_packet,					/* Called when need main app data packet*/
	send_packet,				/* Send data packet to main app */

	NULL,						/* Filled by Main App */
	NULL,						/* Filled by Main App */
	NULL,						/* Filled by Main App */
	0,								/* Filled by Main App */
	LATIN,							/* Filled by Main App */
	CLIENT_PLAYER					/* Player type */
};
/** END of Plugin definitions setup */

/** Single Plugin Export */
#ifdef PSP
InputPlugin *get_iplugin_info()
#else
InputPlugin *dll_get_iplugin_info()
#endif
{
	return &itable;
}

void init(void)
{
	strcpy(format, "HVL");
	subtunes = 1;

	kbps = 0x6C;
	aformat = FMT_S16_LE;

	hvl_InitReplayer();

	itable.in_packet->ID = NO_PACKET;
}

void reset(void)
{
	channels   = ReadIntValue("channels", itable.config);
	samplerate = ReadFreqValue("samplerate", itable.config);
	stereosep  = ReadIntValue("stereosep", itable.config);

	strcpy(title, "");
	strcpy(album, "");
	strcpy(copyright, "");
	strcpy(artist, "");

	length = 0;
	playing = 0;

	// Cleanup
	if( tmpbuf ) {
		free(tmpbuf);
		tmpbuf = NULL;
	}

	if( ht ) {
		hvl_FreeTune(ht);
		ht = NULL;
	}

	itable.in_packet->ID = NO_PACKET;
}

int load( char *file )
{
	ht = hvl_LoadTune( file, samplerate, stereosep );
	if( !ht ) return -1;

	if( strlen(ht->ht_Name) > 0 ) strncpy(title, ht->ht_Name, 128);

	// Big enough to play all song
    length     = hvl_GetLen(ht, 0);
	hvl_InitSubsong(ht, 0);
	if( length <= 0 )
		length = 1800 * 1000;  // Just big enough for now (30 min)

	subtunes = ht->ht_SongNum+1;
	if( subtunes <= 0 ) subtunes = 1;

	tmpbuf = (char *)malloc((samplerate*(bps/8)*channels) * 2);
	if( !tmpbuf ) return -2;

	writeposbytes = readposbytes = 0;

	playing = 1;
	return 1;
}

int  load_rw( char *filename, char *addr, int size )
{
	ht = hvl_LoadTune_RW( addr, size, samplerate, stereosep );
	if( !ht ) return -1;

	if( strlen(ht->ht_Name) > 0 ) strncpy(title, ht->ht_Name, 128);

	// Big enough to play all song
    length     = hvl_GetLen(ht, 0);
	hvl_InitSubsong(ht, 0);
	if( length <= 0 )
		length = 1800 * 1000;  // Just big enough for now (30 min)

	subtunes = ht->ht_SongNum+1;
	if( subtunes <= 0 ) subtunes = 1;

	tmpbuf = (char *)malloc((samplerate*(bps/8)*channels) * 2);
	if( !tmpbuf ) return -2;

	writeposbytes = readposbytes = 0;

	playing = 1;
	return 1;
}

long seek( long time )
{
	if( playing ) {
		// Seek function (time in msec)
		hvl_Seek(ht, time);
		return time;
	}

    return 0;
}

void set_tune( int tune )
{
	// Big enough to play all song
    length     = hvl_GetLen(ht, tune);
	hvl_InitSubsong(ht, tune);
	if( length <= 0 )
		length = 1800 * 1000;  // Just big enough for now (30 min)

	memset(tmpbuf, 0, sizeof(tmpbuf));
	writeposbytes = readposbytes = 0;
}

void cleanup( void )
{
	// Cleanup
	reset();
	__psp_free_heap();
}

int write_buffer( signed short *dest, int len )
{
	if (playing)
    {
        int writtenbytes = 0;
        for (;;)
        {
            int bytesleft = len - writtenbytes;
            int bufbytes  = writeposbytes - readposbytes;

            if (bytesleft <= bufbytes)
            {
                memcpy((char*)dest + writtenbytes, (char*)tmpbuf + readposbytes, bytesleft);
                readposbytes += bytesleft;
                writtenbytes = len;
                break;
            }
            else
            {
                memcpy((char*)dest + writtenbytes, (char*)tmpbuf + readposbytes, bufbytes);
                writtenbytes += bufbytes;
                readposbytes = writeposbytes = 0;

                hvl_DecodeFrame(ht, tmpbuf, tmpbuf+(bps/8), (bps/8)*channels);
				writeposbytes = (samplerate*(bps/8)*channels)/50;
            }
        }

        return writtenbytes;
    }

    return 0;
}

void get_song_title( char *filename, char **title, int *length )
{
}

void get_packet(void)
{
	switch( itable.in_packet->ID ) {
		case NO_PACKET:
		default:
			break;
	};
}

void send_packet(void)
{
	switch( itable.out_packet->ID ) {
		/* song data */
		case SONG_TITLE:
			itable.out_packet->data = title;
			break;
		case SONG_ARTIST:
			itable.out_packet->data = artist;
			break;
		case SONG_ALBUM_GAME:
			itable.out_packet->data = album;
			break;
		case SONG_COPYRIGHT:
			itable.out_packet->data = copyright;
			break;
		case SONG_FORMAT:
			itable.out_packet->data = format;
			break;
		case SONG_LENGTH:
			itable.out_packet->data = &length;
			break;
		case SONG_SUBTUNES:
			itable.out_packet->data = &subtunes;
			break;

		/* audio info */
		case AUDIO_CHANNELS:
			itable.out_packet->data = &channels;
			break;
		case AUDIO_SAMPLERATE:
			itable.out_packet->data = &samplerate;
			break;
		case AUDIO_FORMAT:
			itable.out_packet->data = &aformat;
			break;
		case AUDIO_KBPS:
			itable.out_packet->data = &kbps;
			break;

		/* extra data */
		case ATTACHED_PICTURE:
			itable.out_packet->size = 0;
			itable.out_packet->data = NULL;
			break;
		case NO_PACKET:
		default:
			itable.out_packet->size = 0;
			itable.out_packet->data = NULL;
			break;
	};
}
