/*  Input Plugin for GameMusicGear ~
 *  Copyright (C) 2005 - 2010 gama, rockMax
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/* Default headers */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <pspkernel.h>
#include <pspkerneltypes.h>
#include <pspiofilemgr.h>

#include <ctype.h>

#include "../plugin.h"
#include "libmad/include/mad.h"

PSP_MODULE_INFO("IN_MAD", 0, 1, 1);
PSP_NO_CREATE_MAIN_THREAD();
PSP_HEAP_SIZE_KB(1024 * 2);

#ifdef __cplusplus
extern "C" {
#endif
int module_start(SceSize argc, void* argp)
{
 	return 0;
}

int module_stop(int args, void *argp)
{
	return 0;
}
#ifdef __cplusplus
}
#endif

#define TAGLIB
#ifdef TAGLIB
#include "fileref.h"
#include "tag.h"
#include "toolkit/tbytevector.h"
#include "mpeg/mpegfile.h"
#include "mpeg/id3v2/id3v2tag.h"
#include "mpeg/id3v2/id3v2frame.h"
#include "mpeg/id3v2/id3v2header.h"
#include "mpeg/id3v2/frames/attachedpictureframe.h"
using namespace TagLib;
#endif

// LINKER PATCH
extern "C" {
void *__dso_handle = NULL;
}

/* Prototypes */
void init(void);
void reset(void);
int load( char *filename );
long seek( long time );
void cleanup( void );
int write_buffer( signed short *dest, int len );
void get_song_title( char *filename, char **title, int *length );

void get_packet(void);
void send_packet(void);

char *get_picture( int *size );
/* End Prototypes */

/* Variables */
#define IN_SIZE (8192)
int sampperframe[] = {0, 384, 1152, 1152};
SceUID mp3_file = -1;

struct mad_stream   *stream = NULL;
struct mad_frame    *frame = NULL;
struct mad_synth    *synth = NULL;

int playing = 0;
int fsize;

int samples_done;

int tag_offset = 0;
static signed short *tmpbuf = NULL;
static unsigned char *inbuf = NULL;
int readposbytes;
int writeposbytes;

/* Required */
int channels = 2;
int samplerate = 44100;
AFormat aformat = FMT_S16_LE;
int kbps = 128;

char title[128];
char artist[128];
char album[128];
char copyright[128];
char format[128];

long length;
int subtunes;

bool string_utf8;

// picture info
char *pic_data;
int   pic_size;

/** START of Plugin definitions setup */
InputPlugin itable =
{
	"LibMAD Plugin",		/* The description of the input plugin */

	init,						/* Called when the plugin is loaded */
	reset, 						/* Called when need to update config */
	load,						/* Load the file */
	NULL,						/* Load the file from data */
	NULL,						/* Set song subtune */
	seek,						/* Seek to the specified time */
	NULL,						/* Set the equalizer, most plugins won't be able to do this */
	cleanup,					/* Called when gmgear exit */
	write_buffer, 				/* Fill audio buffer */
	get_song_title,				/* Function to grab the title string */

	get_packet,					/* Called when need main app data packet*/
	send_packet,				/* Send data packet to main app */

	NULL,						/* Filled by Main App */
	NULL,						/* Filled by Main App */
	NULL,						/* Filled by Main App */
	0,							/* Filled by Main App */
	LATIN,						/* Filled by Main App */
	CLIENT_PLAYER					/* Player type */
};
/** END of Plugin definitions setup */

/** Single Plugin Export */
#ifdef PSP
InputPlugin *get_iplugin_info()
#else
InputPlugin *dll_get_iplugin_info()
#endif
{
	return &itable;
}


/* LIBMAD HANDLERS */
void mad_stream_buffer_clear(struct mad_stream *stream)
{
	int left = stream->bufend - stream->this_frame;
	int size = stream->this_frame - stream->buffer;

    memcpy((void *)stream->buffer, (void *)stream->this_frame, left);

	stream->bufend = stream->buffer + left;
	stream->this_frame -= size;
	stream->next_frame -= size;
	stream->ptr.byte -= size;
}

int handle_input(struct mad_stream *stream)
{
	mad_stream_buffer_clear(stream);
	int len = sceIoRead(mp3_file, (void *)stream->bufend, IN_SIZE);

	if(len < IN_SIZE)
		playing = 0;
	stream->bufend += len;
    return len;
}

int handle_output(struct mad_pcm *pcm)
{
	signed int sample;
	int nSamples  = pcm->length; // Per channel
	mad_fixed_t *left_ch   = pcm->samples[0];
    mad_fixed_t *right_ch;

    if (channels == 2)
        right_ch  = pcm->samples[1];

	short *outp = tmpbuf;

	while(nSamples--)
	{
		sample = *left_ch++;
		sample += (1L << (MAD_F_FRACBITS - 16));
		if (sample >= MAD_F_ONE)
			sample = MAD_F_ONE - 1;
		else if (sample < -MAD_F_ONE)
			sample = -MAD_F_ONE;
        signed int s = sample >> (MAD_F_FRACBITS + 1 - 16);
        *outp++ = s;

        if (channels == 2)
        {
            sample = *right_ch++;
            sample += (1L << (MAD_F_FRACBITS - 16));
            if (sample >= MAD_F_ONE)
                sample = MAD_F_ONE - 1;
            else if (sample < -MAD_F_ONE)
                sample = -MAD_F_ONE;
            *outp++ = sample >> (MAD_F_FRACBITS + 1 - 16);
        }
	}
	return pcm->length*2*channels;
}
/* END LIBMAD HANDLERS */

void init(void)
{
	strcpy(format, "MP3");
	subtunes = 1;
	aformat = FMT_S16_LE;

	pic_data = NULL;
	pic_size = 0;

	if( itable.charset == UTF8 ) string_utf8 = true;
	else string_utf8 = false;

	itable.in_packet->ID = NO_PACKET;
}

void read_tag(char *file)
{
#ifdef TAGLIB
	MPEG::File *tagfile = NULL;
	tagfile = new MPEG::File(file, true, MPEG::Properties::Accurate);

	if( !tagfile ) {
		printf("Warning: Cannot create MPEG tag\n");
		return;
	}

	if( tagfile->tag() && !tagfile->tag()->isEmpty() )
	{
		TagLib::Tag *tag = tagfile->tag();

		if( tagfile->ID3v2Tag() ) {
			tag_offset = tagfile->ID3v2Tag()->header()->completeTagSize();

			// Get the list of frames for a specific frame type
			ID3v2::FrameList l = tagfile->ID3v2Tag()->frameListMap()["APIC"];

			if( !l.isEmpty() ) {
				pspDebugScreenPrintf("[Picture Found]\n");
				ID3v2::AttachedPictureFrame *pic =
				static_cast<ID3v2::AttachedPictureFrame *> (l.front());
				pic_size = pic->picture().size();

				if( pic->picture().size() > 0 ) pic_data = new char [pic->picture().size()];

				if( pic_data != NULL )
					memcpy(pic_data, pic->picture().data(), pic->picture().size());
			}
		}

		if( !tag->title().isEmpty() ) strncpy(title, tag->title().toCString(), 128);
		if( !tag->album().isEmpty() ) strncpy(album, tag->album().toCString(), 128);
		if( tag->year() > 0 ) snprintf(copyright, 128, "%d", tag->year(), 128);
		if( !tag->artist().isEmpty() ) strncpy(artist, tag->artist().toCString(), 128);
	}

	if( tagfile->audioProperties() )
	{
		AudioProperties *properties = tagfile->audioProperties();
		length = properties->length() * 1000;

		channels = properties->channels();
		samplerate = properties->sampleRate();
		kbps = properties->bitrate();
	}

	delete tagfile;
#endif
}

int load( char *file )
{
	read_tag(file);

	tmpbuf = new signed short [1024*8];
	if( tmpbuf == NULL ) return 0;
	inbuf = new unsigned char [IN_SIZE*6];
	if( inbuf == NULL ) return 0;

	stream = new struct mad_stream;
	if( stream == NULL ) return 0;
	frame = new struct mad_frame;
	if( frame == NULL ) return 0;
	synth = new struct mad_synth;
	if( synth == NULL ) return 0;

	mad_stream_init(stream);
    mad_frame_init(frame);
    mad_synth_init(synth);

    mp3_file = sceIoOpen(file, PSP_O_RDONLY, 0777);
	if( mp3_file < 0 ) return 0;

	fsize = sceIoLseek(mp3_file, 0, PSP_SEEK_END);
	sceIoLseek(mp3_file, tag_offset, PSP_SEEK_SET);

	int len = sceIoRead(mp3_file, inbuf, IN_SIZE);
    mad_stream_buffer(stream, (const unsigned char *)inbuf, len);

	readposbytes = writeposbytes = 0;

    samples_done  = 0;
	playing = 1;

	return 1;
}


int _skip_bytes( int len )
{
	if(playing)
	{
        int writtenbytes = 0;
        for (;;)
        {
            int bytesleft = len - writtenbytes;
            int bufbytes  = writeposbytes - readposbytes;
            if (bytesleft <= bufbytes)
            {
                //memcpy((char*)dest + writtenbytes, (char*)tmpbuf + readposbytes, bytesleft);
                readposbytes += bytesleft;
                writtenbytes = len;
                break;
            }
            else
            {
                //memcpy((char*)dest + writtenbytes, (char*)tmpbuf + readposbytes, bufbytes);
                writtenbytes += bufbytes;
                readposbytes = writeposbytes = 0;
                for (;;)
                {
                    if ((stream->bufend - stream->next_frame) < IN_SIZE)
                        if (handle_input(stream) <= 0)
                            return writtenbytes;

                    if (mad_frame_decode(frame, stream) != -1)
                        break;
                }

                mad_synth_frame(synth, frame);
                writeposbytes = handle_output(&synth->pcm);
                if (!writeposbytes)
                    break;
            }
        }
        samples_done += writtenbytes/(2*channels);
        return writtenbytes;
	}

	return len;
}

//Seek next valid frame
//NOTE: this function comes from Music prx 0.55 source
//      all credits goes to joek2100.
int SeekNextFrameMP3(SceUID fd)
{
    int offset = 0;
    unsigned char buf[1024];
    unsigned char *pBuffer;
    int i;
    int size = 0;

    offset = sceIoLseek32(fd, 0, PSP_SEEK_CUR);
    sceIoRead(fd, buf, sizeof(buf));
    if (!strncmp((char*)buf, "ID3", 3) || !strncmp((char*)buf, "ea3", 3)) //skip past id3v2 header, which can cause a false sync to be found
    {
        //get the real size from the syncsafe int
        size = buf[6];
        size = (size<<7) | buf[7];
        size = (size<<7) | buf[8];
        size = (size<<7) | buf[9];

        size += 10;

        if (buf[5] & 0x10) //has footer
            size += 10;
    }

    sceIoLseek32(fd, offset, PSP_SEEK_SET); //now seek for a sync
    while(1)
    {
        offset = sceIoLseek32(fd, 0, PSP_SEEK_CUR);
        size = sceIoRead(fd, buf, sizeof(buf));

        if (size <= 2)//at end of file
            return -1;

        if (!strncmp((char*)buf, "EA3", 3))//oma mp3 files have non-safe ints in the EA3 header
        {
            sceIoLseek32(fd, (buf[4]<<8)+buf[5], PSP_SEEK_CUR);
            continue;
        }

        pBuffer = buf;
        for( i = 0; i < size; i++)
        {
            //if this is a valid frame sync (0xe0 is for mpeg version 2.5,2+1)
            if ( (pBuffer[i] == 0xff) && ((pBuffer[i+1] & 0xE0) == 0xE0))
            {
                offset += i;
                sceIoLseek32(fd, offset, PSP_SEEK_SET);
                return offset;
            }
        }
       //go back two bytes to catch any syncs that on the boundary
        sceIoLseek32(fd, -2, PSP_SEEK_CUR);
    }
}

long seek( long time )
{
	if (time)
    {
        int bitrate   = frame->header.bitrate;
        // Should i have plugin.channels in here? Not sure
        int bytesperframe = sampperframe[frame->header.layer]*bitrate/**channels*//
            (samplerate*8);

        int seekbytes = (int)((float)bitrate*time/(8*1000));
		int curpos = sceIoLseek(mp3_file, 0, PSP_SEEK_CUR);

        if (seekbytes < curpos )
			sceIoLseek(mp3_file, seekbytes, PSP_SEEK_SET);
        else
			sceIoLseek(mp3_file, seekbytes - curpos, PSP_SEEK_CUR);

		// Sync frames
		int newpos = SeekNextFrameMP3(mp3_file);

        // Reduce the delay? but how much? unsure... this sucks ass.
        mad_stream_skip(stream, 48000);
		mad_stream_sync(stream);
        readposbytes = writeposbytes = 0;
		memset(tmpbuf, 0, sizeof(tmpbuf));

		int actual_msecs = (int)(8 * 1000.0f * (newpos - curpos) / bitrate);
		int samp = samples_done + (int)((float)actual_msecs / 1000.0f *
                    samplerate * channels / 2);
        samples_done = samp > 0 ? samp : 0;
        return actual_msecs;
    }

	return 0;
}

void reset()
{
	strcpy(title, "");
	strcpy(album, "");
	strcpy(copyright, "");
	strcpy(artist, "");

	tag_offset = 0;
	length = 200000;
	kbps = 128;

	if( pic_data ) {
		delete [] pic_data;
		pic_data = NULL;
		pic_size = 0;
	}

	if (mp3_file >= 0) {
        sceIoClose(mp3_file);
        mp3_file = -1;
    }

	if (tmpbuf) {
		delete [] tmpbuf;
	}

	if (inbuf) {
		delete [] inbuf;
	}

	if( playing ) {
		mad_synth_finish(synth);
		mad_frame_finish(frame);
		mad_stream_finish(stream);

		if( stream ) delete stream;
		if( frame ) delete frame;
		if( synth ) delete synth;
		playing = 0;
	}

	itable.in_packet->ID = NO_PACKET;
}

void cleanup( void )
{
	reset();
	__psp_free_heap();
}

int write_buffer( signed short *dest, int len )
{
	if(playing)
	{
        int writtenbytes = 0;
        for (;;)
        {
            int bytesleft = len - writtenbytes;
            int bufbytes  = writeposbytes - readposbytes;
            if (bytesleft <= bufbytes)
            {
                memcpy((char*)dest + writtenbytes, (char*)tmpbuf + readposbytes, bytesleft);
                readposbytes += bytesleft;
                writtenbytes = len;
                break;
            }
            else
            {
                memcpy((char*)dest + writtenbytes, (char*)tmpbuf + readposbytes, bufbytes);
                writtenbytes += bufbytes;
                readposbytes = writeposbytes = 0;
                for (;;)
                {
                    if ((stream->bufend - stream->next_frame) < IN_SIZE)
                        if (handle_input(stream) <= 0)
                            return writtenbytes;

                    if (mad_frame_decode(frame, stream) != -1)
                        break;
                }

                mad_synth_frame(synth, frame);
                writeposbytes = handle_output(&synth->pcm);
                if (!writeposbytes)
                    break;
            }
        }
        samples_done += writtenbytes/(2*channels);
        return writtenbytes;
	}
	else
	{
		memset(dest, 0, len);
	}

	return len;
}

void get_song_title( char *filename, char **title, int *length )
{
}

void get_packet(void)
{
	switch( itable.in_packet->ID ) {
		case NO_PACKET:
		default:
			break;
	};
}

void send_packet(void)
{
	switch( itable.out_packet->ID ) {
		/* song data */
		case SONG_TITLE:
			itable.out_packet->data = title;
			break;
		case SONG_ARTIST:
			itable.out_packet->data = artist;
			break;
		case SONG_ALBUM_GAME:
			itable.out_packet->data = album;
			break;
		case SONG_COPYRIGHT:
			itable.out_packet->data = copyright;
			break;
		case SONG_FORMAT:
			itable.out_packet->data = format;
			break;
		case SONG_LENGTH:
			itable.out_packet->data = &length;
			break;
		case SONG_SUBTUNES:
			itable.out_packet->data = &subtunes;
			break;

		/* audio info */
		case AUDIO_CHANNELS:
			itable.out_packet->data = &channels;
			break;
		case AUDIO_SAMPLERATE:
			itable.out_packet->data = &samplerate;
			break;
		case AUDIO_FORMAT:
			itable.out_packet->data = &aformat;
			break;
		case AUDIO_KBPS:
			itable.out_packet->data = &kbps;
			break;

		/* extra data */
		case ATTACHED_PICTURE:
			itable.out_packet->data = get_picture( &itable.out_packet->size );
			break;
		case NO_PACKET:
		default:
			itable.out_packet->size = 0;
			itable.out_packet->data = NULL;
			break;
	};
}

char *get_picture( int *size )
{
	*size = pic_size;
	return pic_data;
}


