/*  Input Plugin for GameMusicGear ~
 *  Copyright (C) 2005 - 2010 gama, rockMax
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/* Default headers */

#ifdef PSP
#include <pspkernel.h>
#include <pspdebug.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <malloc.h>

#include <ctype.h>

#include "../plugin.h"

extern "C" {
#include "libtimidity-0.1.0/src/timidity.h"
}

#define DEBUG(...) pspDebugScreenPrintf(__VA_ARGS__)

#ifdef PSP
PSP_MODULE_INFO("IN_MIDI", 0, 1, 1);
PSP_NO_CREATE_MAIN_THREAD();
PSP_HEAP_SIZE_KB( 1024 * 8 );

#ifdef __cplusplus
extern "C" {
#endif
int module_start(SceSize argc, void* argp)
{
 	return 0;
}

int module_stop(int args, void *argp)
{
	return 0;
}
#ifdef __cplusplus
}
#endif
#endif

/* Prototypes */
void init(void);
void reset(void);
int load( char *filename );
int load_rw( char *filename, char *addr, int size );
long seek( long time );
void cleanup( void );
int write_buffer( signed short *dest, int len );
void get_song_title( char *filename, char **title, int *length );

void get_packet(void);
void send_packet(void);
/* End Prototypes */

/* Variables */
char config_path[256];
MidIStream *stream = NULL;
MidSong *song = NULL;
MidSongOptions options;

int playing = 0;
int buf_size = 4096;
int loaded = 0;

/* Required */
int channels = 2;
int samplerate = 44100;
AFormat aformat = FMT_S16_LE;
int kbps = 128;

char title[128];
char artist[128];
char album[128];
char copyright[128];
char format[128];

long length;
int subtunes;

/* End Variables */

/** START of Plugin definitions setup */
InputPlugin itable =
{
	/* Populated by Plugin */
	"Timidity",						/* The description of the input plugin */

	init,							/* Called when the plugin is loaded */
	reset, 							/* Called when need to update config */
	load,							/* Load the file */
	load_rw,						/* Load the file from data */
	NULL,							/* Set song subtune */
	seek,							/* Seek to the specified time */
	NULL,							/* Set the equalizer, most plugins won't be able to do this */
	cleanup,						/* Called when gmgear exit */
	write_buffer, 					/* Fill audio buffer */
	get_song_title,					/* Function to grab the title string */

	get_packet,					/* Called when need main app data packet*/
	send_packet,				/* Send data packet to main app */

	NULL,						/* Filled by Main App */
	NULL,						/* Filled by Main App */
	NULL,						/* Filled by Main App */
	0,								/* Filled by Main App */
	LATIN,							/* Filled by Main App */
	CLIENT_PLAYER					/* Player type */
};
/** END of Plugin definitions setup */

/** Single Plugin Export */
#ifdef PSP
InputPlugin *get_iplugin_info()
#else
InputPlugin *dll_get_iplugin_info()
#endif
{
	return &itable;
}

void init(void)
{
	strcpy(format, "MIDI");
	subtunes = 1;

	kbps = 0x6C;
	aformat = FMT_S16_LE;

	const char *token = ReadStringValue("config_path", itable.config);
	if( token ) strncpy(config_path, token, 256);
	else strncpy(config_path, "ms0:/.gmgear/timidity/timidity.cfg", 256);

	itable.in_packet->ID = NO_PACKET;
}

void reset(void)
{
	channels = ReadIntValue("channels", itable.config);
	samplerate = ReadFreqValue("samplerate", itable.config);
	buf_size = ReadIntValue("buf_size", itable.config);
	if( buf_size <= 0 ) buf_size = 4096;

	strcpy(title, "");
	strcpy(album, "");
	strcpy(copyright, "");
	strcpy(artist, "");

	length = 0;

	if( song ) {
        mid_song_free(song);
        song = NULL;
    }

	if( stream ) {
		mid_istream_close(stream);
		stream = NULL;
	}

	itable.in_packet->ID = NO_PACKET;
}

int load( char *file )
{
	if( !loaded ) {
		if( mid_init(config_path) < 0 ) return -1;
		loaded = 1;
	}

	if( !(stream = mid_istream_open_file(file)) ) {
        mid_exit();
        return 0;
    }

    options.rate = samplerate;
    options.format = MID_AUDIO_S16LSB;
    options.channels = channels;
    options.buffer_size = buf_size; // Ugh.. i don't want to specify this here.

	song = mid_song_load(stream, &options);
    mid_istream_close (stream);
	stream = NULL;

    if (!song) {
        mid_exit ();
        return -2;
    }

	mid_song_set_volume(song, 60);
    mid_song_start(song);

	char *mtitle, *mcopyright;
	mtitle = mid_song_get_meta(song, MID_SONG_TEXT);
	mcopyright = mid_song_get_meta(song, MID_SONG_COPYRIGHT);

	if( mtitle ) strncpy(title, mtitle, 128);
	if( mcopyright ) strncpy(copyright, mcopyright, 128);

	length = mid_song_get_total_time(song);

	playing = 1;
	return 1;
}

int load_rw( char *file, char *addr, int size )
{
	if( !loaded ) {
		if( mid_init(config_path) < 0 ) return 0;
		loaded = 1;
	}

	if( !(stream = mid_istream_open_mem( addr, size, 0)) ) {
		mid_exit();
		return 0;
	}

    options.rate = samplerate;
    options.format = MID_AUDIO_S16LSB;
    options.channels = channels;
    options.buffer_size = buf_size; // Ugh.. i don't want to specify this here.

	song = mid_song_load(stream, &options);
    mid_istream_close (stream);

    if (!song) {
        mid_exit ();
        return 0;
    }

	mid_song_set_volume(song, 60);
    mid_song_start(song);

	char *mtitle, *mcopyright;
	mtitle = mid_song_get_meta(song, MID_SONG_TEXT);
	mcopyright = mid_song_get_meta(song, MID_SONG_COPYRIGHT);

	if( mtitle ) strncpy(title, mtitle, 128);
	if( mcopyright ) strncpy(copyright, mcopyright, 128);

	length = mid_song_get_total_time(song);

	playing = 1;
	return 1;
}

long seek( long time )
{
	if( playing ) {
		mid_song_seek(song, time);
		return mid_song_get_time(song);
	}

    return 0;
}

void cleanup( void )
{
	if( loaded ) {
		if (song) {
			mid_song_free(song);
			song = NULL;
		}

		if( stream ) {
			mid_istream_close(stream);
			stream = NULL;
		}

		mid_exit();
	}

	__psp_free_heap();
}

int write_buffer( signed short *dest, int len )
{
	if(playing)
    {
        int writtenbytes = mid_song_read_wave(song, dest, len);
        return writtenbytes;
    }
    else
        memset(dest, 0, len);
    return len;
}

void get_song_title( char *filename, char **title, int *length )
{
}

void get_packet(void)
{
	switch( itable.in_packet->ID ) {
		case NO_PACKET:
		default:
			break;
	};
}

void send_packet(void)
{
	switch( itable.out_packet->ID ) {
		/* song data */
		case SONG_TITLE:
			itable.out_packet->data = title;
			break;
		case SONG_ARTIST:
			itable.out_packet->data = artist;
			break;
		case SONG_ALBUM_GAME:
			itable.out_packet->data = album;
			break;
		case SONG_COPYRIGHT:
			itable.out_packet->data = copyright;
			break;
		case SONG_FORMAT:
			itable.out_packet->data = format;
			break;
		case SONG_LENGTH:
			itable.out_packet->data = &length;
			break;
		case SONG_SUBTUNES:
			itable.out_packet->data = &subtunes;
			break;

		/* audio info */
		case AUDIO_CHANNELS:
			itable.out_packet->data = &channels;
			break;
		case AUDIO_SAMPLERATE:
			itable.out_packet->data = &samplerate;
			break;
		case AUDIO_FORMAT:
			itable.out_packet->data = &aformat;
			break;
		case AUDIO_KBPS:
			itable.out_packet->data = &kbps;
			break;

		/* extra data */
		case ATTACHED_PICTURE:
			itable.out_packet->size = 0;
			itable.out_packet->data = NULL;
			break;
		case NO_PACKET:
		default:
			itable.out_packet->size = 0;
			itable.out_packet->data = NULL;
			break;
	};
}


