
/*  Input Plugin for GameMusicGear ~
 *  Copyright (C) 2005 - 2010 gama, rockMax
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/* Default headers */

#include <pspkernel.h>
#include <pspdebug.h>
#include <pspiofilemgr.h>
#include <pspiofilemgr_fcntl.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <malloc.h>

#include <ctype.h>

#include "../plugin.h"
#include "../SharedLib/libutils/utils.h"

#include "libmodplug-0.8.7/src/libmodplug/modplug.h"

// LINKER PATCH
extern "C" {
void *__dso_handle = NULL;
}

PSP_MODULE_INFO("IN_MODPLUG", 0, 1, 1);
PSP_NO_CREATE_MAIN_THREAD();
PSP_HEAP_SIZE_KB( 1024 * 6 );

#ifdef __cplusplus
extern "C" {
#endif
int module_start(SceSize argc, void* argp)
{
 	return 0;
}

int module_stop(int args, void *argp)
{
	return 0;
}
#ifdef __cplusplus
}
#endif

/* Prototypes */
void init(void);
void reset(void);
int load( char *filename );
int load_rw( char *filename, char *addr, int size );
long seek( long time );
void cleanup( void );
int write_buffer( signed short *dest, int len );
void get_song_title( char *filename, char **title, int *length );

void get_packet(void);
void send_packet(void);
/* End Prototypes */

/* Variables */

int playing = 0;

/* Required */
int channels = 2;
int samplerate = 44100;
AFormat aformat = FMT_S16_LE;
int kbps = 128;

char title[128];
char artist[128];
char album[128];
char copyright[128];
char format[128];

long length;
int subtunes;

short resampling_mode = 1;
bool oversampling = true;
bool noise_reduction = true;
bool reverb = true;
bool megabass = false;
bool surround = false;
int reverb_depth = 40;
int reverb_delay = 80;
int bass_amount = 40;
int bass_range = 40;
int surround_depth = 40;
int surround_delay = 10;

ModPlugFile *mpfile = NULL;
ModPlug_Settings settings;

extern char TIMIDITYCFG[256];
extern char PATHFORPAT[256];
/* End Variables */

/* Extra Functions */
char *getFileData(char *filename, long *size)
{
	SceUID mFile;
    char *data;
	size_t readbytes;

	mFile = sceIoOpen(filename, PSP_O_RDONLY, 0777);
	if( mFile < 0 ) return NULL;
    *size = sceIoLseek(mFile, 0, PSP_SEEK_END);
	sceIoLseek(mFile, 0, PSP_SEEK_SET);
    data = new char [*size];
	readbytes = sceIoRead(mFile, data, *size);

	if (mFile > 0)
		sceIoClose(mFile);

    return(data);
}

/* End Extra Functions */

/* START of Plugin definitions setup */
InputPlugin itable =
{
	/* Populated by Plugin */
	"ModPlug Plugin",				/* The description of the input plugin */

	init,							/* Called when the plugin is loaded */
	reset, 							/* Called when need to update config */
	load,							/* Load the file */
	load_rw,						/* Load the file from data */
	NULL,							/* Set song subtune */
	seek,							/* Seek to the specified time */
	NULL,							/* Set the equalizer, most plugins won't be able to do this */
	cleanup,						/* Called when gmgear exit */
	write_buffer, 					/* Fill audio buffer */
	get_song_title,					/* Function to grab the title string */

	get_packet,					/* Called when need main app data packet*/
	send_packet,				/* Send data packet to main app */

	NULL,						/* Filled by Main App */
	NULL,						/* Filled by Main App */
	NULL,						/* Filled by Main App */
	0,								/* Filled by Main App */
	LATIN,							/* Filled by Main App */
	CLIENT_PLAYER					/* Player type */
};
/** END of Plugin definitions setup */

/** Single Plugin Export */
InputPlugin *get_iplugin_info()
{
	return &itable;
}

void init(void)
{
	subtunes = 1;

	kbps = 0x6C;
	aformat = FMT_S16_LE;

	const char *token = ReadStringValue("timidity_cfg", itable.config);
	if( token ) strncpy(TIMIDITYCFG, token, 256);
	else strncpy(TIMIDITYCFG, "ms0:/.gmgear/timidity/timidity.cfg", 256);
	token = ReadStringValue("path_for_pat", itable.config);
	if( token ) strncpy(PATHFORPAT, token, 256);
	else strncpy(PATHFORPAT, "ms0:/.gmgear/timidity/", 256);

	itable.in_packet->ID = NO_PACKET;
}

void reset(void)
{
	channels = ReadIntValue("channels", itable.config);
	samplerate = ReadFreqValue("samplerate", itable.config);

	// Settings
	resampling_mode = ReadIntValue("resampling_mode", itable.config);
	oversampling    = ReadBoolValue("oversampling", itable.config);
	noise_reduction = ReadBoolValue("noise_reduction", itable.config);

	reverb		    = ReadBoolValue("reverb", itable.config);
	reverb_depth 	= ReadIntValue("reverb_depth", itable.config);
	reverb_delay	= ReadIntValue("reverb_delay", itable.config);

	megabass		= ReadBoolValue("megabass", itable.config);
	bass_amount	 	= ReadIntValue("bass_amount", itable.config);
	bass_range		= ReadIntValue("bass_range", itable.config);

	surround		= ReadBoolValue("surround", itable.config);
	surround_depth  = ReadIntValue("surround_depth", itable.config);
	surround_delay	= ReadIntValue("surround_delay", itable.config);

	ModPlug_GetSettings(&settings);
    settings.mResamplingMode = resampling_mode;
    settings.mChannels = channels;
    settings.mBits = 16;
    settings.mFrequency = samplerate;

	settings.mFlags = 0;
    if( oversampling ) settings.mFlags |= MODPLUG_ENABLE_OVERSAMPLING;
    if( noise_reduction ) settings.mFlags |= MODPLUG_ENABLE_NOISE_REDUCTION;
    if( reverb ) {
		settings.mFlags |= MODPLUG_ENABLE_REVERB;
		settings.mReverbDepth   = reverb_depth;
		settings.mReverbDelay   = reverb_delay;
	}
    if( megabass ) {
		settings.mFlags |= MODPLUG_ENABLE_MEGABASS;
		settings.mBassAmount    = bass_amount;
		settings.mBassRange     = bass_range;
	}
    if( surround ) {
		settings.mFlags |= MODPLUG_ENABLE_SURROUND;
		settings.mSurroundDepth = surround_depth;
		settings.mSurroundDelay = surround_delay;
	}


	strcpy(title, "");
	strcpy(album, "");
	strcpy(copyright, "");
	strcpy(artist, "");

	length = 0;

	if( mpfile ) {
		ModPlug_Unload( mpfile );
		mpfile = NULL;
	}

	itable.in_packet->ID = NO_PACKET;
}

int load( char *file )
{
	char *data = NULL;
	long size;

	data = getFileData( file, &size );
	if( !data ) return 0;

    mpfile = ModPlug_Load(data, size);
	if( mpfile == NULL ) {
		delete [] data;
        return 0;
    }

	// Delete data here?
	delete [] data;

	ModPlug_SetSettings(&settings);
	ModPlug_SetMasterVolume( mpfile, 512 );

	const char *name = ModPlug_GetName(mpfile);
	if( name ) strncpy(title, name, 128);
	strcpy(format, get_extension(file, true));

	length = ModPlug_GetLength(mpfile);
	if( length <= 0 ) length = 200000;
	playing = 1;

	return 1;
}

int load_rw( char *file, char *addr, int size )
{
	if (!(mpfile = ModPlug_Load(addr, size)))
    {
        fprintf(stderr,  "load - Couldn't load data\n");
        return 0;
    }

	ModPlug_SetMasterVolume( mpfile, 512 );

	const char *name = ModPlug_GetName(mpfile);
	strcpy(title, name ? name : get_basename(file));
	strcpy(format, get_extension(file, true));

	length = ModPlug_GetLength(mpfile);
	playing = 1;

	return 1;
}

long seek( long time )
{
	ModPlug_Seek(mpfile, time);
    return time;
}

void cleanup( void )
{
	if( mpfile ) {
		ModPlug_Unload( mpfile );
		mpfile = NULL;
	}

	__psp_free_heap();
}

int write_buffer( signed short *dest, int len )
{
	if(playing && mpfile)
    {
		if (settings.mChannels == 2)
			ModPlug_Read(mpfile, dest, len);
		else {
			ModPlug_Read(mpfile, dest, len>>1);
			for (int i = (len>>2)-1, j = (len>>1)-1; i >= 0; )
			{
				dest[j--] = dest[i];
				dest[j--] = dest[i--];
			}
		}

		return len;
	}

    return 0;
}

void get_song_title( char *filename, char **title, int *length )
{
}

void get_packet(void)
{
	switch( itable.in_packet->ID ) {
		case NO_PACKET:
		default:
			break;
	};
}

void send_packet(void)
{
	switch( itable.out_packet->ID ) {
		/* song data */
		case SONG_TITLE:
			itable.out_packet->data = title;
			break;
		case SONG_ARTIST:
			itable.out_packet->data = artist;
			break;
		case SONG_ALBUM_GAME:
			itable.out_packet->data = album;
			break;
		case SONG_COPYRIGHT:
			itable.out_packet->data = copyright;
			break;
		case SONG_FORMAT:
			itable.out_packet->data = format;
			break;
		case SONG_LENGTH:
			itable.out_packet->data = &length;
			break;
		case SONG_SUBTUNES:
			itable.out_packet->data = &subtunes;
			break;

		/* audio info */
		case AUDIO_CHANNELS:
			itable.out_packet->data = &channels;
			break;
		case AUDIO_SAMPLERATE:
			itable.out_packet->data = &samplerate;
			break;
		case AUDIO_FORMAT:
			itable.out_packet->data = &aformat;
			break;
		case AUDIO_KBPS:
			itable.out_packet->data = &kbps;
			break;

		/* extra data */
		case ATTACHED_PICTURE:
			itable.out_packet->size = 0;
			itable.out_packet->data = NULL;
			break;
		case NO_PACKET:
		default:
			itable.out_packet->size = 0;
			itable.out_packet->data = NULL;
			break;
	};
}


