
/*  Input Plugin for GameMusicGear ~
 *  Copyright (C) 2005 - 2010 gama, rockMax
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/* Default headers */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <pspkernel.h>
#include "../plugin.h"

#include "kssplay.h"

PSP_MODULE_INFO("IN_MSX", 0, 1, 1);
PSP_NO_CREATE_MAIN_THREAD();
PSP_HEAP_SIZE_KB( 4 * 1024 );

#ifdef __cplusplus
extern "C" {
#endif
int module_start(SceSize argc, void* argp)
{
 	return 0;
}

int module_stop(int args, void *argp)
{
	return 0;
}
#ifdef __cplusplus
}
#endif

/* Prototypes */
void init(void);
void reset(void);
int  load( char *filename );
int  load_rw( char *filename, char *addr, int size );
void set_tune( int tune );
long  seek( long time );
void cleanup( void );
int  write_buffer( signed short *dest, int len );
void get_song_title( char *filename, char **title, int *length );

void get_packet(void);
void send_packet(void);
/* End Prototypes */

#ifdef DEBUG
#define _log(fmt, args...) do { pspDebugScreenPrintf("%s:%d: %s: " fmt, __FILE__, __LINE__, __func__, ## args); } while(0); \
						   sceKernelDelayThread(2 * 1000000);
#else
#define _log(fmt, args...)
#endif

/* Variables */
KSSPLAY *kssplay = NULL;
KSS     *kss = NULL;

/* CONFIG */
int  MBM_MODE = 0;			// 0 - 2
int  MBM_ENABLE_STEREO = 0;  // 0 - 1
int  MBM_SYNC = 0;			// 0 - 1
char MBK_PATH[256];
char MBK_DUMMY[256];

int  OPLL_STEREO = 1;		// 0 - 1
int  CPU_SPEED   = 0;		// 0 - 16

int  PSG_HQ = 0;				// 0 - 1
int  SCC_HQ = 0;				// 0 - 1
int  OPLL_HQ = 0;			// 0 - 1

int  PSG_MODE = 0;			// 0 - 2
int  SCC_MODE = 0;			// 0 - 2
int  OPLL_MODE = 0;			// 0 - 2

char drivers_path[256];
int  play_time = 200;
int  global_volume = 80;

static int playing = 0;
static int channels = 2;
static int samplerate = 44100;
AFormat aformat = FMT_S16_LE;
int kbps = 128;

char title[128];
char artist[128];
char album[128];
char copyright[128];
char format[128];

long length;
int subtunes;
/* End Variables */

/** START of Plugin definitions setup */
InputPlugin itable =
{
	/* Populated by Plugin */
	"MSXPlug",						/* The description of the input plugin */

	init,							/* Called when the plugin is loaded */
	reset, 							/* Called when need to update config */
	load,							/* Load the file */
	load_rw,						/* Load the file from data */
	set_tune,						/* Set song subtune */
	seek,							/* Seek to the specified time */
	NULL,							/* Set the equalizer, most plugins won't be able to do this */
	cleanup,						/* Called when gmgear exit */
	write_buffer, 					/* Fill audio buffer */
	get_song_title,					/* Function to grab the title string */

	get_packet,					/* Called when need main app data packet*/
	send_packet,				/* Send data packet to main app */

	NULL,						/* Filled by Main App */
	NULL,						/* Filled by Main App */
	NULL,						/* Filled by Main App */
	0,								/* Filled by Main App */
	LATIN,							/* Filled by Main App */
	CLIENT_PLAYER					/* Player type */
};
/** END of Plugin definitions setup */

/** Single Plugin Export */
InputPlugin *get_iplugin_info()
{
	return &itable;
}

/* DRIVERS */
#define ILL_NAME "in_msx.ill"
typedef int (*pfnLoadDriver)(const char *filename);
static struct MsxDriver {
    const char *cszDriverName;
    pfnLoadDriver LoadDriver;
} MsxDrivers[] = {
    {"MGSDRV.COM",   KSS_load_mgsdrv},
    {"KINROU5.DRV",  KSS_load_kinrou},
    {"MPK.BIN",      KSS_load_mpk106},
    {"MPK103.BIN",   KSS_load_mpk103},
    {"OPX4KSS.BIN",  KSS_load_opxdrv},
    {"FMPAC.ROM",    KSS_load_fmbios},
    {"MBR143.001",   KSS_load_mbmdrv},
    {NULL, NULL}
};

static int illflag = 0;
static unsigned char illdata[16*19] ;
static void LoadDrivers(char *path)
{
    char szDriverPath[1024];

    struct MsxDriver *drv = MsxDrivers;
    while(drv->cszDriverName){
        strcpy(szDriverPath, path);
        strcat(szDriverPath, drv->cszDriverName);
        drv->LoadDriver(szDriverPath);
        drv++;
    }

    strcpy(szDriverPath, path) ;
    strcat(szDriverPath, ILL_NAME) ;

    FILE *fp = fopen(szDriverPath,"rb") ;
    if( fp != NULL )
    {
		fread(illdata, sizeof(illdata), 1, fp) ;
		illflag = 1 ;
		fclose(fp) ;
    }
}

void init(void)
{
	strcpy(format, "MSX");
	subtunes = 1;

	kbps = 0x6C;
	aformat = FMT_S16_LE;

#ifdef DEBUG
	pspDebugScreenInit();
#endif

	strncpy(drivers_path, ReadStringValue("drivers_path", itable.config), 256);
	LoadDrivers(drivers_path);

	itable.in_packet->ID = NO_PACKET;
}

void reset(void)
{
	// Read settings
	samplerate = ReadFreqValue("samplerate", itable.config);
	play_time  = ReadIntValue("play_time", itable.config);
	global_volume = ReadIntValue("global volume", itable.config);

	MBM_MODE   = ReadIntValue("MBM_MODE", itable.config);
	MBM_ENABLE_STEREO = ReadBoolValue("MBM_ENABLE_STEREO", itable.config);
	MBM_SYNC   = ReadIntValue("MBM_SYNC", itable.config);
	strncpy(MBK_PATH, ReadStringValue("MBK_PATH", itable.config), 256);
	strncpy(MBK_DUMMY, ReadStringValue("MBK_DUMMY", itable.config), 256);

	OPLL_STEREO = ReadBoolValue("OPLL_STEREO", itable.config);
	CPU_SPEED   = ReadIntValue("CPU_SPEED", itable.config);
	PSG_HQ      = ReadBoolValue("PSG_HQ", itable.config);
	SCC_HQ      = ReadBoolValue("SCC_HQ", itable.config);
	OPLL_HQ     = ReadBoolValue("OPLL_HQ", itable.config);

	PSG_MODE    = ReadIntValue("PSG_MODE", itable.config);
	SCC_MODE    = ReadIntValue("SCC_MODE", itable.config);
	OPLL_MODE   = ReadIntValue("OPLL_MODE", itable.config);

	strcpy(title, "");
	strcpy(album, "");
	strcpy(copyright, "");
	strcpy(artist, "");

	length = 0;
	playing = 0;

	// Cleanup
	if( kssplay ) {
		KSSPLAY_delete(kssplay);
		kssplay = NULL;
	}

	if( kss ) {
		KSS_delete(kss);
		kss = NULL;
	}

	itable.in_packet->ID = NO_PACKET;
}

int load( char *file )
{
	unsigned char *read_buf = NULL;
    int           file_length = 0;
    int           real_length = 0;

	FILE *fp;
	if( (fp = fopen(file,"rb")) == NULL ){
		return 0;
    }

    fseek(fp,0L,SEEK_END);
    file_length = ftell(fp);
    fseek(fp,0L,SEEK_SET);
    if((read_buf = (unsigned char *)malloc(file_length))==NULL){
        fclose(fp);
        return -1;
    }
    real_length = fread(read_buf, 1, file_length,fp);
    fclose(fp) ;

	if( KSS_check_type(read_buf, real_length, file) == MBMDATA )
    {
		KSS_set_mbmparam(MBM_MODE, MBM_ENABLE_STEREO, MBM_SYNC);
		KSS_autoload_mbk(file, MBK_PATH, MBK_DUMMY);
    }

	if( (kss = KSS_bin2kss(read_buf, real_length, file)) == NULL ){
        free(read_buf) ;
        return -2;
    }
    free(read_buf);

	/* Get info */
	length = play_time * 1000;
	subtunes = kss->trk_max;
	if( subtunes <= 0 ) subtunes = 1;
	strncpy(album, kss->idstr, 128);

	if( kss->info ) {
		length   = kss->info->time_in_ms;
		if( kss->info_num > 0 ) subtunes = kss->info_num;

		snprintf(title, 128, "%s", kss->title) ;
		if( kss->extra ) snprintf(artist, 128, "%s", kss->extra);
	}

	// Set numbeer of channels
	channels = kss->stereo||(OPLL_STEREO&&kss->fmpac)?2:1;

	kssplay = KSSPLAY_new(samplerate, channels, 16) ;
	if( !kssplay ) return -3;

	if( OPLL_STEREO ) kssplay->opll_stereo = 1; else kssplay->opll_stereo = 0;
	KSSPLAY_set_data(kssplay, kss);

	int disable_opll = 0;
	int disable_opl  = 0;
	int disable_psg  = 0;
	int disable_scc  = 0;
	int disable_filters = 1;

	// Disable devices and filters
	if (disable_opll)
		KSSPLAY_set_device_mute(kssplay, EDSC_OPLL, 1);

	if (disable_opl)
		KSSPLAY_set_device_mute(kssplay, EDSC_OPL, 1);

	if (disable_psg)
		KSSPLAY_set_device_mute(kssplay, EDSC_PSG, 1);

	if (disable_scc)
		KSSPLAY_set_device_mute(kssplay, EDSC_SCC, 1);

	if (disable_filters) {
		RCF_disable(kssplay->rcf[0]);
		RCF_disable(kssplay->rcf[1]);
		DCF_disable(kssplay->dcf[0]);
		DCF_disable(kssplay->dcf[1]);
	}

	/* CONFIG UPDATE */
	KSSPLAY_set_device_quality(kssplay, EDSC_PSG, PSG_HQ) ;
	KSSPLAY_set_device_quality(kssplay, EDSC_SCC, SCC_HQ) ;
	KSSPLAY_set_device_quality(kssplay, EDSC_OPLL, OPLL_HQ);

	KSSPLAY_set_device_type(kssplay, EDSC_PSG, PSG_MODE);
	KSSPLAY_set_device_type(kssplay, EDSC_SCC, SCC_MODE);
	KSSPLAY_set_device_type(kssplay, EDSC_OPLL, OPLL_MODE);

	KSSPLAY_set_master_volume(kssplay, global_volume);
	/* CONFIG END */

	KSSPLAY_reset(kssplay, kss->trk_min, CPU_SPEED) ;
	if(illflag) KSSPLAY_set_opll_patch(kssplay, illdata) ;

	playing = 1;
	return 1;
}

int load_rw( char *file, char *addr, int size )
{
	if( KSS_check_type(addr, size, file) == MBMDATA )
    {
		KSS_set_mbmparam(MBM_MODE, MBM_ENABLE_STEREO, MBM_SYNC);
		KSS_autoload_mbk(file, MBK_PATH, MBK_DUMMY);
    }

	if( (kss = KSS_bin2kss(addr, size, file)) == NULL ){
        return -2;
    }

	/* Print title strings */
	snprintf(title, 128, "[%s] %s", kss->idstr, kss->title) ;
	if( kss->extra ) snprintf(album, 128, "%s", kss->extra) ;

	if( kss->info ) {
		length = kss->info->time_in_ms;
	}
	else length = play_time * 1000;
	subtunes = kss->info_num;

	// Set numbeer of channels
	channels = kss->stereo||(OPLL_STEREO&&kss->fmpac)?2:1;

	kssplay = KSSPLAY_new(samplerate, channels, 16) ;

	if( OPLL_STEREO ) kssplay->opll_stereo = 1; else kssplay->opll_stereo = 0;
	KSSPLAY_set_data(kssplay, kss);

	int disable_opll = 1;
	int disable_opl  = 0;
	int disable_psg  = 0;
	int disable_scc  = 0;
	int disable_filters = 1;

	// Disable devices and filters
	if (disable_opll)
		KSSPLAY_set_device_mute(kssplay, EDSC_OPLL, 1);

	if (disable_opl)
		KSSPLAY_set_device_mute(kssplay, EDSC_OPL, 1);

	if (disable_psg)
		KSSPLAY_set_device_mute(kssplay, EDSC_PSG, 1);

	if (disable_scc)
		KSSPLAY_set_device_mute(kssplay, EDSC_SCC, 1);

	if (disable_filters) {
		RCF_disable(kssplay->rcf[0]);
		RCF_disable(kssplay->rcf[1]);
		DCF_disable(kssplay->dcf[0]);
		DCF_disable(kssplay->dcf[1]);
	}

	/* CONFIG UPDATE */
	KSSPLAY_set_device_quality(kssplay, EDSC_PSG, PSG_HQ) ;
	KSSPLAY_set_device_quality(kssplay, EDSC_SCC, SCC_HQ) ;
	KSSPLAY_set_device_quality(kssplay, EDSC_OPLL, OPLL_HQ);

	KSSPLAY_set_device_type(kssplay, EDSC_PSG, PSG_MODE);
	KSSPLAY_set_device_type(kssplay, EDSC_SCC, SCC_MODE);
	KSSPLAY_set_device_type(kssplay, EDSC_OPLL, OPLL_MODE);

	KSSPLAY_set_master_volume(kssplay, global_volume);
	/* CONFIG END */

	KSSPLAY_reset(kssplay, 0, CPU_SPEED) ;
	if(illflag) KSSPLAY_set_opll_patch(kssplay, illdata) ;

	playing = 1;
	return 1;
}

void set_tune( int tune )
{
	if( kssplay->kss->info_num > 0 ) {
		memcpy(title, kssplay->kss->info[tune%kssplay->kss->info_num].title, KSS_TITLE_MAX);
		length = kss->info[tune%kss->info_num].time_in_ms;
		if( length <= 0 ) length = play_time * 1000;
		KSSPLAY_reset(kssplay, kssplay->kss->info[tune%kssplay->kss->info_num].song, CPU_SPEED);
	}
	else KSSPLAY_reset(kssplay, tune, CPU_SPEED);

	if(illflag) KSSPLAY_set_opll_patch(kssplay, illdata) ;
}

/* TODO */
/* position to time and time to position */
#define POS2MS(x) ((int)((x)/samplerate*1000))
#define MS2POS(x) ((unsigned int)((x)/1000*samplerate))
long seek( long time )
{
	if( playing ) {
		// Seek function (time in msec)
		long samples = MS2POS(time);

		KSSPLAY_calc_silent(kssplay, samples/2/channels);
	}

    return 0;
}

void cleanup( void )
{
	// Cleanup
	reset();
	__psp_free_heap();
}

int write_buffer( signed short *dest, int len )
{
	if (playing)
    {
		KSSPLAY_calc(kssplay, (k_int16 *)dest, len/2/channels);
        return len;
    }

    return 0;
}



void get_song_title( char *filename, char **title, int *length )
{
}

void get_packet(void)
{
	switch( itable.in_packet->ID ) {
		case NO_PACKET:
		default:
			break;
	};
}

void send_packet(void)
{
	switch( itable.out_packet->ID ) {
		/* song data */
		case SONG_TITLE:
			itable.out_packet->data = title;
			break;
		case SONG_ARTIST:
			itable.out_packet->data = artist;
			break;
		case SONG_ALBUM_GAME:
			itable.out_packet->data = album;
			break;
		case SONG_COPYRIGHT:
			itable.out_packet->data = copyright;
			break;
		case SONG_FORMAT:
			itable.out_packet->data = format;
			break;
		case SONG_LENGTH:
			itable.out_packet->data = &length;
			break;
		case SONG_SUBTUNES:
			itable.out_packet->data = &subtunes;
			break;

		/* audio info */
		case AUDIO_CHANNELS:
			itable.out_packet->data = &channels;
			break;
		case AUDIO_SAMPLERATE:
			itable.out_packet->data = &samplerate;
			break;
		case AUDIO_FORMAT:
			itable.out_packet->data = &aformat;
			break;
		case AUDIO_KBPS:
			itable.out_packet->data = &kbps;
			break;

		/* extra data */
		case ATTACHED_PICTURE:
			itable.out_packet->size = 0;
			itable.out_packet->data = NULL;
			break;
		case NO_PACKET:
		default:
			itable.out_packet->size = 0;
			itable.out_packet->data = NULL;
			break;
	};
}
