/*
  ym2413_emu2413_wrapper.c

  made by Daisuke Nagano <breeze.nagano@nifty.com>
  Dec.14.2005
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "emu2413.h"

#include "ym2413.h"

/* class definition */
#define SAMP_BUFLEN (48000/60)
typedef struct _YM2413Wrapper {
  OPLL* opll;
  SAMP music[SAMP_BUFLEN];
  SAMP rhythm[SAMP_BUFLEN];
  int avail;

  int rate;
  int maxlen;
} YM2413Wrapper;

/* interfaces */
void
OPLL_set_quality (OPLL * opll, e_uint32 q)
{
  /* just ignore */
}

void
OPLL_set_pan (OPLL * opll, e_uint32 ch, e_uint32 pan)
{
  /* just ignore */
}

void
OPLL_setPatch (OPLL * opll, const e_uint8 * dump)
{
  /* just ignore ? */
}

e_uint32
OPLL_setMask (OPLL * opll, e_uint32 mask)
{
  /* just ignore ? */
  return mask;
}

void
OPLL_reset_patch (OPLL * opll, e_int32 type)
{
  /* just ignore */
}


OPLL *
OPLL_new (e_uint32 clk, e_uint32 rate)
{
  YM2413Wrapper* self = NULL;
  void* opll = NULL;

  self = (YM2413Wrapper *)malloc(sizeof(YM2413Wrapper));
  if (!self) {
    return NULL;
  }
  memset(self, 0, sizeof(YM2413Wrapper));
  
  opll = YM2413Init(clk, rate);
  if (!opll) {
    free(self);
    return NULL;
  }
  self->opll = opll;

  self->rate = rate;
  self->maxlen = rate / 60;
  self->avail = 0;
  return (OPLL *)self;
}

void
OPLL_delete (OPLL * opll)
{
  YM2413Wrapper* self = (void *)opll;
  if (self) {
    YM2413Shutdown(self->opll);
    free(self);
  }
}

void
OPLL_reset (OPLL * opll)
{
  YM2413Wrapper* self = (void *)opll;
  YM2413ResetChip(self->opll);
}

void
OPLL_writeIO (OPLL * opll, e_uint32 adr, e_uint32 val)
{
  YM2413Wrapper* self = (void *)opll;
  YM2413Write(self->opll, adr, val);
}


e_int16
OPLL_calc (OPLL * opll)
{
  YM2413Wrapper *self = (void *)opll;
  SAMP* pcmbuf[2];
  e_int16 v = 0;
  int ptr = self->maxlen - self->avail;

  if (self->avail>0) {
    v = self->music[ptr] + self->rhythm[ptr];
    self->avail--;
    return v;
  }

  pcmbuf[0] = self->music;
  pcmbuf[1] = self->rhythm;

  YM2413UpdateOne(self->opll, pcmbuf, self->maxlen);
  self->avail = self->maxlen-1;

  v = self->music[0] + self->rhythm[0];
  return v;
}

void
OPLL_calc_stereo (OPLL * opll, e_int32 out[2])
{
  e_int16 v = OPLL_calc(opll);

  out[0] = v;
  out[1] = v;
}
