
/*  Input Plugin for GameMusicGear ~
 *  Copyright (C) 2005 - 2010 gama, rockMax
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/* Default headers */

#include <pspkernel.h>
#include <pspkerneltypes.h>
#include <pspiofilemgr.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <malloc.h>

#include <ctype.h>

#include "../plugin.h"

#include <ivorbisfile.h>

PSP_MODULE_INFO("IN_OGG", 0, 1, 1);
PSP_NO_CREATE_MAIN_THREAD();
PSP_HEAP_SIZE_KB( 1024 * 4 );

#ifdef __cplusplus
extern "C" {
#endif
int module_start(SceSize argc, void* argp)
{
 	return 0;
}

int module_stop(int args, void *argp)
{
	return 0;
}
#ifdef __cplusplus
}
#endif

/* Prototypes */
void init(void);
void reset(void);
int load( char *filename );
long seek( long time );
void cleanup( void );
int write_buffer( signed short *dest, int len );
void get_song_title( char *filename, char **title, int *length );

void get_packet(void);
void send_packet(void);
/* End Prototypes */

/* Variables */
int current_section;
OggVorbis_File vf;
vorbis_info *vi;
SceUID fp = -1;

int playing = 0;

/* Required */
int channels = 2;
int samplerate = 44100;
AFormat aformat = FMT_S16_LE;
int kbps = 128;

char title[128];
char artist[128];
char album[128];
char copyright[128];
char format[128];

long length;
int subtunes;

/* End Variables */

/** START of Plugin definitions setup */
InputPlugin itable =
{
	/* Populated by Plugin */
	"Vorbis Ogg Plugin",				/* The description of the input plugin */

	init,							/* Called when the plugin is loaded */
	reset, 						/* Called when need to update config */
	load,							/* Load the file */
	NULL,							/* Load the file from data */
	NULL,							/* Set song subtune */
	seek,							/* Seek to the specified time */
	NULL,							/* Set the equalizer, most plugins won't be able to do this */
	cleanup,						/* Called when gmgear exit */
	write_buffer, 					/* Fill audio buffer */
	get_song_title,					/* Function to grab the title string */

	get_packet,					/* Called when need main app data packet*/
	send_packet,				/* Send data packet to main app */

	NULL,
	NULL,							/* Filled by Main App */
	NULL,							/* Filled by Main App */
	0,								/* Filled by Main App */
	LATIN,							/* Filled by Main App */
	CLIENT_PLAYER					/* Player type */
};
/** END of Plugin definitions setup */

/** Single Plugin Export */
InputPlugin *get_iplugin_info()
{
	return &itable;
}

void init(void)
{
	strcpy(format, "OGG");

	length = 0;
	subtunes = 1;

	aformat = FMT_S16_LE;

	itable.in_packet->ID = NO_PACKET;
}

void reset(void)
{
	strcpy(title, "");
	strcpy(album, "");
	strcpy(copyright, "");
	strcpy(artist, "");

	length = 0;
	kbps = 0;

	ov_clear(&vf);
	fp = -1;
	//SceIoClose(fp); // Already done in ov_clear. Not the way it should work i think

	playing = 0;
    current_section = 0;
}

int load( char *file )
{
#ifdef TREMORDEBUG
    fprintf(stderr, "vorbisplugin::init_file - Opening file\n");
#endif
	fp = sceIoOpen(file, PSP_O_RDONLY, 0777);
    if(ov_open(fp, &vf, NULL, 0) < 0)
    {
        fprintf(stderr, "vorbisplugin::init_file - Couldn't open file\n");

		// already closed?
		//sceIoClose(fp);
        return -1;
    }
#ifdef TREMORDEBUG
    fprintf(stderr, "vorbisplugin::init_file - File ok!\n");
#endif
    vi = ov_info(&vf,-1);
    channels = vi->channels;
    samplerate = vi->rate;
	kbps = vi->bitrate_nominal / 1000;
    //long samples = ov_pcm_total(&vf,-1);
    //length = (int)((1000.0f*samples) * channels
    //                            / (samplerate * 2));
	length = ov_time_total(&vf,-1);
#ifdef TREMORDEBUG
    fprintf(stderr, "vorbisplugin::init_file - %d chn, %d rate, %d len\n",
                    channels, samplerate, length);
#endif

    vorbis_comment *vc = ov_comment(&vf,-1);
#ifdef TREMORDEBUG
    fprintf(stderr, "vorbisplugin::init_file - Comments found: %d\n",
                    vc->comments);
#endif

    for (int i = 0; i < vc->comments; i++)
    {
        char* comment = vc->user_comments[i];
        if (!strncasecmp(comment,"title=", 6) && strlen(comment+6))
            strncpy(title, comment+6, 128);
        else if (!strncasecmp(comment,"artist=", 7) && strlen(comment+7))
            strncpy(artist, comment+7, 128);
        else if (!strncasecmp(comment,"album=", 6) && strlen(comment+6))
            strncpy(album, comment+6, 128);
        else if (!strncasecmp(comment,"copyright=",10) && strlen(comment+10))
            strncpy(copyright, comment+10, 128);
    }

	playing = 1;

	return 1;
}

long seek( long time )
{
	if( playing ) {
		int ogg_time = (int)ov_time_tell(&vf);
        int seek_time = time;
        if (seek_time < 0 && ogg_time < -seek_time)
            seek_time = 0;
        else
            seek_time = ogg_time + seek_time;
        if( ov_time_seek(&vf,seek_time) ) return 0;
        return (seek_time - ogg_time);
	}

    return 0;
}

void cleanup( void )
{
	reset();
	__psp_free_heap();
}

int write_buffer( signed short *dest, int len )
{
	if(playing)
    {
        long read = 0;
        for (int left = len; left > 0; left = len - read)
        {
            long read2 = ov_read(&vf,(char*)dest+read, left, &current_section);
            if (read2 > 0)
                read += read2;
            else if (read2 < 0)
                printf("vorbisplugin:fill_buffer - Error %d\n", read2);
            else
            {
                playing = 0;
                return read;
            }
        }

		return read;
    }

    return 0;
}

void get_song_title( char *filename, char **title, int *length )
{
}

void get_packet(void)
{
	switch( itable.in_packet->ID ) {
		case NO_PACKET:
		default:
			break;
	};
}

void send_packet(void)
{
	switch( itable.out_packet->ID ) {
		/* song data */
		case SONG_TITLE:
			itable.out_packet->data = title;
			break;
		case SONG_ARTIST:
			itable.out_packet->data = artist;
			break;
		case SONG_ALBUM_GAME:
			itable.out_packet->data = album;
			break;
		case SONG_COPYRIGHT:
			itable.out_packet->data = copyright;
			break;
		case SONG_FORMAT:
			itable.out_packet->data = format;
			break;
		case SONG_LENGTH:
			itable.out_packet->data = &length;
			break;
		case SONG_SUBTUNES:
			itable.out_packet->data = &subtunes;
			break;

		/* audio info */
		case AUDIO_CHANNELS:
			itable.out_packet->data = &channels;
			break;
		case AUDIO_SAMPLERATE:
			itable.out_packet->data = &samplerate;
			break;
		case AUDIO_FORMAT:
			itable.out_packet->data = &aformat;
			break;
		case AUDIO_KBPS:
			itable.out_packet->data = &kbps;
			break;

		/* extra data */
		case ATTACHED_PICTURE:
			itable.out_packet->size = 0;
			itable.out_packet->data = NULL;
			break;
		case NO_PACKET:
		default:
			itable.out_packet->size = 0;
			itable.out_packet->data = NULL;
			break;
	};
}

