
/*  Input Plugin for GameMusicGear ~
 *  Copyright (C) 2005 - 2010 gama, rockMax
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/* Default headers */

#include <pspkernel.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <malloc.h>

#include <ctype.h>

#include "../plugin.h"
#include "../SharedLib/libutils/utils.h"

#include "ao.h"
#include "eng_protos.h"

PSP_MODULE_INFO("IN_QSF", 0, 1, 1);
PSP_NO_CREATE_MAIN_THREAD();
PSP_HEAP_SIZE_KB(6 * 1024);

#ifdef __cplusplus
extern "C" {
#endif
int module_start(SceSize argc, void* argp)
{
 	return 0;
}

int module_stop(int args, void *argp)
{
	return 0;
}
#ifdef __cplusplus
}
#endif

/* Prototypes */
void init(void);
void reset(void);
int load( char *filename );
int load_rw( char *filename, char *addr, int size );
long seek( long time );
void cleanup( void );
int write_buffer( signed short *dest, int len );
void get_song_title( char *filename, char **title, int *length );

void get_packet(void);
void send_packet(void);
/* End Prototypes */

/* Variables */

char QSFName[256];

/* Required */
int channels = 2;
int samplerate = 44100;
AFormat aformat = FMT_S16_LE;
int kbps = 128;

char title[128];
char artist[128];
char album[128];
char copyright[128];
char format[128];

long length;
int subtunes;

/* redirect stubs to interface the Z80 core to the QSF engine */
uint8 memory_read(uint16 addr)
{
	return qsf_memory_read(addr);
}

uint8 memory_readop(uint16 addr)
{
	return memory_read(addr);
}

uint8 memory_readport(uint16 addr)
{
	return qsf_memory_readport(addr);
}

void memory_write(uint16 addr, uint8 byte)
{
	qsf_memory_write(addr, byte);
}

void memory_writeport(uint16 addr, uint8 byte)
{
	qsf_memory_writeport(addr, byte);
}

/* ao_get_lib: called to load secondary files */
int ao_get_lib(char *filename, SceUID *lib_fd)
{
	SceUID auxfile;
	auxfile = sceIoOpen(filename, PSP_O_RDONLY, 0777);
	//auxfile = fopen(filename, "rb");
	if (!auxfile)
	{
#ifdef DEBUG_LOADER
		printf("Unable to find auxiliary file %s\n", filename);
#endif
		return AO_FAIL;
	}

	*lib_fd = auxfile;

	return AO_SUCCESS;
}

/** START of Plugin definitions setup */
InputPlugin itable =
{
	"Capcom QSound (QSF) Plugin",	/* The description of the input plugin */

	init,							/* Called when the plugin is loaded */
	reset, 							/* Called when need to update config */
	load,							/* Load the file */
	NULL,							/* Load the file from data */
	NULL,							/* Set song subtune */
	NULL,							/* Seek to the specified time */
	NULL,							/* Set the equalizer, most plugins won't be able to do this */
	cleanup,						/* Called when gmgear exit */
	write_buffer, 					/* Fill audio buffer */
	get_song_title,					/* Function to grab the title string */

	get_packet,					/* Called when need main app data packet*/
	send_packet,				/* Send data packet to main app */

	NULL,
	NULL,							/* Filled by Main App */
	NULL,							/* Filled by Main App */
	0,								/* Filled by Main App */
	LATIN,							/* Filled by Main App */
	CLIENT_PLAYER					/* Player type */
};
/** END of Plugin definitions setup */

/** Single Plugin Export */
InputPlugin *get_iplugin_info()
{
	return &itable;
}

void init(void)
{
	strcpy(format, "QSF");
	subtunes = 1;
	kbps = 0xFF;

	itable.in_packet->ID = NO_PACKET;
}

void reset(void)
{
	// channels = ReadIntValue("channels", itable.config);

	strcpy(title, "");
	strcpy(album, "");
	strcpy(copyright, "");
	strcpy(artist, "");

	qsf_stop();

	length = 0;
	subtunes = 1;
}

int load( char *file )
{
	strncpy(QSFName, file, 256);

	itable.in_packet->data = qsf_get_qsamples();
    itable.in_packet->ID = CMD_REQUEST_FREE;
	return 1;
}

int ex_load(char *pQSamples)
{
	if( pQSamples == NULL ) {
		sceKernelExitGame();
	}

	qsf_set_qsamples(pQSamples);
	if (qsf_start(QSFName) != AO_SUCCESS) {
#ifdef DEBUG_LOADER
		printf("Error: Cannot load %s\n", QSFName);
#endif
		return -1;
	}

	length = 200000;
	ao_display_info info;
	if( qsf_fill_info(&info) == AO_SUCCESS ) {
		strncpy(title, info.info[1], 128);
		strncpy(album, info.info[2], 128);
		strncpy(copyright, info.info[4], 128);
		strncpy(artist, info.info[3], 128);
		length = psfTimeToMS(info.info[6]);
		if( length <= 0 ) length = 200000;
	}

	return 1;
}

#if 0
int load_rw( char *file, char *addr, int size )
{
	/* Dirty hack to load vgz from memory */
	if( is_ext(file, ".vgz") )
	{
		char tmpmod[256];
		sprintf(tmpmod, "_%s", file);
		FILE *fp = fopen(tmpmod, "wb");
		if( fp == NULL ) return -2;
		fwrite(addr, 1, size, fp);
		fclose(fp);
		int ret = load(tmpmod);
		remove(tmpmod);
		return ret;
	}

	if( (error = gme_open_data( addr, size, &emu, samplerate )) ) {
		return 0;
    }

	if( emu == NULL ) return -1;

	/* Enable most emulation accuracy */
	gme_enable_accuracy( emu, enable_accuracy );

	/* Add some stereo enhancement */
	gme_set_stereo_depth( emu, stereo_depth );

	/* Adjust equalizer for crisp, bassy sound */
	if( equalizer_effect )
	{
		gme_equalizer_t eq;
		eq.treble = treble;
		eq.bass   = bass;
		gme_set_equalizer( emu, &eq );
	}

	strcpy(filename, get_filename( file ));
    playing = 1;
    set_tune( 0 );
    return 1;
}

void set_tune( int tune )
{
	char *tmp = get_extension(filename, true);
	strncpy(format, tmp, 32);
	free(tmp);

	gme_start_track( emu, tune );
    if( !(error = gme_track_info( emu, &t_info, tune )) )
    {
		if( force_length > 0 ) length = force_length * 1000;
        else if (t_info->length > 0)
            length = t_info->length;
        else if (t_info->intro_length >= 0 && t_info->loop_length > 0)
            length = t_info->intro_length + t_info->loop_length * 2;
        else
            length = (2*60 + 30) * 1000; // 2:30

        if ( t_info->game )
        {
			strncpy( album, t_info->game, 32 );
        }

        if ( t_info->song )
        {
            strncpy( title, t_info->song, 32 );
        }
		/* else
        {
			char *base = get_base(filename);
            sprintf( title, "%s (%d)", base, tune+1 );
			free(base);
        }
		*/

        if ( t_info->author )
        {
            strncpy( artist, t_info->author, 32 );
        }

        if ( t_info->copyright )
        {
            strncpy( copyright, t_info->copyright, 32 );
        }

        subtunes = gme_track_count( emu );
		if( subtunes <= 0 ) subtunes = 1;
    }
}
#endif

void cleanup( void )
{
	qsf_stop();
	__psp_free_heap();
}

int write_buffer( signed short *dest, int len )
{
	signed short *pbuffer = dest;
	int tmplen = len>>1;
	while( tmplen > 0 ) {
		qsf_gen(pbuffer, 1024);
		pbuffer += 2048;
		tmplen -= 2048;
	}

	return len;
}

void get_song_title( char *filename, char **title, int *length )
{
}

void get_packet(void)
{
	switch( itable.in_packet->ID ) {
		case CMD_REQUEST_MALLOC:
			ex_load((char *)itable.in_packet->data);
			itable.in_packet->ID = NO_PACKET;
			break;
		case CMD_REQUEST_FREE:
			// TODO - Malloc only once?
			itable.in_packet->size = 8 * 1024 * 1024;
			itable.in_packet->ID = CMD_REQUEST_MALLOC;
			break;
		case NO_PACKET:
		default:
			break;
	};
}

void send_packet(void)
{
	switch( itable.out_packet->ID ) {
		/* song data */
		case SONG_TITLE:
			itable.out_packet->data = title;
			break;
		case SONG_ARTIST:
			itable.out_packet->data = artist;
			break;
		case SONG_ALBUM_GAME:
			itable.out_packet->data = album;
			break;
		case SONG_COPYRIGHT:
			itable.out_packet->data = copyright;
			break;
		case SONG_FORMAT:
			itable.out_packet->data = format;
			break;
		case SONG_LENGTH:
			itable.out_packet->data = &length;
			break;
		case SONG_SUBTUNES:
			itable.out_packet->data = &subtunes;
			break;

		/* audio info */
		case AUDIO_CHANNELS:
			itable.out_packet->data = &channels;
			break;
		case AUDIO_SAMPLERATE:
			itable.out_packet->data = &samplerate;
			break;
		case AUDIO_FORMAT:
			itable.out_packet->data = &aformat;
			break;
		case AUDIO_KBPS:
			itable.out_packet->data = &kbps;
			break;

		/* extra data */
		case ATTACHED_PICTURE:
			itable.out_packet->size = 0;
			itable.out_packet->data = NULL;
			break;
		case NO_PACKET:
		default:
			itable.out_packet->data = NULL;
			break;
	};
}
