
/*  Input Plugin for GameMusicGear ~
 *  Copyright (C) 2005 - 2010 gama, rockMax
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/* Default headers */

#include <pspkernel.h>
#include <pspdebug.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <malloc.h>

#include <ctype.h>

#include "../plugin.h"

#include "kmp_pi.h"

PSP_MODULE_INFO("IN_S98", 0, 1, 1);
PSP_NO_CREATE_MAIN_THREAD();
PSP_HEAP_SIZE_KB( 1024 * 4 );

#ifdef __cplusplus
extern "C" {
#endif
int module_start(SceSize argc, void* argp)
{
 	return 0;
}

int module_stop(int args, void *argp)
{
	return 0;
}
#ifdef __cplusplus
}
#endif

/* Prototypes */
void init(void);
void reset(void);
int load( char *filename );
int load_rw( char *filename, char *addr, int size );
long seek( long time );
void cleanup( void );
int write_buffer( signed short *dest, int len );
void get_song_title( char *filename, char **title, int *length );

void get_packet(void);
void send_packet(void);
/* End Prototypes */

/* Variables */
static void *s98 = NULL;
static SOUNDINFO info;

int playing = 0;
int loaded = 0;

/* Required */
int channels = 2;
int samplerate = 44100;
AFormat aformat = FMT_S16_LE;
int kbps = 128;

char title[128];
char artist[128];
char album[128];
char copyright[128];
char format[128];

long length;
int subtunes;

/* End Variables */

/** START of Plugin definitions setup */
InputPlugin itable =
{
	/* Populated by Plugin */
	"S98",						/* The description of the input plugin */

	init,							/* Called when the plugin is loaded */
	reset, 							/* Called when need to update config */
	load,							/* Load the file */
	load_rw,						/* Load the file from data */
	NULL,							/* Set song subtune */
	seek,							/* Seek to the specified time */
	NULL,							/* Set the equalizer, most plugins won't be able to do this */
	cleanup,						/* Called when gmgear exit */
	write_buffer, 					/* Fill audio buffer */
	get_song_title,					/* Function to grab the title string */

	get_packet,					/* Called when need main app data packet*/
	send_packet,				/* Send data packet to main app */

	NULL,						/* Filled by Main App */
	NULL,						/* Filled by Main App */
	NULL,						/* Filled by Main App */
	0,								/* Filled by Main App */
	LATIN,							/* Filled by Main App */
	CLIENT_PLAYER					/* Player type */
};
/** END of Plugin definitions setup */

/** Single Plugin Export */
InputPlugin *get_iplugin_info()
{
	return &itable;
}

// support functions
char *get_file_data(char *filename, int *size)
{
	SceUID mFile;
    char *data;
	size_t readbytes;

	mFile = sceIoOpen(filename, PSP_O_RDONLY, 0777);
	if( mFile < 0 ) return NULL;
    *size = sceIoLseek(mFile, 0, PSP_SEEK_END);
	sceIoLseek(mFile, 0, PSP_SEEK_SET);
    data = new char [*size];
	readbytes = sceIoRead(mFile, data, *size);

	if (mFile > 0)
		sceIoClose(mFile);

    return(data);
}


void init(void)
{
	strcpy(format, "S98");
	subtunes = 1;

	kbps = 0x6C;
	aformat = FMT_S16_LE;

	itable.in_packet->ID = NO_PACKET;
}

void reset(void)
{
	strcpy(title, "");
	strcpy(album, "");
	strcpy(copyright, "");
	strcpy(artist, "");

	length = 0;

	if( s98 ) {
		S98_Close(s98);
		s98 = NULL;
	}

	itable.in_packet->ID = NO_PACKET;
}

int load( char *file )
{
	// Initialize engine and load file
	int   size;
	char *data = get_file_data(file, &size);
	if( !data ) {
		return 0;
	}

	s98 = S98_OpenFromBuffer((BYTE *)data, size, &info);
	if( !s98 ) {
		delete [] data;
		return 0;
	}

	// CHECKME - can delete data?
	delete [] data;

	samplerate = info.dwSamplesPerSec;
	channels = info.dwChannels;
	kbps = info.dwBitsPerSample;
	length = info.dwLength;


	// Get audio info
	samplerate = info.dwSamplesPerSec;
	channels = info.dwChannels;
	length = info.dwLength;

	playing = 1;
	return 1;
}

int load_rw( char *file, char *addr, int size )
{
	s98 = S98_OpenFromBuffer((BYTE *)addr, size, &info);
	if( !s98 ) {
		return 0;
	}

	samplerate = info.dwSamplesPerSec;
	channels = info.dwChannels;
	kbps = info.dwBitsPerSample;
	length = info.dwLength;

	// Get audio info
	samplerate = info.dwSamplesPerSec;
	channels = info.dwChannels;
	length = info.dwLength;

	playing = 1;
	return 1;
}

long seek( long time )
{
	if( playing ) {
		return S98_SetPosition (s98, time);
	}

    return 0;
}

void cleanup( void )
{
	// Terminate engine and free memory
	if( s98 ) S98_Close(s98);

	__psp_free_heap();
}

int write_buffer( signed short *dest, int len )
{
	if(playing)
    {
        int writtenbytes = S98_Render(s98, (BYTE *)dest, len/2/channels);
        return len;
    }
    else
        memset(dest, 0, len);
    return len;
}

void get_song_title( char *filename, char **title, int *length )
{
}

void get_packet(void)
{
	switch( itable.in_packet->ID ) {
		case NO_PACKET:
		default:
			break;
	};
}

void send_packet(void)
{
	switch( itable.out_packet->ID ) {
		/* song data */
		case SONG_TITLE:
			itable.out_packet->data = title;
			break;
		case SONG_ARTIST:
			itable.out_packet->data = artist;
			break;
		case SONG_ALBUM_GAME:
			itable.out_packet->data = album;
			break;
		case SONG_COPYRIGHT:
			itable.out_packet->data = copyright;
			break;
		case SONG_FORMAT:
			itable.out_packet->data = format;
			break;
		case SONG_LENGTH:
			itable.out_packet->data = &length;
			break;
		case SONG_SUBTUNES:
			itable.out_packet->data = &subtunes;
			break;

		/* audio info */
		case AUDIO_CHANNELS:
			itable.out_packet->data = &channels;
			break;
		case AUDIO_SAMPLERATE:
			itable.out_packet->data = &samplerate;
			break;
		case AUDIO_FORMAT:
			itable.out_packet->data = &aformat;
			break;
		case AUDIO_KBPS:
			itable.out_packet->data = &kbps;
			break;

		/* extra data */
		case ATTACHED_PICTURE:
			itable.out_packet->size = 0;
			itable.out_packet->data = NULL;
			break;
		case NO_PACKET:
		default:
			itable.out_packet->size = 0;
			itable.out_packet->data = NULL;
			break;
	};
}


