//	$Id: file.cpp,v 1.6 1999/12/28 11:14:05 cisc Exp $

#include "headers.h"
#include "File.h"
#ifdef __amigaos4__
#include <proto/dos.h>
#endif

// ---------------------------------------------------------------------------
//	\z/
// ---------------------------------------------------------------------------

FileIO::FileIO()
{
	flags = 0;
}

FileIO::FileIO(const char* filename, uint flg)
{
	flags = 0;
	Open(filename, flg);
}

FileIO::~FileIO()
{
	Close();
}

// ---------------------------------------------------------------------------
//	t@CJ
// ---------------------------------------------------------------------------

bool FileIO::Open(const char* filename, uint flg)
{
	Close();

	strncpy(path, filename, MAX_PATH);

	int flags = 0;
	flags |= flg & readonly ? PSP_O_RDONLY : PSP_O_RDWR;
	if( flg & create ) flags |= PSP_O_CREAT;
	hfile = sceIoOpen(filename, flags, 0777);
	
	flags = (flg & readonly) | (hfile < 0 ? 0 : open);
	if (!(flags & open))
	{
		error = file_not_found;
	}

	SetLogicalOrigin(0);
	return !!(flags & open);
}

// ---------------------------------------------------------------------------
//	t@CȂꍇ͍쐬
// ---------------------------------------------------------------------------

bool FileIO::CreateNew(const char* filename)
{
	Close();

	strncpy(path, filename, MAX_PATH);
	
	int flags = PSP_O_CREAT | PSP_O_RDWR;
	hfile = sceIoOpen(filename, flags, 0777);
	
	flags = (hfile < 0 ? 0 : open);
	
	SetLogicalOrigin(0);
	return !!(flags & open);
}

// ---------------------------------------------------------------------------
//	t@C蒼
// ---------------------------------------------------------------------------

bool FileIO::Reopen(uint flg)
{
	if (!(flags & open)) return false;
	if ((flags & readonly) && (flg & create)) return false;

	if (flags & readonly) flg |= readonly;

	Close();

	int flags = 0;
	flags |= flg & readonly ? PSP_O_RDONLY : PSP_O_RDWR;
	if( flg & create ) flags |= PSP_O_CREAT;
	hfile = sceIoOpen(path, flags, 0777);
	
	flags = (flg & readonly) | (hfile < 0 ? 0 : open);

	SetLogicalOrigin(0);
	return !!(flags & open);
}

// ---------------------------------------------------------------------------
//	t@C
// ---------------------------------------------------------------------------

void FileIO::Close()
{
	if (GetFlags() & open)
	{
		sceIoClose(hfile);
		flags = 0;
	}
}

// ---------------------------------------------------------------------------
//	t@Ck̓ǂݏo
// ---------------------------------------------------------------------------

int32 FileIO::Read(void* dest, int32 size)
{
	if (!(GetFlags() & open))
		return -1;
	
	return sceIoRead(hfile, dest, size);
}

// ---------------------------------------------------------------------------
//	t@Cւ̏o
// ---------------------------------------------------------------------------

int32 FileIO::Write(const void* dest, int32 size)
{
	if (!(GetFlags() & open) || (GetFlags() & readonly))
		return -1;
	
	return sceIoWrite(hfile, dest, size);
}

// ---------------------------------------------------------------------------
//	t@CV[N
// ---------------------------------------------------------------------------

bool FileIO::Seek(int32 pos, SeekMethod method)
{
	if (!(GetFlags() & open))
		return false;

	SceOff mode;
	switch(method) {
		case begin:
			mode = PSP_SEEK_SET;	pos += lorigin;
			break;
		case current:
			mode = PSP_SEEK_CUR;
			break;
		case end:
			mode = PSP_SEEK_END;
			break;
		default:
			return false;
	}
	return 0xffffffff != sceIoLseek(hfile, pos, mode);
}

// ---------------------------------------------------------------------------
//	t@C̈ʒu𓾂
// ---------------------------------------------------------------------------

int32 FileIO::Tellp()
{
	if (!(GetFlags() & open))
		return 0;

	return sceIoLseek(hfile, 0, PSP_SEEK_CUR);
}

// ---------------------------------------------------------------------------
//	݂̈ʒut@C̏I[Ƃ
// ---------------------------------------------------------------------------

bool FileIO::SetEndOfFile()
{
	if (!(GetFlags() & open))
		return false;
	
	return sceIoLseek(hfile, 0, PSP_SEEK_END) != 0;
}
