
/*  Input Plugin for GameMusicGear ~
 *  Copyright (C) 2005 - 2010 gama, rockMax
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/* Default headers */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/stat.h>

#include <pspkernel.h>
#include "../plugin.h"

#include "asap.h"

PSP_MODULE_INFO("IN_SAP", 0, 1, 1);
PSP_NO_CREATE_MAIN_THREAD();
PSP_HEAP_SIZE_KB( 2 * 1024 );

#ifdef __cplusplus
extern "C" {
#endif
int module_start(SceSize argc, void* argp)
{
 	return 0;
}

int module_stop(int args, void *argp)
{
	return 0;
}
#ifdef __cplusplus
}
#endif

/* Prototypes */
void init(void);
void reset(void);
int  load( char *filename );
int  load_rw( char *filename, char *addr, int size );
void set_tune( int tune );
long seek( long time );
void cleanup( void );
int  write_buffer( signed short *dest, int len );
void get_song_title( char *filename, char **title, int *length );

void get_packet(void);
void send_packet(void);
/* End Prototypes */

/* Variables */
ASAP_State asap;
byte module[ASAP_MODULE_MAX];
int  module_len;
int  play_time;
int  detect_silence;
int	 silence_timeout;

int playing = 0;
int channels = 1;
int samplerate = 44100;
AFormat aformat = FMT_S16_LE;
int kbps = 128;

char title[128];
char artist[128];
char album[128];
char copyright[128];
char format[128];

long length;
int subtunes;
/* End Variables */

/** START of Plugin definitions setup */
InputPlugin itable =
{
	/* Populated by Plugin */
	"ASAP 1.2.0",							 /* The description of the input plugin */

	init,							/* Called when the plugin is loaded */
	reset, 							/* Called when need to update config */
	load,							/* Load the file */
	load_rw,						/* Load the file from data */
	set_tune,						/* Set song subtune */
	seek,							/* Seek to the specified time */
	NULL,							/* Set the equalizer, most plugins won't be able to do this */
	cleanup,						/* Called when gmgear exit */
	write_buffer, 					/* Fill audio buffer */
	get_song_title,					/* Function to grab the title string */

	get_packet,					/* Called when need main app data packet*/
	send_packet,				/* Send data packet to main app */

	NULL,						/* Filled by Main App */
	NULL,						/* Filled by Main App */
	NULL,						/* Filled by Main App */
	0,								/* Filled by Main App */
	LATIN,							/* Filled by Main App */
	CLIENT_PLAYER					/* Player type */
};
/** END of Plugin definitions setup */

/** Single Plugin Export */
InputPlugin *get_iplugin_info()
{
	return &itable;
}

/* EXTRA FUNCTIONS */
static char *asap_stpcpy(char *dest, const char *src)
{
	size_t len = strlen(src);
	memcpy(dest, src, len);
	return dest + len;
}

#define USE_STDIO
static int asap_load_file(const char *filename)
{
#ifdef USE_STDIO
	FILE *fp;
	fp = fopen(filename, "rb");
	if (fp == NULL)
		return FALSE;
	module_len = (int) fread(module, 1, sizeof(module), fp);
	fclose(fp);
#else
	int fd;
	fd = open(filename, O_RDONLY);
	if (fd == -1)
		return FALSE;
	module_len = read(fd, module, sizeof(module));
	close(fd);
	if (module_len < 0)
		return FALSE;
#endif
	return TRUE;
}
/* END EXTRA FUNCTIONS */

void init(void)
{
	strcpy(format, "SAP");
	subtunes = 1;

	detect_silence  = 1;
	silence_timeout = 5;
	play_time       = 300;

	kbps    = 0x6C;
	aformat = FMT_S16_LE;

	itable.in_packet->ID = NO_PACKET;
}

void reset(void)
{
	play_time = ReadIntValue("play_time", itable.config);
	detect_silence = ReadBoolValue("detect_silence", itable.config);
	if( detect_silence )
		silence_timeout = ReadIntValue("silence_timeout", itable.config);

	strcpy(title, "");
	strcpy(album, "");
	strcpy(copyright, "");
	strcpy(artist, "");

	length = 0;
	playing = 0;

	// Cleanup

	itable.in_packet->ID = NO_PACKET;
}

int load( char *file )
{
	if (!asap_load_file(file))
		return -1;

	if (!ASAP_Load(&asap, file, module, module_len))
		return -2;

	if( detect_silence )
		ASAP_DetectSilence(&asap, silence_timeout);

	length = asap.module_info.durations[0];
	if( length <= 0 ) length = play_time * 1000;

    channels   = asap.module_info.channels;
    samplerate = ASAP_SAMPLE_RATE;

	if( asap.module_info.name[0] )
		strncpy(title, asap.module_info.name, 128);
	if( asap.module_info.author[0] )
		strncpy(artist, asap.module_info.author, 128);
	if( asap.module_info.date[0] )
		strncpy(copyright, asap.module_info.date, 128);

	subtunes = asap.module_info.songs;
	if( subtunes <= 0 ) subtunes = 1;

	ASAP_PlaySong(&asap, 0, length);

	playing = 1;
	return 1;
}

int load_rw( char *file, char *addr, int size )
{
	if (!ASAP_Load(&asap, file, addr, size))
		return -2;

	if( detect_silence )
		ASAP_DetectSilence(&asap, silence_timeout);

	length = asap.module_info.durations[0];
	if( length <= 0 ) length = play_time * 1000;

    channels   = asap.module_info.channels;
    samplerate = ASAP_SAMPLE_RATE;

	if( asap.module_info.name[0] )
		strncpy(title, asap.module_info.name, 128);
	if( asap.module_info.author[0] )
		strncpy(artist, asap.module_info.author, 128);
	if( asap.module_info.date[0] )
		strncpy(copyright, asap.module_info.date, 128);

	subtunes = asap.module_info.songs;
	if( subtunes <= 0 ) subtunes = 1;

	ASAP_PlaySong(&asap, 0, length);

	playing = 1;
	return 1;
}

long seek( long time )
{
	if( playing ) {
		ASAP_Seek(&asap, time);
		return time;
	}

    return 0;
}

void set_tune( int tune )
{
	length = asap.module_info.durations[tune];
	if( length <= 0 ) length = play_time * 1000;

	if( detect_silence )
		ASAP_DetectSilence(&asap, silence_timeout);

	ASAP_PlaySong(&asap, tune, length);
}

void cleanup( void )
{
	// Cleanup
	reset();
	__psp_free_heap();
}

int write_buffer( signed short *dest, int len )
{
	if (playing)
    {
		int ret = ASAP_Generate(&asap, dest, len, ASAP_FORMAT_S16_LE);
		if( ret <= 0 ) playing = 0;
		return ret;
    }

    return 0;
}

void get_song_title( char *filename, char **title, int *length )
{
}

void get_packet(void)
{
	switch( itable.in_packet->ID ) {
		case NO_PACKET:
		default:
			break;
	};
}

void send_packet(void)
{
	switch( itable.out_packet->ID ) {
		/* song data */
		case SONG_TITLE:
			itable.out_packet->data = title;
			break;
		case SONG_ARTIST:
			itable.out_packet->data = artist;
			break;
		case SONG_ALBUM_GAME:
			itable.out_packet->data = album;
			break;
		case SONG_COPYRIGHT:
			itable.out_packet->data = copyright;
			break;
		case SONG_FORMAT:
			itable.out_packet->data = format;
			break;
		case SONG_LENGTH:
			itable.out_packet->data = &length;
			break;
		case SONG_SUBTUNES:
			itable.out_packet->data = &subtunes;
			break;

		/* audio info */
		case AUDIO_CHANNELS:
			itable.out_packet->data = &channels;
			break;
		case AUDIO_SAMPLERATE:
			itable.out_packet->data = &samplerate;
			break;
		case AUDIO_FORMAT:
			itable.out_packet->data = &aformat;
			break;
		case AUDIO_KBPS:
			itable.out_packet->data = &kbps;
			break;

		/* extra data */
		case ATTACHED_PICTURE:
			itable.out_packet->size = 0;
			itable.out_packet->data = NULL;
			break;
		case NO_PACKET:
		default:
			itable.out_packet->size = 0;
			itable.out_packet->data = NULL;
			break;
	};
}
