
/*  Input Plugin for GameMusicGear ~
 *  Copyright (C) 2005 - 2010 gama, rockMax
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/* Default headers */

#include <pspkernel.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <malloc.h>

#include <ctype.h>

#include "../plugin.h"

/* sc68 includes */
extern "C" {
#include "api68/api68.h"
}

#define printf pspDebugScreenPrintf

PSP_MODULE_INFO("IN_SC68", 0, 1, 1);
PSP_NO_CREATE_MAIN_THREAD();
PSP_HEAP_SIZE_KB( 1024 * 4 );

#ifdef __cplusplus
extern "C" {
#endif
int module_start(SceSize argc, void* argp)
{
 	return 0;
}

int module_stop(int args, void *argp)
{
	return 0;
}
#ifdef __cplusplus
}
#endif

/* Prototypes */
void init(void);
void reset(void);
int load( char *filename );
int load_rw( char *file, char *addr, int size );
void set_tune(int tune);
long seek( long time );
void cleanup( void );
int write_buffer( signed short *dest, int len );
void get_song_title( char *filename, char **title, int *length );
void set_tune(int tune);

void get_packet(void);
void send_packet(void);
/* End Prototypes */

/* Variables */
api68_t *sc68 = NULL;

int playing = 0;
int loaded = 0;

/* Required */
int channels = 2;
int samplerate = 44100;
AFormat aformat = FMT_S16_LE;
int kbps = 128;

char title[128];
char artist[128];
char album[128];
char copyright[128];
char format[128];

long length;
int subtunes;

/* End Variables */

/** START of Plugin definitions setup */
InputPlugin itable =
{
	/* Populated by Plugin */
	"SC68 Plugin",				/* The description of the input plugin */

	init,							/* Called when the plugin is loaded */
	reset, 							/* Called when need to update config */
	load,							/* Load the file */
	load_rw,						/* Load the file from data */
	set_tune,						/* Set song subtune */
	seek,							/* Seek to the specified time */
	NULL,							/* Set the equalizer, most plugins won't be able to do this */
	cleanup,						/* Called when gmgear exit */
	write_buffer, 					/* Fill audio buffer */
	get_song_title,					/* Function to grab the title string */

	get_packet,					/* Called when need main app data packet*/
	send_packet,				/* Send data packet to main app */

	NULL,
	NULL,							/* Filled by Main App */
	NULL,							/* Filled by Main App */
	0,								/* Filled by Main App */
	LATIN,							/* Filled by Main App */
	CLIENT_PLAYER					/* Player type */
};
/** END of Plugin definitions setup */

/** Single Plugin Export */
InputPlugin *get_iplugin_info()
{
	return &itable;
}

void init(void)
{
	strcpy(format, "SC68");

	channels = 2;
	kbps = 0x6C;
	aformat = FMT_S16_LE;

	itable.in_packet->ID = NO_PACKET;
}

void reset(void)
{
	samplerate = ReadFreqValue("samplerate", itable.config);

	strcpy(title, "");
	strcpy(album, "");
	strcpy(copyright, "");
	strcpy(artist, "");

	subtunes = 1;

	api68_close(sc68);
	playing = 0;
}

// SC68 ALLOC FUNCTIONS
void *myAlloc(unsigned int n)
{
	unsigned int *data = new unsigned int [n];
	return data;
}

void myFree(void * data)
{
	delete(data);
}

void get_sc68_info(int track)
{
	api68_music_info_t info;
	if( !api68_music_info(sc68, &info, track, 0) ) {
		if( info.title ) strncpy(title, info.title, 128);
		if( info.composer ) strncpy(artist, info.composer, 128);
		if( info.author ) strncpy(copyright, info.author, 128);

		subtunes = info.tracks;
		length = info.time_ms ? info.time_ms : 200000;
	}
	else {
		subtunes = 1;
		length = 120000;
	}
}

int load( char *file )
{
	if( !loaded ) {
		api68_init_t init68;
		memset(&init68, 0, sizeof(init68));
		init68.alloc = myAlloc;
		init68.free = myFree;
		init68.user_path = ReadStringValue("user_path", itable.config);

		sc68 = api68_init(&init68);
		if( !sc68 ) goto error;
	}

	if (api68_verify_file(file) < 0) goto error;
	if (api68_load_file(sc68, file)) goto error;
	if (api68_play(sc68, 1)) goto error;

	samplerate = api68_sampling_rate(sc68, samplerate);
	get_sc68_info(0);

	playing = 1;
	loaded = 1;

	return 1;

error:
	api68_shutdown(sc68);
	return 0;
}

int load_rw( char *file, char *addr, int size )
{
	reset();

	if( !loaded ) {
		api68_init_t init68;
		memset(&init68, 0, sizeof(init68));
		init68.alloc = myAlloc;
		init68.free = myFree;
		init68.user_path = ReadStringValue("user_path", itable.config);

		sc68 = api68_init(&init68);
		if( !sc68 ) goto error;
	}

	if (api68_verify_mem(addr, size) < 0) goto error;
	if (api68_load_mem(sc68, addr, size)) goto error;
	if (api68_play(sc68, 1)) goto error;

	samplerate = api68_sampling_rate(sc68, samplerate);
	get_sc68_info(0);

	playing = 1;
	loaded = 1;

	return 1;

error:
	api68_shutdown(sc68);
	return 0;
}

long seek( long time )
{
	if( playing ) {
		return api68_seek(sc68, time);
	}

    return 0;
}

void cleanup( void )
{
	api68_shutdown(sc68);
	__psp_free_heap();
}

int write_buffer( signed short *dest, int len )
{
	if(playing)
    {
		unsigned int code = api68_process(sc68, dest, len >> 2);

		if (code == API68_MIX_ERROR) {
			playing = 0;
		}

		if ( (code & API68_END) ) {
			playing = 0;
		}

		return len;
    }
    else
        memset(dest, 0, len);
    return len;
}

void get_song_title( char *filename, char **title, int *length )
{
}

void set_tune(int tune)
{
	api68_stop(sc68);
	api68_play(sc68, tune);
	get_sc68_info(tune);
}

void get_packet(void)
{
	switch( itable.in_packet->ID ) {
		case NO_PACKET:
		default:
			break;
	};
}

void send_packet(void)
{
	switch( itable.out_packet->ID ) {
		/* song data */
		case SONG_TITLE:
			itable.out_packet->data = title;
			break;
		case SONG_ARTIST:
			itable.out_packet->data = artist;
			break;
		case SONG_ALBUM_GAME:
			itable.out_packet->data = album;
			break;
		case SONG_COPYRIGHT:
			itable.out_packet->data = copyright;
			break;
		case SONG_FORMAT:
			itable.out_packet->data = format;
			break;
		case SONG_LENGTH:
			itable.out_packet->data = &length;
			break;
		case SONG_SUBTUNES:
			itable.out_packet->data = &subtunes;
			break;

		/* audio info */
		case AUDIO_CHANNELS:
			itable.out_packet->data = &channels;
			break;
		case AUDIO_SAMPLERATE:
			itable.out_packet->data = &samplerate;
			break;
		case AUDIO_FORMAT:
			itable.out_packet->data = &aformat;
			break;
		case AUDIO_KBPS:
			itable.out_packet->data = &kbps;
			break;

		/* extra data */
		case ATTACHED_PICTURE:
			itable.out_packet->size = 0;
			itable.out_packet->data = NULL;
			break;
		case NO_PACKET:
		default:
			itable.out_packet->size = 0;
			itable.out_packet->data = NULL;
			break;
	};
}

