
/*  Input Plugin for GameMusicGear ~
 *  Copyright (C) 2005 - 2010 gama, rockMax
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/* Default headers */

#include <pspkernel.h>
#include <pspdebug.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <malloc.h>
#include <sys/stat.h>

#include <player.h>
#include <fformat.h>
#include <myendian.h>

#include "xs_length.h"

#include "../plugin.h"

// LINKER PATCH
extern "C" {
void *__dso_handle = NULL;
}

// Filters defaults
#define XS_SIDPLAY1_FS        (400.0f)
#define XS_SIDPLAY1_FM        (60.0f)
#define XS_SIDPLAY1_FT        (0.05f)

PSP_MODULE_INFO("IN_SID", 0, 1, 1);
PSP_NO_CREATE_MAIN_THREAD();
PSP_HEAP_SIZE_KB( 1024 * 4 );

#ifdef __cplusplus
extern "C" {
#endif
int module_start(SceSize argc, void* argp)
{
 	return 0;
}

int module_stop(int args, void *argp)
{
	return 0;
}
#ifdef __cplusplus
}
#endif

/* Prototypes */
void init(void);
void reset(void);
int  load( char *filename );
int  load_rw( char *filename, char *addr, int size );
void set_tune( int tune );
// CHECKME - int seek( int time );
void cleanup( void );
int  write_buffer( signed short *dest, int len );
void get_song_title( char *filename, char **title, int *length );

void get_packet(void);
void send_packet(void);
/* End Prototypes */

/* Variables */

#define SID_DOUBLE_MONO

class SidState
{
    public:
        SidState(emuEngine *e = NULL, sidTune *t = NULL)
        {
            engine = e;
            tune = t;
            playing = false;
        }

        emuEngine *engine;
        sidTune *tune;
        bool playing;
        int song;
};

static SidState *state = NULL;
static xs_sldb_t *xs_sldb_db = NULL;
sldb_node_t *xs_lengths = NULL;
static char sldb_path[256];

int playing  = 0;
int loaded   = 0;
int db_found = 0;

/* Required */
int channels = 2;
int samplerate = 44100;
AFormat aformat = FMT_S16_LE;
int kbps = 128;

char title[128];
char artist[128];
char album[128];
char copyright[128];
char format[128];

long length;
int subtunes;

/* End Variables */

/** START of Plugin definitions setup */
InputPlugin itable =
{
	/* Populated by Plugin */
	"Comodore 64 (sidplay1)",						/* The description of the input plugin */

	init,							/* Called when the plugin is loaded */
	reset, 							/* Called when need to update config */
	load,							/* Load the file */
	load_rw,						/* Load the file from data */
	set_tune,						/* Set song subtune */
	NULL,							/* Seek to the specified time */
	NULL,							/* Set the equalizer, most plugins won't be able to do this */
	cleanup,						/* Called when gmgear exit */
	write_buffer, 					/* Fill audio buffer */
	get_song_title,					/* Function to grab the title string */

	get_packet,					/* Called when need main app data packet*/
	send_packet,				/* Send data packet to main app */

	NULL,						/* Filled by Main App */
	NULL,						/* Filled by Main App */
	NULL,						/* Filled by Main App */
	0,								/* Filled by Main App */
	LATIN,							/* Filled by Main App */
	CLIENT_PLAYER					/* Player type */
};
/** END of Plugin definitions setup */

/** Single Plugin Export */
InputPlugin *get_iplugin_info()
{
	return &itable;
}

// Init sidplay engine
static void init_engine()
{
	if( !state ) {
		state = new SidState;
		if( !state ) sceKernelExitGame();
	}

    SidState *s = state;

    s->playing = false;

    if(!s->engine)
    {
        s->engine = new emuEngine();

        if(!s->engine->verifyEndianess())
        {
            sceKernelExitGame();
        }

        // Get the default configuration.
        struct emuConfig myEmuConfig;
        s->engine->getConfig(myEmuConfig);
        myEmuConfig.frequency = samplerate;
        myEmuConfig.channels = channels == 2 ? SIDEMU_STEREO : SIDEMU_MONO;
		if( channels == 2 ) {
			myEmuConfig.autoPanning = SIDEMU_CENTEREDAUTOPANNING;
			myEmuConfig.volumeControl = SIDEMU_FULLPANNING;
		}

        myEmuConfig.bitsPerSample = SIDEMU_16BIT;
        myEmuConfig.sampleFormat = SIDEMU_SIGNED_PCM;

		// filters
		myEmuConfig.mos8580 = 0;
		myEmuConfig.emulateFilter = 1;
		myEmuConfig.filterFs = XS_SIDPLAY1_FS;
		myEmuConfig.filterFm = XS_SIDPLAY1_FM;
		myEmuConfig.filterFt = XS_SIDPLAY1_FT;

		s->engine->setConfig(myEmuConfig);
    }
}

static int get_length(sldb_node_t *tmpLength, int subtune)
{
	int result;
	if (tmpLength && (subtune < tmpLength->nlengths))
		result = tmpLength->lengths[subtune] * 1000;
    else result = 200000;

	return result;
}

void init(void)
{
	strcpy(format, "SID");
	subtunes = 1;

	kbps = 0x6C;
	aformat = FMT_S16_LE;

	itable.in_packet->ID = NO_PACKET;

	const char *token = ReadStringValue("sldb_path", itable.config);
	if( !token ) strncpy(sldb_path, "ms0:/.gmgear/Songlengths.txt", 256);
	else strncpy(sldb_path, token, 256);

	db_found = 0;

	// Check for database to exist
	struct stat st;
	memset(&st, 0, sizeof(struct stat));
	if( stat(sldb_path, &st) == 0 && st.st_size > 0 ) {
		/* Allocate database */
		xs_sldb_db = (xs_sldb_t *) malloc(sizeof(xs_sldb_t));
		if (!xs_sldb_db) {
			return;
		}

		memset(xs_sldb_db, 0, sizeof(xs_sldb_t));

		/* Read the database */
		if (xs_sldb_read(xs_sldb_db, sldb_path) != 0) {
			xs_sldb_free(xs_sldb_db);
			xs_sldb_db = NULL;
			return;
		}

		/* Create index */
		if (xs_sldb_index(xs_sldb_db) != 0) {
			xs_sldb_free(xs_sldb_db);
			xs_sldb_db = NULL;
			return;
		}
			db_found = 1;
	}
}

void reset(void)
{
	channels = ReadIntValue("channels", itable.config);
	samplerate = ReadFreqValue("samplerate", itable.config);

	strcpy(title, "");
	strcpy(album, "");
	strcpy(copyright, "");
	strcpy(artist, "");

	length = 0;

	itable.in_packet->ID = NO_PACKET;
}

int load( char *file )
{
	// Initialize the SID-Emulator Engine to defaults.
    init_engine();
    SidState *s = state;

    if(file)
    {
        if(s->tune)
            delete s->tune;
        s->tune = NULL;

        sidTune *myTune = new sidTune(file);
        if (!myTune)
            return 0;

        struct sidTuneInfo mySidInfo;
        myTune->getInfo(mySidInfo);

        if (mySidInfo.nameString && strlen(mySidInfo.nameString))
		{
			strncpy(title, mySidInfo.nameString, 128);
		}

		if (mySidInfo.authorString && strlen(mySidInfo.authorString))
		{
			strncpy(artist, mySidInfo.authorString, 128);
		}

		if (mySidInfo.copyrightString && strlen(mySidInfo.copyrightString))
		{
			strncpy(copyright, mySidInfo.copyrightString, 128);
		}

        subtunes  = mySidInfo.songs;
		if( subtunes <= 0 ) subtunes = 1;
        s->song = mySidInfo.startSong;
        s->tune = myTune;

		if( db_found ) {
			xs_lengths = xs_sldb_get(xs_sldb_db, file);
			length = get_length(xs_lengths, 0);
		}
		else length = 200000;
    }

    if(!s->tune)
        return 0;

    if(!sidEmuInitializeSong(*(s->engine), *(s->tune), s->song))
        return 0;

    s->engine->resetSecondsThisSong();
    s->playing = true;

	playing = 1;
	return 1;
}

int load_rw( char *file, char *addr, int size )
{
	 // Initialize the SID-Emulator Engine to defaults.
    init_engine();
    SidState *s = state;

    if(s->tune)
        delete s->tune;

	s->tune = NULL;

    sidTune *myTune = new sidTune((const ubyte *)addr, size);
    if (!myTune)
        return 0;

    if(!myTune->getStatus())
    {
        delete myTune;
        return 0;
    }

    struct sidTuneInfo mySidInfo;
    myTune->getInfo(mySidInfo);

    if (mySidInfo.nameString && strlen(mySidInfo.nameString))
	{
		strncpy(title, mySidInfo.nameString, 128);
	}

	if (mySidInfo.authorString && strlen(mySidInfo.authorString))
	{
		strncpy(artist, mySidInfo.authorString, 128);
	}

	if (mySidInfo.copyrightString && strlen(mySidInfo.copyrightString))
	{
		strncpy(copyright, mySidInfo.copyrightString, 128);
	}

    subtunes  = mySidInfo.songs;
	if( subtunes <= 0 ) subtunes = 1;
    s->song = mySidInfo.startSong;
    s->tune = myTune;

	if( db_found ) {
		xs_lengths = xs_sldb_get_rw(xs_sldb_db, addr, size);
		length = get_length(xs_lengths, 0);
	}
	else length = 200000;

    if(!s->tune)
        return 0;

    if(!sidEmuInitializeSong(*(s->engine), *(s->tune), s->song))
        return 0;

    s->engine->resetSecondsThisSong();
    s->playing = true;

	playing = 1;
	return 1;
}

void set_tune( int tune )
{
	state->song = tune+1;

	state->playing = false;
	if(!sidEmuInitializeSong(*(state->engine), *(state->tune), state->song))
		return;

	if( db_found ) length = get_length(xs_lengths, tune);
	else length = 200000;

	state->engine->resetSecondsThisSong();
	state->playing = true;
}

void cleanup( void )
{
	if( state ) {
		if( state->tune ) delete state->tune;
		if( state->engine ) delete state->engine;
		delete state;
	}

	if( db_found ) {
		// free songlengths database
		xs_sldb_free(xs_sldb_db);
		xs_sldb_db = NULL;
	}

	__psp_free_heap();
}

int write_buffer( signed short *dest, int len )
{
	if(playing)
    {
#ifdef DOUBLE_MONO_SID
        signed short tmp[len>>2];
        sidEmuFillBuffer(*(state->engine), *(state->tune), tmp, len>>1);
        for (int i=0; i < len>>1; i++)
        {
            *dest++ = tmp[i];
            *dest++ = tmp[i];
        }
#else
        sidEmuFillBuffer(*(state->engine), *(state->tune), dest, len);
#endif
        return len;
    }

    return 0;
}

void get_song_title( char *filename, char **title, int *length )
{
}

void get_packet(void)
{
	switch( itable.in_packet->ID ) {
		case NO_PACKET:
		default:
			break;
	};
}

void send_packet(void)
{
	switch( itable.out_packet->ID ) {
		/* song data */
		case SONG_TITLE:
			itable.out_packet->data = title;
			break;
		case SONG_ARTIST:
			itable.out_packet->data = artist;
			break;
		case SONG_ALBUM_GAME:
			itable.out_packet->data = album;
			break;
		case SONG_COPYRIGHT:
			itable.out_packet->data = copyright;
			break;
		case SONG_FORMAT:
			itable.out_packet->data = format;
			break;
		case SONG_LENGTH:
			itable.out_packet->data = &length;
			break;
		case SONG_SUBTUNES:
			itable.out_packet->data = &subtunes;
			break;

		/* audio info */
		case AUDIO_CHANNELS:
			itable.out_packet->data = &channels;
			break;
		case AUDIO_SAMPLERATE:
			itable.out_packet->data = &samplerate;
			break;
		case AUDIO_FORMAT:
			itable.out_packet->data = &aformat;
			break;
		case AUDIO_KBPS:
			itable.out_packet->data = &kbps;
			break;

		/* extra data */
		case ATTACHED_PICTURE:
			itable.out_packet->size = 0;
			itable.out_packet->data = NULL;
			break;
		case NO_PACKET:
		default:
			itable.out_packet->size = 0;
			itable.out_packet->data = NULL;
			break;
	};
}


