
/*  Input Plugin for GameMusicGear ~
 *  Copyright (C) 2005 - 2010 gama, rockMax
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/* Default headers */

#include <pspkernel.h>
#include <pspdebug.h>

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <malloc.h>
#include <assert.h>
#include <sys/stat.h>

#include <ctype.h>

#include "../plugin.h"

extern "C" {
#include "amifilemagic.h"
#include "uadecontrol.h"
#include "uadeconf.h"
#include "include/uadeipc.h"
#include "mutex.h"
void uade_init();
void uade_go();
void uae_quit();
}

PSP_MODULE_INFO("IN_UADE", 0, 1, 1);
PSP_NO_CREATE_MAIN_THREAD();
PSP_HEAP_SIZE_KB( 1024 * 6 );

#ifdef __cplusplus
extern "C" {
#endif
int module_start(SceSize argc, void* argp)
{
 	return 0;
}

int module_stop(int args, void *argp)
{
	return 0;
}
#ifdef __cplusplus
}
#endif

/* Prototypes */
void init(void);
void reset(void);
int  load( char *filename );
void set_tune( int tune );
long  seek( long time );
void cleanup( void );
int  write_buffer( signed short *dest, int len );
void get_song_title( char *filename, char **title, int *length );

void get_packet(void);
void send_packet(void);
/* End Prototypes */

/* Variables */
char filename[256];
char uade_path[256];
char lastext[16] = "";
char current_format[80] = "";
int  uade_thread_id;

struct uade_config uadeconf;
uint8_t space[UADE_MAX_MESSAGE_SIZE];
struct uade_msg *um = (struct uade_msg *) space;
struct uade_ep_options ep_options;
int left = 0;
int next_song = 0;
int tailbytes = 0;
int playbytes = 0;

int uade_song_end_trigger = 0;
int subsong_end = 0;

enum uade_control_state state = UADE_S_STATE;

short soundbuffer[32768];
short *soundptr = soundbuffer;

int new_subsong = -1;
int cur_subsong = 0;
int playing = 0;
int song_info = 0;
static struct uade_ipc uadeipc;

/* Required */
int channels = 2;
int samplerate = 44100;
AFormat aformat = FMT_S16_LE;
int kbps = 128;

char title[128];
char artist[128];
char album[128];
char copyright[128];
char format[128];

long length;
int subsongs;
/* End Variables */

/* Extra Functions */
#define LINESIZE (1024)
#define WS_DELIMITERS " \t\n"
#define OPTION_DELIMITER ","

#ifdef DEBUG
#define _log(fmt, args...) do { pspDebugScreenPrintf("%s:%d: %s: " fmt, __FILE__, __LINE__, __func__, ## args); } while(0); \
						   sceKernelDelayThread(5 * 1000000)
#else
#define _log(fmt, args...)
#endif

char **split_line(size_t *nitems, size_t *lineno, FILE *f,
			 const char *delimiters)
{
	char line[LINESIZE], templine[LINESIZE];
	char **items = NULL;
	size_t pos;
	char *sp, *s;

	*nitems = 0;

	while (fgets(line, sizeof line, f) != NULL) {

		if (lineno != NULL)
			(*lineno)++;

		/* Skip, if a comment line */
		if (line[0] == '#')
			continue;

		/* strsep() modifies line that it touches, so we make a copy of it, and
		then count the number of items on the line */
		strlcpy(templine, line, sizeof(templine));
		sp = templine;
		while ((s = strsep(&sp, delimiters)) != NULL) {
			if (*s == 0)
				continue;
			(*nitems)++;
		}

		if (*nitems > 0)
			break;
	}

	if (*nitems == 0)
		return NULL;

	if ((items = (char **)malloc(sizeof(items[0]) * (*nitems + 1))) == NULL) {
		_log( "No memory for nws items.\n");
		sceKernelExitGame();
	}

	sp = line;
	pos = 0;
	while ((s = strsep(&sp, delimiters)) != NULL) {
		if (*s == 0)
			continue;
		if ((items[pos] = strdup(s)) == NULL) {
			_log( "No memory for an nws item.\n");
			sceKernelExitGame();
		}
		pos++;
	}
	items[pos] = NULL;

	if( pos != *nitems ) return NULL;

	return items;
}

char *uade_get_eagleplayer(const char *ext, const char *config, int *ep_attributes)
{
	FILE *f;
	char playername[128];
	int found = 0;
	size_t lineno = 0;
	size_t exti;
	size_t i;

	// Extensions
	int len = strlen(ext), tmp;

	f = fopen(config, "r");
	if (f == NULL)
		goto error;

	while (1) {
		char **items;
		size_t nitems;

		if ((items = split_line(&nitems, &lineno, f, WS_DELIMITERS)) == NULL)
			break;

		if( nitems <= 0 ) {
			break;
		}

		strcpy(playername, items[0]);

		// Reset eagleplayer attributes and options
		*ep_attributes = 0;
		memset(&ep_options, 0, sizeof(struct uade_ep_options));

		for (i = 1; i < nitems; i++) {
			if (strncasecmp(items[i], "prefixes=", 9) == 0) {
				char prefixes[LINESIZE];
				char *prefixstart = items[i] + 9;
				char *sp, *s;
				size_t pos;

				pos = 0;
				sp = prefixstart;
				while ((s = strsep(&sp, OPTION_DELIMITER)) != NULL) {
					if (*s == 0) continue;
					//_log("Ext: %s, prefix: %s\n", ext, s);
					tmp = strlen(s);
					if( tmp == len )
						if( strcasecmp(ext, s) == 0 ) {
							found = 1;
							break;
						}

					pos++;
				}
			}
			// Get attributes
			else if (strcasecmp(items[i], "content_detection") == 0) {
				*ep_attributes |= EP_CONTENT_DETECTION;
			}
			else if (strcasecmp(items[i], "detect_format_by_content") == 0) {
				*ep_attributes |= EP_CONTENT_DETECTION;
			}
			else if (strcasecmp(items[i], "a500") == 0) {
				*ep_attributes |= EP_A500;
			} else if (strcasecmp(items[i], "a1200") == 0) {
				*ep_attributes |= EP_A1200;
			} else if (strcasecmp(items[i], "always_ends") == 0) {
				*ep_attributes |= EP_ALWAYS_ENDS;
			} else if (strcasecmp(items[i], "speed_hack") == 0) {
				*ep_attributes |= EP_SPEED_HACK;
			} else if (strcasecmp(items[i], "ignore_player_check") == 0) {
				*ep_attributes |= ES_IGNORE_PLAYER_CHECK;
			} else if (strcasecmp(items[i], "broken_song_end") == 0) {
				*ep_attributes |= ES_BROKEN_SONG_END;
			} else if (strcasecmp(items[i], "detect_format_by_name") == 0) {
				*ep_attributes |= ES_NAME_DETECTION;
			} else if (strncasecmp(items[i], "epopt=", 6) == 0) {
				uade_add_ep_option(&ep_options, items[i] + 6);
			} else if (strncasecmp(items[i], "comment:", 8) == 0) {
				break;
			} else {
				_log( "Unrecognized option: %s\n", items[i]);
			}
		}

		// Free memory
		for (i = 0; i < nitems; i++)
			free(items[i]);
		free(items);
		if( found ) break;
    }

	fclose(f);

	if( found ) {
		return strdup(playername);
	}

	return NULL;

error:
	if( playername ) free(playername);
	if (f != NULL) fclose(f);
	return NULL;
}

/* SUPPORT */
size_t atomic_fread(void *ptr, size_t size, size_t nmemb, FILE *stream)
{
  uint8_t *dest = (uint8_t *)ptr;
  size_t readmembers = 0;
  size_t ret;

  while (readmembers < nmemb) {
    ret = fread(dest + size * readmembers, size, nmemb - readmembers, stream);
    if (ret == 0)
      break;
    readmembers += ret;
  }

  assert(readmembers <= nmemb);

  return readmembers;
}

char *analyze_format(const char *name, const char *lname, int *ep_attributes)
{
	const char *sext = strrchr(lname, '.');
	char prefix[strlen(lname)+1];

	char ext[128];
	size_t bufsize, bytesread;
	uint8_t buf[8192];

	// Get some magic
	FILE *f;
	int size;
	if ((f = fopen(name, "rb")) == NULL) {
		_log("ERROR\n");
		return NULL;
	}

	// Get file size
	size = fseek(f, 0, SEEK_END);
	fseek(f, 0, SEEK_SET);

	bufsize = sizeof buf;
	bytesread = atomic_fread(buf, 1, bufsize, f);
	fclose(f);
	if (bytesread == 0)
		return NULL;

	memset(&buf[bytesread], 0, bufsize - bytesread);

	uade_filemagic(buf, bytesread, ext, size, name, 1);

	if (strcmp(ext, "reject") == 0)
		return NULL;

	if (ext[0] != 0)
		_log("Content recognized: %s (%s)\n", ext, name);

	if (strcmp(ext, "packed") == 0)
		return NULL;

	// Try prefix and extension first
	if( !sext )
    {
		sext = "";
        prefix[0] = '\0';
    }
	else
    {
		sext++;
        const char* name2 = strrchr(lname,'/');
        name2 = name2 ? name2+1 : lname;
        strcpy(prefix,name2);
        *strchr(prefix,'.') = '\0';

		/* HACK */
		char *tmp = strchr(prefix, '~');
		if( tmp ) *tmp = '\0';
    }

	char tmp[256];
	sprintf(tmp, "%s/eagleplayer.conf", uade_path);
	int i;
	for(i = 0; i < strlen(prefix); i++) prefix[i] = tolower(prefix[i]);

	/* First do filename detection (we'll later do content detection) */
	char *name_pl = uade_get_eagleplayer(prefix, tmp, ep_attributes);
	if( name_pl == NULL ) {
		strncpy(title, prefix, 128);
		name_pl = uade_get_eagleplayer(sext, tmp, ep_attributes);
	}
	else {
		strncpy(title, sext, 128);
	}

	// If name detection skip content detection
	if( name_pl && (*ep_attributes & ES_NAME_DETECTION) ) return name_pl;

	/* If filemagic found a match, we'll use player plugins associated with
	   that extension */
	if (ext[0]) {
		int _ep_attributes = 0;
		char *content_pl = uade_get_eagleplayer(ext, tmp, &_ep_attributes);
		if( content_pl ) {
			*ep_attributes = _ep_attributes;
			return content_pl;
		}
	}

	return name_pl;
}

int cprintf(char *fmt, ...);
unsigned int inline htonl(unsigned int l)
{
	return ((l>>24) & 0xff) | ((l>>8) & 0xFF00) | ((l<<8) & 0xFF0000) | (l<<24);
}

unsigned int inline ntohl(unsigned int l)
{
	return ((l>>24) & 0xff) | ((l>>8) & 0xFF00) | ((l<<8) & 0xFF0000) | (l<<24);
}

short inline ntohs(short l)
{
	return ((l>>8) & 0xff) | ((l<<8) & 0xFF00);
}

extern "C" {
	void client_sleep()
	{
		sceKernelDelayThread(500);
	}

	int get_write_mutex(void **m)
	{
		if( !*m ) *m = create_mutex();
		return (mutexp((mutex*)*m) == 0);
	}

	int release_write_mutex(void **m)
	{
		return (mutexv((mutex*)*m) == 0);
	}

	int get_read_mutex(void **m)
	{
		if( !*m ) *m = create_mutex();
		return (mutexp((mutex*)*m) == 0);
	}

	int release_read_mutex(void **m)
	{
		return (mutexv((mutex*)*m) == 0);
	}
}

static int uade_thread(SceSize args, void *argp)
{
    _log("SERVER\n");
	uade_init();
	uade_go();
	_log("SERVER END\n");
	return 0;
}

int run_client()
{
	char *reason;
	int i;

	if(state == UADE_S_STATE)
	{
		if(uade_song_end_trigger || subsong_end)
			return -1;

		if(new_subsong >= 0)
		{
			uade_change_subsong(new_subsong, &uadeipc);
			new_subsong = -1;
		}

		left = uade_read_request(&uadeipc); // Request another batch of sample data from uadecore

		if(uade_send_short_message(UADE_COMMAND_TOKEN, &uadeipc))
		{
			_log( "Can not send token\n");
			return 0;
		}

		state = UADE_R_STATE;
	}

	/* receive state */
	if(state == UADE_R_STATE)
	{
		uint16_t *sm;

		if (uade_receive_message(um, sizeof(space), &uadeipc) <= 0)
		{
			_log(  "Can not receive events from uade\n");
			return 0;
		}

		switch (um->msgtype)
		{
		case UADE_COMMAND_TOKEN:
			state = UADE_S_STATE;
			break;

		case UADE_REPLY_DATA:
			sm = (uint16_t *)um->data;

			if(subsong_end)
			{
				playbytes = tailbytes; // Determined by UADE_REPLY_SONG_END
				tailbytes = 0;
			}
			else
				playbytes = um->size;


			if(samplerate == 44100)
			{
				int i;
				for (i = 0; i < um->size/2; i++)
					soundptr[i] = ntohs(sm[i]);

				soundptr += (playbytes/2);
			}
			else
			if(samplerate == 22050)
			{
				int i;
				for (i = 0; i < um->size/2; i+=4)
				{
					soundptr[i/2] = (ntohs(sm[i]) + ntohs(sm[i+2])) / 2;
					soundptr[i/2+1] = (ntohs(sm[i+1]) + ntohs(sm[i+3])) / 2;
				}

				soundptr += (playbytes/4);
			}

			assert(left >= um->size );
			left -= um->size;
			break;

		case UADE_REPLY_FORMATNAME:
			uade_check_fix_string(um, 128);
			_log(  "Format name: %s\n", (char*) um->data);
			strcpy(current_format, (char *)um->data);
			break;

		case UADE_REPLY_MODULENAME:
			uade_check_fix_string(um, 128);
			_log(  "Module name: %s\n", (char*) um->data);
			if( strncmp((char*)um->data, "<no songtitle>", 14) )
				strcpy(title, (char *)um->data);
			break;

		case UADE_REPLY_MSG:
			uade_check_fix_string(um, 128);
			_log(  "Message: %s\n", (char *) um->data);
			break;

		case UADE_REPLY_PLAYERNAME:
			uade_check_fix_string(um, 128);
			_log(  "Player name: %s\n", (char*) um->data);
			strcpy(album, (char *)um->data);
			break;

		case UADE_REPLY_SONG_END:
			if (um->size < 9)
			{
				_log(  "Invalid song end reply\n");
			}

			tailbytes = ntohl(((uint32_t *) um->data)[0]);
			/* next ntohl() is only there for a principle. it is not useful */
			if (ntohl(((uint32_t *) um->data)[1]) == 0) {
				/* normal happy song end. go to next subsong if any */
				subsong_end = 1;

			} else {
				/* unhappy song end (error in the 68k side). skip to next song
				ignoring possible subsongs */
				uade_song_end_trigger = 1;
			}

			_log( "Song end signal received\n");
			break;

		case UADE_REPLY_SUBSONG_INFO:
			if(um->size != 12)
			{
				_log( "subsong info: too short a message\n");
			}

			{
				unsigned int *u32ptr = (unsigned int *)um->data;
				int min_sub = ntohl(u32ptr[0]);
				int max_sub = ntohl(u32ptr[1]);
				int cur_sub = ntohl(u32ptr[2]);
				subsongs = max_sub ? max_sub : 1;
				_log( "subsong: %d from range [%d, %d]\n", cur_sub, min_sub, max_sub);
			}

			song_info = 1;
			break;
		case UADE_REPLY_CANT_PLAY:
		case UADE_REPLY_CAN_PLAY:


		default:
			_log(  "Expected sound data. got %d.\n", um->msgtype);
			return 0;
		}
	}

}

static int wait_token()
{
	do
	{
		int ret = uade_receive_message(um, sizeof(space), &uadeipc);
		if(ret < 0)
		{
			_log( "Can not receive events (TOKEN) from uade.\n");
			return 0;
		}
		if (ret == 0)
		{
			_log( "End of input after reboot.\n");
			return 0;
		}
	} while (um->msgtype != UADE_COMMAND_TOKEN);
	return 1;
}

/* End Extra Functions */

/** START of Plugin definitions setup */
InputPlugin itable =
{
	/* Populated by Plugin */
	"Unix_Amiga_Delitracer_Emulator", 				/* The description of the input plugin */

	init,							/* Called when the plugin is loaded */
	reset, 							/* Called when need to update config */
	load,							/* Load the file */
	NULL,							/* Load the file from data */
	set_tune,						/* Set song subtune */
	NULL,							/* Seek to the specified time */
	NULL,							/* Set the equalizer, most plugins won't be able to do this */
	cleanup,						/* Called when gmgear exit */
	write_buffer, 					/* Fill audio buffer */
	get_song_title,					/* Function to grab the title string */

	get_packet,					/* Called when need main app data packet*/
	send_packet,				/* Send data packet to main app */

	NULL,						/* Filled by Main App */
	NULL,						/* Filled by Main App */
	NULL,						/* Filled by Main App */
	0,								/* Filled by Main App */
	LATIN,							/* Filled by Main App */
	CLIENT_PLAYER					/* Player type */
};
/** END of Plugin definitions setup */

/** Single Plugin Export */
InputPlugin *get_iplugin_info()
{
	return &itable;
}

void init(void)
{
 	SceKernelThreadInfo status;
	int priority = 32;
	char tmp[256];

	strcpy(format, "Amiga");
	subsongs = 1;
	playing = 0;

	kbps = 0x6C;
	aformat = FMT_S16_LE;

#ifdef DEBUG
	pspDebugScreenInit();
#endif

	const char *token = ReadStringValue("uade_path", itable.config);
	if( token ) strncpy(uade_path, token, 256);
	else strncpy(uade_path, "ms0:/.gmgear", 256);

	uade_set_peer(&uadeipc, 1, "client", "server");

	// Create UADE Thread
	/* Set priority of new thread to the same as the current thread */
	status.size = sizeof(SceKernelThreadInfo);
	if (sceKernelReferThreadStatus(sceKernelGetThreadId(), &status) == 0) {
		priority = status.currentPriority;
	}

	uade_thread_id = sceKernelCreateThread("UADE Thread", uade_thread,
					       priority, 0x8000,
					       PSP_THREAD_ATTR_USER, NULL);
	if ( uade_thread_id < 0 ) {
		_log("thread error\n");
		sceKernelExitGame();
	}
	sceKernelStartThread(uade_thread_id, 0, NULL);
	// End UADE Thread

	sprintf(tmp, "%s/uaerc", uade_path);
	if (uade_send_string(UADE_COMMAND_CONFIG, tmp, &uadeipc))
    {
		sceKernelDeleteThread(uade_thread_id);
	    sceKernelExitGame();
	}

	// Wait 1/2 second for things to finish
	sceKernelDelayThread(1000000);
	itable.in_packet->ID = NO_PACKET;
}

void reset(void)
{
	samplerate = ReadFreqValue("samplerate", itable.config);

	strcpy(title, "");
	strcpy(album, "");
	strcpy(copyright, "");
	strcpy(artist, "");
	strcpy(filename, "");

	length = 0;
	subsongs = 1;
	cur_subsong = 0;

	if( playing ) {
		if(state == UADE_R_STATE)
			wait_token();

		if(uade_send_short_message(UADE_COMMAND_REBOOT, &uadeipc))
		{
			_log( "uadeplugin:close - Can not send reboot\n");
			return;
		}

		if (uade_send_short_message(UADE_COMMAND_TOKEN, &uadeipc))
		{
			_log( "uadeplugin:close - Can not send token\n");
			return;
		}

		wait_token();
		state = UADE_S_STATE;
		playing = 0;
	}

	// clear sound buffer
	memset(soundbuffer, 0, sizeof(soundbuffer));
	soundptr = soundbuffer;

	itable.in_packet->ID = NO_PACKET;
}

int load( char *file )
{
	/* HACK - to allow in_uade handle the extensions correctly we need
		     a long name */
	length = 800000;
	strncpy(filename, file, 256);
	itable.in_packet->ID = SONG_LONGFILENAME;
	return 1;
}

int _load( char *file, char *lname )
{
	char *playername;
	char tmp[256], tmp2[256], score[256];
	char *src = file;
	char *dst = tmp2;
	int rc;

	int ep_attributes;
	playername = analyze_format(file, lname, &ep_attributes);
	if( playername == NULL ) return 0;

	pspDebugScreenPrintf("Player: %s\n", playername);
	sprintf(tmp, "%s/players/%s", uade_path, playername);

	while(*src)
	{
		if(*src == '\\')
			*dst++ = '/';
		else
			*dst++ = *src;
		src++;
	}
	*dst = 0;

	strcpy(current_format, "");
	sprintf(score, "%s/score", uade_path);
	if(strcmp("custom", playername) == 0)
	{
		rc = uade_song_initialization(score, tmp2, "", &uadeipc);
		strcpy(current_format, "Amiga Custom");
	}
	else
		rc = uade_song_initialization(score, tmp, tmp2, &uadeipc);

	if(rc) return 0;
    if( playername ) free( playername );

	uade_song_end_trigger = 0;
	subsong_end = 0;

	uade_config_set_defaults(&uadeconf);

	// Set Eagleplayer attributes
	if (ep_attributes & ES_IGNORE_PLAYER_CHECK) {
		if (uade_send_short_message(UADE_COMMAND_IGNORE_CHECK, &uadeipc) < 0) {
			_log("Can not send ignore check message.\n");
			return -1;
		}
    }

    if (ep_attributes & ES_BROKEN_SONG_END) {
		if (uade_send_short_message(UADE_COMMAND_SONG_END_NOT_POSSIBLE, &uadeipc) < 0) {
			_log("Can not send 'song end not possible'.\n");
			return -2;
		}
    }

	// Disable filters
	uadeconf.filter_type = 0;

    uade_send_filter_command(uadeconf.filter_type, uadeconf.led_state, uadeconf.led_forced, &uadeipc);
    uade_send_interpolation_command(uadeconf.interpolator, &uadeipc);

	// Speed hack
	if (ep_attributes & ES_SPEED_HACK) {
		if (uade_send_short_message(UADE_COMMAND_SPEED_HACK, &uadeipc)) {
			_log("Can not send speed hack command.\n");
			return -3;
		}
    }

	// Send ep options
	if (uade_send_ep_options(&ep_options, &uadeipc))
		return -4;

	// Try some times to get song info
	int tries = 32;
	song_info = 0;
	while( tries-- > 0 ) {
		if (song_info) break;
		run_client();
		sceKernelDelayThread(100);
	}

	playing = 1;
	return 1;
}

void set_tune( int tune )
{
	new_subsong = tune;
	cur_subsong = tune;
	subsong_end = 0;
}

void cleanup( void )
{
	reset();
	uae_quit();
	sceKernelTerminateDeleteThread(uade_thread_id);

	// Wait a little for terminate thread
	sceKernelDelayThread(500000);
	__psp_free_heap();
}

int write_buffer( signed short *dest, int len )
{
	if (playing)
    {
        if(run_client() < 0) {
			memset(dest, 0, len);
			return len;
		}

		int filled = (soundptr - soundbuffer)*2;

		while(filled < len) {
			if(run_client() < 0) {
				memset(dest, 0, len);
				return 0;
			}
			filled = (soundptr - soundbuffer)*2;
		}

		{
			memcpy(dest, soundbuffer, len);
			memmove(soundbuffer, soundbuffer+(len/2), filled-len);
			soundptr -= (len/2);
		}

		return len;
    }

    return 0;
}

void get_song_title( char *filename, char **title, int *length )
{
}

void get_packet(void)
{
	switch( itable.in_packet->ID ) {
		case SONG_LONGFILENAME:
			if( itable.in_packet->data ) {
				_load(filename, (char *)itable.in_packet->data);
			}
			itable.in_packet->ID = NO_PACKET;
			break;
		case NO_PACKET:
		default:
			break;
	};
}

void send_packet(void)
{
	switch( itable.out_packet->ID ) {
		/* song data */
		case SONG_TITLE:
			itable.out_packet->data = title;
			break;
		case SONG_ARTIST:
			itable.out_packet->data = artist;
			break;
		case SONG_ALBUM_GAME:
			itable.out_packet->data = album;
			break;
		case SONG_COPYRIGHT:
			itable.out_packet->data = copyright;
			break;
		case SONG_FORMAT:
			itable.out_packet->data = format;
			break;
		case SONG_LENGTH:
			itable.out_packet->data = &length;
			break;
		case SONG_SUBTUNES:
			itable.out_packet->data = &subsongs;
			break;

		/* audio info */
		case AUDIO_CHANNELS:
			itable.out_packet->data = &channels;
			break;
		case AUDIO_SAMPLERATE:
			itable.out_packet->data = &samplerate;
			break;
		case AUDIO_FORMAT:
			itable.out_packet->data = &aformat;
			break;
		case AUDIO_KBPS:
			itable.out_packet->data = &kbps;
			break;

		/* extra data */
		case ATTACHED_PICTURE:
			itable.out_packet->size = 0;
			itable.out_packet->data = NULL;
			break;
		case NO_PACKET:
		default:
			itable.out_packet->size = 0;
			itable.out_packet->data = NULL;
			break;
	};
}


