
/*  GMGear - Game Music Gear for PSP
 *  Copyright (C) 2005 - 2010 gama, rockMax
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef PLUGIN_H
#define PLUGIN_H

#include <pspkernel.h>
#include "SharedLib/libconfig/config.h"

#ifdef __cplusplus
extern "C" int __psp_free_heap(void);
#else
int __psp_free_heap(void);
#endif

typedef enum {
	FMT_U8 = 0x0, FMT_S8, FMT_U16_LE, FMT_U16_BE, FMT_U16_NE, FMT_S16_LE, FMT_S16_BE, FMT_S16_NE
} AFormat;

typedef enum {
	CLIENT_PLAYER = 0x0,
	STANDALONE_PLAYER
} PlayerType;

typedef enum {
	NO_PACKET = 0x0,
	SONG_FULLPATH,
	SONG_FILENAME,
	SONG_LONGFILENAME,
	SONG_TITLE,
	SONG_ARTIST,
	SONG_ALBUM_GAME,
	SONG_COPYRIGHT,

	SONG_FORMAT,
	SONG_LENGTH,
	SONG_SUBTUNES,

	AUDIO_SAMPLERATE,
	AUDIO_CHANNELS,
	AUDIO_FORMAT,
	AUDIO_KBPS,

	PLAYTIME,

	ATTACHED_PICTURE,
	ATTACHED_LYRICS,

	SCE_CTRL_DATA,

	// STANDALONE_PLAYER commands
	CMD_PAUSE,
	CMD_RESUME,
	CMD_STOP,
	CMD_VOLUP,
	CMD_VOLDOWN,
	CMD_SET_EQUALIZER,
	CMD_SEND_VISUAL_BUFFER,
	CMD_SEND_PLAYTIME,
	CMD_SEND_TOTALSAMPLES,

	CMD_REQUEST_XSFLIBN,
	CMD_REQUEST_XSFLIBN_FILE,
	CMD_FREE_XSFLIBN,

	// System resources
	SYSTEM_FONT,

	CMD_REQUEST_MALLOC,
	CMD_REQUEST_FREE

} QueryID;

typedef struct
{
	int channels;
	int samplerate;
	int kbps;
	AFormat format;
} AProperties;

typedef struct
{
	void *data;
	int size;
	QueryID ID;
} Packet;


/* Based on XMMS's and PSPRadio plugin format */

typedef enum {
	LATIN = 0, UTF8
} Charset;

typedef struct
{
	const char *description;						/* The description of the input plugin */

	void (*init) (void);					/* Called when the plugin is loaded */
	void (*reset) (void); 					/* Called when need to update config */
	int  (*load) (char *filename);
	int  (*load_rw) (char *filename, char *addr, int size);

	void (*set_tune) (int tune);			/* Guess what... */
	long (*seek) (long time);				/* Seek to the specified time */
	void (*set_eq) (int on, float preamp, float *bands);	/* Set the equalizer, most plugins won't be able to do this */
	void (*cleanup) (void);					/* Called when gmgear exit */

	int  (*write_buffer)(signed short *dest, int len); /* Fill audio buffer */

	void (*get_song_title) (char *filename, char **title, int *length);	/* Function to grab the title string */

	void (*get_packet)(void);
	void (*send_packet)(void);

	configType *config;
	Packet *in_packet;
	Packet *out_packet;

	SceUID ID;

	Charset charset;
	PlayerType type;
} InputPlugin;


typedef struct
{
	const char *description;		/* Description of the plugin */
	const char *ext;				/* Format extension */

	void (*open) (char *finelame);	/* Called when the plugin is started */
	void (*reset) (void); 			/* Called when need to update config */
	void (*close) (void);			/* Called when the plugin is closed */

	void (*write)(short *buf, long count);	/* Called when need plugin write */

	long (*sample_count)(void);

	void (*get_packet)(void);
	void (*send_packet)(void);

	configType *config;
	Packet *in_packet;
	Packet *out_packet;

	SceUID ID;
} OutputPlugin;


typedef struct
{
	const char *description;		/* Description of the plugin */

	void (*init) (void);	/* Called when the plugin is loaded */
	void (*reset) (void); 	/* Called when need to update config */
	void (*cleanup) (void);			/* Called when gmgear exit */
	//int need_3D(void);		/* Called by main app */

	void (*update)(void);
	void (*update_pcm)(short *pcm_buffer, int len);	/* Called when need plugin update */
	void (*update_freq)(float freq_data[2][257]);

	void (*render)(void *fbp0);

	void (*get_packet)(void);
	void (*send_packet)(void);

	configType *config;
	Packet *in_packet;
	Packet *out_packet;

	SceUID ID;

} VisualPlugin;


/* The plugin exports this function */
#ifdef __cplusplus
extern "C" InputPlugin *get_iplugin_info();
extern "C" VisualPlugin *get_vplugin_info();
extern "C" OutputPlugin *get_oplugin_info();
#endif


#endif /* PLUGIN_H */
